/*
 * Author: Tony Olech - Elan Digital Systems Limited
 * $Revision: 1.37 $
 * $Date: 2007/12/20 15:06:18 $
 * 
 * VMB HCD (Host Controller Driver) for USB.
 *
 * Interfaces to kernel 2.6.19
 *
 */

#include <linux/version.h>
#include <linux/kernel.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
#error you need at least 2.6.13 to compile the driver
#endif
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/usb.h>
#include <linux/workqueue.h>
#include <linux/platform_device.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/byteorder.h>

#include "../../usb/core/hcd.h"
/* #include "/usr/src/linux/drivers/usb/core/hcd.h" */

#define DRIVER_VERSION "$Revision: 1.1 $"

MODULE_AUTHOR("Tony Olech");
MODULE_VERSION(DRIVER_VERSION);
MODULE_DESCRIPTION("VMB USB Host Controller Driver for 2.6.19+");
MODULE_LICENSE("GPL");

#define PRINTK(args...) do{}while(0)

#define LOCK_DEBUGGING_COMPILED_IN 0
#if defined(DEBUG) || defined(CONFIG_USB_DEBUG) || defined(PCMCIA_DEBUG)

#define DEBUGGING_COMPILED_IN 1

static int pc_debug = 0; module_param(pc_debug, int, 0644);
static int pc_trace = 0; module_param(pc_trace, int, 0644);
static int pc_retry = 0; module_param(pc_retry, int, 0644);

#define	DBG(n, args...) do{}while(0)
#define	URB_DBG(n, args...) do{}while(0)
#define	SCHED_DBG(n, args...) do{}while(0)
#define URB_QUEUED(vmb,urb,slot) do{}while(0)
#define URB_DEBUG(n,urb) do{}while(0)
#define URB_SETUP(n,urb) do{}while(0)
#define PTR_DEBUG(n,vmb) do{}while(0)
#define SLOT_DEBUG(n,slot) do{}while(0)
#define VMB_DEBUG(n,vmb) do{}while(0)
#define VMB_DEBUG_LOCKED(n,vmb) do{}while(0)
#define UTL_DEBUG(n,vmb) do{}while(0)
#define QUEUE_DEBUG(n,vmb) do{}while(0)

#else
#define DEBUGGING_COMPILED_IN 0
#define DBG(n, args...) do{}while(0)
#define URB_DBG(n, args...) do{}while(0)
#define	SCHED_DBG(n, args...) do{}while(0)
#define URB_QUEUED(vmb,urb,slot) do{}while(0)
#define URB_DEBUG(n,urb) do{}while(0)
#define URB_SETUP(n,urb) do{}while(0)
#define PTR_DEBUG(n,vmb) do{}while(0)
#define SLOT_DEBUG(n,slot) do{}while(0)
#define VMB_DEBUG(n,vmb) do{}while(0)
#define VMB_DEBUG_LOCKED(n,vmb) do{}while(0)
#define UTL_DEBUG(n,vmb) do{}while(0)
#define QUEUE_DEBUG(n,vmb) do{}while(0)
#endif  /* no debugging */

#define INT_MODULE_PARM(n, v) static int n = v; module_param(n, int, 0444)

#define ERR(stuff...)     PRINTK(KERN_ERR "vmb_hcd: " stuff)
#define WARN(stuff...)    PRINTK(KERN_WARNING "vmb_hcd: " stuff)
#define INFO(stuff...)    PRINTK(KERN_INFO "vmb_hcd: " stuff)

#define ASSERT(cond) do{}while(0)
#define UTL_ASSERT(cond) do{}while(0)
#define URB_ASSERT(cond) do{}while(0)
#define SLOT_ASSERT(cond) do{}while(0)
#define TLER_ASSERT(cond) do{}while(0)


#define PKT_DEBUG(n,urb) do{}while(0)
#define PKT_ERROR(n,vmb) do{}while(0)

/*-------------------------------------------------------------------------*/
#if LOCK_DEBUGGING_COMPILED_IN
#error only use in phase one testing because it will eventually ASSERT in the (very very small) hole
#define HW_LOCK(vmb) unsigned long  flags;if(vmb->hw_flag) PRINTK(KERN_ERR "%s line %d ****LOCK ERROR**** previously locked at line %d\n",__FUNCTION__,__LINE__,vmb->hw_line);ASSERT(0==vmb->hw_flag++);spin_lock_irqsave(&vmb->hw_lock, flags);vmb->hw_line=__LINE__
#define HW_UNLOCK(vmb) ASSERT(--vmb->hw_flag==0); spin_unlock_irqrestore(&vmb->hw_lock, flags)
#define HW_ASSERT_LOCKED(vmb)	ASSERT(1==vmb->hw_flag)

#define SW_LOCK(vmb) {ASSERT(! in_atomic());if(vmb->sw_flag) PRINTK(KERN_ERR "%s line %d ****LOCK ERROR**** previously locked at line %d\n",__FUNCTION__,__LINE__,vmb->sw_line);ASSERT(0==vmb->sw_flag++);down(&vmb->sw_lock);vmb->sw_line=__LINE__;}
#define SW_UNLOCK(vmb) {ASSERT(--vmb->sw_flag==0); up(&vmb->sw_lock);}
#define SW_ASSERT_LOCKED(vmb)	ASSERT(1==vmb->sw_flag)
#else
#define HW_LOCK(vmb) unsigned long  flags; spin_lock_irqsave(&vmb->hw_lock, flags)
#define HW_UNLOCK(vmb) spin_unlock_irqrestore(&vmb->hw_lock, flags)
#define HW_ASSERT_LOCKED(vmb)

#define SW_LOCK(vmb) down(&vmb->sw_lock)
#define SW_UNLOCK(vmb) up(&vmb->sw_lock)
#define SW_ASSERT_LOCKED(vmb)
#endif
static int max_assert_reports = 100; 
static void assertion_failure(const char*function,const char*filename,const unsigned line,const char*failure)
{
	if (max_assert_reports > 0) {
		max_assert_reports--;
		PRINTK(KERN_ERR "ASSERTION(%s) in %s at %s:%d\n",failure,function,filename,line);
		dump_stack();
		if (!in_atomic()) msleep(2);
		pc_debug=10;
	}
}
/*-------------------------------------------------------------------------*/
static spinlock_t vmb_static_lock;
static struct list_head vmb_static_list;
static int vmb_instances;
static struct workqueue_struct*workqueue;
/*-------------------------------------------------------------------------*/
static int in_vmb_scheduler = 0;
/*-------------------------------------------------------------------------*/
#include "vmb_hcd.h"
#include "../vmb_cs/usb_vmb.h"
static void slot_debug(const char*function,const unsigned line,struct vmb_slot *slot);
static void urb_debug(const char*function,const unsigned line,struct urb *urb);
static void utl_debug(const char*function,const unsigned line,struct vmb *vmb);

static void slot_assertion_failure(const char*function,const char*filename,const unsigned line,const char*failure,struct vmb_slot *slot)
{
	if (max_assert_reports > 0) {
		max_assert_reports--;
		PRINTK(KERN_ERR "SLOT_ASSERTION(%s) in %s SLOT[%X] at %s:%d\n",failure,function,slot->td,filename,line);
		slot_debug(function,line,slot);
		dump_stack();
		if (!in_atomic()) {
			PRINTK(KERN_ERR "sleeping to allow diagnostics to escape\n");
			msleep(2);
		} else {
			PRINTK(KERN_ERR "end of assertion failure report I=%lX A=%d\n",in_interrupt(),in_atomic());
		}
		pc_debug=10;
	}
}
static void urb_assertion_failure(const char*function,const char*filename,const unsigned line,const char*failure,struct urb *urb)
{
	if (max_assert_reports > 0) {
		max_assert_reports--;
		PRINTK(KERN_ERR "URB_ASSERTION(%s) in %s at %s:%d\n",failure,function,filename,line);
		urb_debug(function,line,urb);
		dump_stack();
		if (!in_atomic()) {
			PRINTK(KERN_ERR "sleeping to allow diagnostics to escape\n");
			msleep(2);
		} else {
			PRINTK(KERN_ERR "end of assertion failure report I=%lX A=%d\n",in_interrupt(),in_atomic());
		}
		pc_debug=10;
	}
}
static void utl_assertion_failure(const char*function,const char*filename,const unsigned line,const char*failure,struct vmb *vmb)
{
	if (max_assert_reports > 0) {
		max_assert_reports--;
		PRINTK(KERN_ERR "UTL_ASSERTION(%s) in %s at %s:%d\n",failure,function,filename,line);
		utl_debug(function,line,vmb);
		dump_stack();
		if (!in_atomic()) {
			PRINTK(KERN_ERR "sleeping to allow diagnostics to escape\n");
			msleep(2);
		} else {
			PRINTK(KERN_ERR "end of assertion failure report I=%lX A=%d\n",in_interrupt(),in_atomic());
		}
		pc_debug=10;
	}
}
static void tler_assertion_failure(const char*function,const char*filename,const unsigned line,const char*failure,struct vmb *vmb,struct vmb_slot *slot)
{
	if (max_assert_reports > 0) {
		max_assert_reports--;
		PRINTK(KERN_ERR "TLER_ASSERTION(%s) in %s SLOT[%X] at %s:%d\n",failure,function,slot->td,filename,line);
		utl_debug(function,line,vmb);
		slot_debug(function,line,slot);
		dump_stack();
		if (!in_atomic()) {
			PRINTK(KERN_ERR "sleeping to allow diagnostics to escape\n");
			msleep(2);
		} else {
			PRINTK(KERN_ERR "end of assertion failure report I=%lX A=%d\n",in_interrupt(),in_atomic());
		}
		pc_debug=10;
	}
}
inline void vmb_debug_with_lock_held(const char*function,const unsigned line,struct vmb *vmb)
{
	PRINTK(KERN_DEBUG "%-21s %4d(CR=%02X SR=%02X TL=%02X TP=%02X PC=%02X FC=%04X rxIT=%04X txIT=%04X %08X*%08X)\n",function,line
			,read_cr(vmb),read_sr(vmb),read_tl(vmb),read_txpc(vmb),read_pcsr(vmb),read_fc(vmb),read_rxit(vmb),read_txit(vmb)
			,vmb->port[0].Status,vmb->port[1].Status);
}
static void vmb_debug(const char*function,const unsigned line,struct vmb *vmb)
{
	HW_LOCK(vmb);
	vmb_debug_with_lock_held(function,line,vmb);
	HW_UNLOCK(vmb);
}
#if DEBUGGING_COMPILED_IN
static void print_queued_urb(const char*function,struct vmb *vmb,struct urb *urb,struct vmb_slot*slot)
{
	int type = usb_pipetype(urb->pipe);
	
	char*pipe_type = PIPE_ISOCHRONOUS==type?"ISOCHRONOUS":PIPE_INTERRUPT==type?"INTERRUPT":PIPE_CONTROL==type?"CONTROL":"BULK" ;

	if (usb_pipein(urb->pipe)) {

		u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));

		if (PIPE_ISOCHRONOUS==type) {
			PRINTK(KERN_DEBUG "%s(S=%X Q=%d X=%d TD=%X %02X:urb=%p e_c=%d start_frame=%d interval=%d INPUT<-%d:%02d %s)\n"
					,function,vmb->slot_count,slot->queue_size,slot->extra,slot->td,slot->queue_next,urb,urb->error_count,urb->start_frame,urb->interval,usb_pipedevice(urb->pipe),usb_pipeendpoint(urb->pipe),pipe_type);
			int pkts = urb->number_of_packets;
			int p = 0;
			char isochronous[128];
			int len = sprintf(isochronous,"pkts=%d len=%d max=%d",urb->number_of_packets,urb->transfer_buffer_length,maxusbpktlen);
			while(pkts-- > 0) {
				struct usb_iso_packet_descriptor*iso_frame_desc = &urb->iso_frame_desc[p++];
				int left = sizeof(isochronous)/sizeof(char) - len;

				len += snprintf(len+isochronous,left," o=%d l=%d a=%d",iso_frame_desc->offset,iso_frame_desc->length,iso_frame_desc->actual_length);
			}
			PRINTK(KERN_DEBUG "%s\n",isochronous);
		} else {
			PRINTK(KERN_DEBUG "%s(S=%X Q=%d X=%d TD=%X %02X:urb=%p error_count=%d len=%d max=%d INPUT<-%d:%02d %s)\n",function,vmb->slot_count,slot->queue_size,slot->extra,slot->td,slot->queue_next,urb,urb->error_count,urb->transfer_buffer_length,maxusbpktlen,usb_pipedevice(urb->pipe),usb_pipeendpoint(urb->pipe),pipe_type);
		}
	} else {

		u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));

		if (PIPE_ISOCHRONOUS==type) {
			PRINTK(KERN_DEBUG "%s(S=%X Q=%d X=%d TD=%X %02X:urb=%p e_c=%d start_frame=%d interval=%d OUTPUT->%d:%02d %s)\n",function,vmb->slot_count,slot->queue_size,slot->extra,slot->td,slot->queue_next,urb,urb->error_count,urb->start_frame,urb->interval,usb_pipedevice(urb->pipe),usb_pipeendpoint(urb->pipe),pipe_type);
			int pkts = urb->number_of_packets;
			int p = 0;
			char isochronous[128];
			int len = sprintf(isochronous,"pkts=%d len=%d max=%d",urb->number_of_packets,urb->transfer_buffer_length,maxusbpktlen);
			while(pkts-- > 0) {
				struct usb_iso_packet_descriptor*iso_frame_desc = &urb->iso_frame_desc[p++];
				int left = sizeof(isochronous)/sizeof(char) - len;

				len += snprintf(len+isochronous,left," o=%d l=%d a=%d",iso_frame_desc->offset,iso_frame_desc->length,iso_frame_desc->actual_length);
			}
			PRINTK(KERN_DEBUG "%s\n",isochronous);
		} else {
			PRINTK(KERN_DEBUG "%s(S=%X Q=%d X=%d TD=%X %02X:urb=%p error_count=%d len=%d max=%d OUTPUT->%d:%02d %s)\n",function,vmb->slot_count,slot->queue_size,slot->extra,slot->td,slot->queue_next,urb,urb->error_count,urb->transfer_buffer_length,maxusbpktlen,usb_pipedevice(urb->pipe),usb_pipeendpoint(urb->pipe),pipe_type);
		}
	}
}
#endif

static void port_reset(struct vmb *vmb,int port_number,int reset_on);
static void port_enable(struct vmb *vmb,int port_number,int enable_on);

#define WRITE_TLER(vmb,td,tler) {DBG(5, "TD=%X TLER=%04X write_tler(%s # %d in function %s)\n",td,tler,__FILE__,__LINE__,__FUNCTION__); write_tler(vmb,td,tler);}

#include "vmb_sched.c"

#if DEBUGGING_COMPILED_IN
void vmb_hcd_switch_on_diagnostics(void)
{
	pc_debug = 0x7fffffff;
	pc_trace = 0xFFFFFFFF;

	struct vmb *vmb;

	list_for_each_entry(vmb,&vmb_static_list,vmb_list) {

		PRINTK(KERN_DEBUG "============= diagnostics for vmb=0x%p =============\n",vmb);

		UTL_DEBUG(0,vmb);	// calls VMB_DEBUG
		PTR_DEBUG(0,vmb);
	}
}
EXPORT_SYMBOL_GPL(vmb_hcd_switch_on_diagnostics);
#endif
/*-------------------------------------------------------------------------*/
static const char vmb_name[] = "vmb_hcd";
/*-------------------------------------------------------------------------*/
#define PORT_C_MASK \
	((USB_PORT_STAT_C_CONNECTION \
	| USB_PORT_STAT_C_ENABLE \
	| USB_PORT_STAT_C_SUSPEND \
	| USB_PORT_STAT_C_OVERCURRENT \
	| USB_PORT_STAT_C_RESET) << 16)
/*-------------------------------------------------------------------------*/

static void vmb_power(struct vmb *vmb, int is_on)
{
	ASSERT(! in_atomic());
	SW_ASSERT_LOCKED(vmb);

	DBG(3, "%s(%d)\n",__FUNCTION__,is_on);

	struct usb_hcd  *hcd = vmb_to_hcd(vmb);

	/* hub is inactive unless the port is powered */
	if (is_on) {
		u8 cr = read_cr(vmb);
		cr |= 0x01;
		write_cr(vmb,cr);
		if (vmb->power)
			return;

		vmb->power = 1;
		hcd->self.controller->power.power_state = PMSG_ON;
	} else {
		u8 cr = read_cr(vmb);
		cr &= ~0x01;
		write_cr(vmb,cr);
		vmb->power = 0;
		hcd->state = HC_STATE_HALT;
		hcd->self.controller->power.power_state = PMSG_SUSPEND;
	}
	DBG(3, "%s(%d) completed power_state=%x\n",__FUNCTION__,is_on,hcd->self.controller->power.power_state.event);
}
static void port_power(struct vmb *vmb,int pn, int is_on)
{
	ASSERT(! in_atomic());
	SW_ASSERT_LOCKED(vmb);
	ASSERT(pn);

	DBG(3, "%s(%d->%d)\n",__FUNCTION__,pn,is_on);

	//struct usb_hcd  *hcd = vmb_to_hcd(vmb);

	vmb->port[pn-1].power = is_on;

	DBG(3, "%s() completed\n",__FUNCTION__);
}
static void vmb_reset(struct vmb *vmb, int reset_on)
{
	ASSERT(! in_atomic());
	SW_ASSERT_LOCKED(vmb);

	DBG(1, "%s(%d)\n",__FUNCTION__,reset_on);

	UTL_DEBUG(1,vmb);

	if (reset_on) {
	    vmb->reset = 1;
		u8 cr = 0x04;
		write_cr(vmb,cr);
		u8 pc = 0x00;
		write_pcsr(vmb,pc);

	} else {
		u8 cr = read_cr(vmb);
		cr &= ~0x04;
		cr |= 0x01;
		write_cr(vmb,cr);
		vmb->reset = 0;
	}
	UTL_DEBUG(1,vmb);

	DBG(1, "%s(%d) completed\n",__FUNCTION__,reset_on);
}
static void port_reset(struct vmb *vmb,int port_number,int reset_on)
{
	ASSERT(! in_atomic());
	SW_ASSERT_LOCKED(vmb);

	DBG(1, "%s(port %d -> %d)\n",__FUNCTION__,port_number,reset_on);

	UTL_DEBUG(1,vmb);

	u8 pc = read_pcsr(vmb);

	u8 usb_addr = MAX_DEV_ADDRS;

	while(usb_addr-- > 0)
	{
		if (port_number == vmb->udev[usb_addr].portnumber)
		{
			if (vmb->udev[usb_addr].usb_device) {

				usb_put_dev(vmb->udev[usb_addr].usb_device);
				vmb->udev[usb_addr].usb_device = NULL;
			}

			vmb->udev[usb_addr].portnumber = 0;
			vmb->udev[usb_addr].devicefail = 0;
		}
	}

	if (reset_on)
	{
		int delay = msecs_to_jiffies(50);

		if (1 == port_number)
		{
			struct vmb_port *port = &vmb->port[0];

			pc &= 0x0A;
			pc |= 0x05;
			port->reset = 1;
			port->enable = 0;
			port->finish = jiffies + delay;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
			queue_delayed_work(workqueue,&port->status_work,delay);
#else
			queue_delayed_work(workqueue,&port->delayed_work,delay);
#endif
		}
		else if (2 == port_number)
		{
			struct vmb_port *port = &vmb->port[1];

			pc &= 0x05;
			pc |= 0x0A;
			port->reset = 1;
			port->enable = 0;
			port->finish = jiffies + delay;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
			queue_delayed_work(workqueue,&port->status_work,delay);
#else
			queue_delayed_work(workqueue,&port->delayed_work,delay);
#endif
		}
		else
		{
			ERR("%s(illegal port number %d)\n",__FUNCTION__,port_number);
		}
	}
	else // reset off
	{
		if (1 == port_number)
		{
			pc &= 0x0A;
			pc |= 0x00;
			vmb->port[0].reset = 0;
			vmb->port[0].enable = 0;
		}
		else if (2 == port_number)
		{
			pc &= 0x05;
			pc |= 0x00;
			vmb->port[1].reset = 0;
			vmb->port[1].enable = 0;
		}
		else
		{
			ERR("%s(illegal port number %d)\n",__FUNCTION__,port_number);
		}
	}
	write_pcsr(vmb,pc);
	u8 cr = read_cr(vmb);
	cr |= 0x01;
	write_cr(vmb,cr);

	UTL_DEBUG(1,vmb);

	DBG(4, "%s() completed\n",__FUNCTION__);
}
static void vmb_enable(struct vmb *vmb)
{
	ASSERT(! in_atomic());
	SW_ASSERT_LOCKED(vmb);

	DBG(3, "%s()\n",__FUNCTION__);
	{
		VMB_DEBUG(3,vmb);

		u8 cr = read_cr(vmb);
		cr |= 0x01;
		write_cr(vmb,cr);
	}
	DBG(3, "%s() completed\n",__FUNCTION__);
}
static void port_enable(struct vmb *vmb,int port_number,int enable_on)
{
	SW_ASSERT_LOCKED(vmb);

	DBG(1, "%s(port %d -> %d)\n",__FUNCTION__,port_number,enable_on);

	VMB_DEBUG(4,vmb);

	u8 pc = read_pcsr(vmb);

	if (enable_on)
	{
		if (1 == port_number)
		{
			pc &= 0x0A;
			pc |= 0x01;
		}
		else if (2 == port_number)
		{
			pc &= 0x05;
			pc |= 0x02;
		}
		else
		{
			ERR("%s(illegal port number %d)\n",__FUNCTION__,port_number);
			return;
		}
		vmb->port[port_number-1].reset = 0;
		vmb->port[port_number-1].enable = 1;
	}
	else // enable off
	{
		if (1 == port_number)
		{
			pc &= 0x0A;
			pc |= 0x00;
		}
		else if (2 == port_number)
		{
			pc &= 0x05;
			pc |= 0x00;
		}
		else
		{
			ERR("%s(illegal port number %d)\n",__FUNCTION__,port_number);
			return;
		}
		vmb->port[port_number-1].reset = 0;
		vmb->port[port_number-1].enable = 0;

		if (VPS_error == vmb->port[port_number-1].status) {
			vmb->port[port_number-1].status = VPS_monitoring;
			vmb->port[port_number-1].udev = NULL;	// CLEAR USAGE COUNT ?????
		}	
	}
	write_pcsr(vmb,pc);
	u8 cr = read_cr(vmb);
	cr |= 0x01;
	write_cr(vmb,cr);

	VMB_DEBUG(4,vmb);

	DBG(4, "%s(pc=%02X) completed\n",__FUNCTION__,pc);
}
/* ----------------------------------------------------------------------------------------------- */
/*		 interface API for USB hub's						     */
/* ----------------------------------------------------------------------------------------------- */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
static irqreturn_t vmb_hcd_irq(struct usb_hcd *hcd, struct pt_regs *regs) // hc_driver API
{
#else
static irqreturn_t vmb_hcd_irq(struct usb_hcd *hcd) // hc_driver API
{
#endif
	ASSERT(in_interrupt());	// cannot use SW_LOCK
	ASSERT(hcd);
	struct vmb	*vmb = hcd_to_vmb(hcd);
	ASSERT(vmb);
	u8 sr = read_sr(vmb);

	if (0xff == sr) {

		if (vmb->unplugged) {
			DBG(3,"%s() UNPLUGGED\n",__FUNCTION__);
			return IRQ_NONE;
		} else {
			ERR("%s() UNPLUGGED\n",__FUNCTION__);
			vmb->unplugged = 1;
			return IRQ_NONE;
		}
	}

	HW_LOCK(vmb);

	u8 sr_bits = sr & ~vmb->sr_mask;

	if (sr_bits) {

		DBG(7, "%s(REALLY FOR US)\n",__FUNCTION__);

		VMB_DEBUG_LOCKED(8,vmb);

		if (0x04 & sr_bits) {
			DBG(4,"%s(TX THRESH)\n",__FUNCTION__);
			vmb->sr_mask |= 0x04;
			vmb->txthresh.kicked += 1;
			queue_work(workqueue,&vmb->txthresh.work);
		}

		if (0x08 & sr_bits) {
			DBG(4,"%s(RX THRESH)\n",__FUNCTION__);
			vmb->sr_mask |= 0x08;
			vmb->rxthresh.kicked += 1;
			queue_work(workqueue,&vmb->rxthresh.work);
		}

		if (0x10 & sr_bits) {
			DBG(4,"%s(RX TO)\n",__FUNCTION__);
			vmb->sr_mask |= 0x10;
			vmb->rxto.kicked += 1;
			queue_work(workqueue,&vmb->rxto.work);
		}

		u8 sr_mask = vmb->sr_mask;

		if (0x20 & sr_bits) {
			DBG(4,"%s(OVR)\n",__FUNCTION__);
			sr_mask |= 0x20;
			vmb->ovr.kicked += 1;
			queue_work(workqueue,&vmb->ovr.work);
		}

		if (0x40 & sr_bits) {
			DBG(4,"%s(EOT)\n",__FUNCTION__);
			sr_mask |= 0x40;
			vmb->eot.kicked += 1;
			queue_work(workqueue,&vmb->eot.work);
		}

		if (0x80 & sr_bits) {
			DBG(4,"%s(RX ERR)\n",__FUNCTION__);
			sr_mask |= 0x80;
			vmb->rxerr.kicked += 1;
			queue_work(workqueue,&vmb->rxerr.work);
		}

		write_sr(vmb,sr_mask);

		DBG(7, "%s() HANDLED\n",__FUNCTION__);

		HW_UNLOCK(vmb);

		return IRQ_HANDLED;

	} else {

		DBG(8, "%s() NOT FOR US\n",__FUNCTION__);

		HW_UNLOCK(vmb);

		return IRQ_NONE;
	}
}
static int vmb_hcd_start(struct usb_hcd *hcd) // hc_driver API
{
	ASSERT(! in_atomic());

	struct vmb	*vmb = hcd_to_vmb(hcd);
	SW_LOCK(vmb);

	DBG(1, "API:%s(%p)\n",__FUNCTION__,hcd);

	ASSERT(vmb->addr > 0);

	int retval = 0;

	{
		u8 rxrplr = read_rxrplr(vmb);
		u8 rxrphr = read_rxrplr(vmb);
		u8 rxwplr = read_rxwplr(vmb);
		u8 rxwphr = read_rxwplr(vmb);
		
		u16 txrp = read_txrp(vmb);
		u16 txwp = read_txwp(vmb);

		u8 rxitlr = read_rxitlr(vmb);
		u8 rxithr = read_rxithr(vmb);
		
		u8 txitlr = read_txitlr(vmb);
		u8 txithr = read_txithr(vmb);

		u8 rxdtor = read_rxdtor(vmb);

		DBG(1,"rxrp h,l = %02X,%02X\n",rxrphr,rxrplr);
		DBG(1,"rxwp h,l = %02X,%02X\n",rxwphr,rxwplr);
		DBG(1,"txrp     = %04X\n",txrp);
		DBG(1,"txwp     = %04X\n",txwp);
		DBG(1,"rxit h,l = %02X,%02X\n",rxithr,rxitlr);
		DBG(1,"txit h,l = %02X,%02X\n",txithr,txitlr);
		DBG(1,"rxdtor   = %02X\n",rxdtor);

		UTL_DEBUG(1,vmb);
	}

	hcd->state = HC_STATE_RUNNING;

	hcd->power_budget = vmb->board->power;

	vmb_power(vmb, 1); /* enable power and interupts */

	vmb_enable(vmb);

	VMB_DEBUG(3,vmb);

	DBG(0, "API:%s(%p power_budget=%d) completed %d\n",__FUNCTION__,hcd,hcd->power_budget,retval);

	SW_UNLOCK(vmb);

	return retval;
}
static void vmb_hcd_stop(struct usb_hcd *hcd) // hc_driver API
{
	ASSERT(! in_atomic());

	struct vmb	*vmb = hcd_to_vmb(hcd);

	SW_LOCK(vmb);

	DBG(1, "API:%s(%p)\n",__FUNCTION__,hcd);

	del_timer_sync(&hcd->rh_timer);

	vmb_power(vmb, 0);

	DBG(1, "API:%s(%p) completed\n",__FUNCTION__,hcd);

	SW_UNLOCK(vmb);
}
static int vmb_hcd_init(struct usb_hcd *hcd) // hc_driver API
{
	ASSERT(! in_atomic());

	struct vmb	*vmb = hcd_to_vmb(hcd);
	SW_LOCK(vmb);

	DBG(3, "API:%s(%p)\n",__FUNCTION__,hcd);

	int retval = 0;

	vmb_reset(vmb, 1);
	msleep(50);
	vmb_reset(vmb, 0);

	DBG(3, "API:%s(%p) completed %d\n",__FUNCTION__,hcd,retval);

	SW_UNLOCK(vmb);

	return retval;
}
/*
 *
 * The USB core subsystem can call this API function from
 *     pretty much any kernel state (referred to as context),
 *     including INTERRUPT context. Which means that all code
 *     execution paths from here cannot use semaphores and
 *     that means the functions such as down() and up(), but
 *     can use spin locks
 *
 * The (usb_host_endpoint *)hep->hcpriv pointer is NOT currently used
 *     if would be available explicitly at the following APIs:
 *     .endpoint_disable
 *     .urb_enqueue
 *
 *
 */
static int vmb_urb_enqueue(struct usb_hcd*hcd,struct usb_host_endpoint*hep,struct urb*urb,unsigned mem_flags) // hc_driver API
{
	DBG(7, "I=%d A=%d C=%p %s(urb=%p)\n",(int)in_interrupt(),in_atomic(),current,__FUNCTION__,urb);
	// possibly in_atomic()
	// in_interrupt()

	ASSERT(0 == urb->error_count);

	ASSERT(-EINPROGRESS == urb->status);

	struct vmb		*vmb = hcd_to_vmb(hcd);

	DBG(6, "API:%s()\n",__FUNCTION__);
	URB_DBG(6, "%s(urb=%p status=%d) en_queue_size=%d en_queue_next=%d\n",__FUNCTION__,urb,urb->status,vmb->en_queue_size,vmb->en_queue_next);

	ASSERT((0 == vmb->current_port) || (1 == vmb->enumerating));

	UTL_DEBUG(7,vmb);

	if (vmb->en_queue_size < URB_EN_QUEUE_SIZE) {

		HW_LOCK(vmb);

		vmb->en_queue[URB_EN_QUEUE_MASK & vmb->en_queue_next++] = urb;

		vmb->en_queue_size += 1;

		HW_UNLOCK(vmb);

		queue_work(workqueue,&vmb->urb_en_queue);

		DBG(6, "API:%s() completed leaving %d queued requests\n",__FUNCTION__,vmb->en_queue_size);

		return 0;

	} else {

		ERR("API:%s() queue of requests is FULL\n",__FUNCTION__);

		return -ENOMEM;
	}
}
static int vmb_urb_dequeue(struct usb_hcd *hcd, struct urb *urb) // hc_driver API
{
	DBG(6, "API:%s(urb=%p)\n",__FUNCTION__,urb);
	DBG(7, "I=%d A=%d C=%p %s()\n",(int)in_interrupt(),in_atomic(),current,__FUNCTION__);
	// possibly in_atomic()
	// in_interrupt()

	ASSERT(-ECONNRESET == urb->status || -ESHUTDOWN == urb->status || -ENOENT == urb->status);

	struct vmb *vmb = hcd_to_vmb(hcd);

	URB_DBG(0, "%s(urb=%p status=%d) de_queue_size=%d de_queue_next=%d\n",__FUNCTION__,urb,urb->status,vmb->de_queue_size,vmb->de_queue_next);

	if (vmb->de_queue_size < URB_DE_QUEUE_SIZE) {

		HW_LOCK(vmb);

		vmb->de_queue[URB_DE_QUEUE_MASK & vmb->de_queue_next++] = urb;

		vmb->de_queue_size += 1;

		HW_UNLOCK(vmb);

		queue_work(workqueue,&vmb->urb_de_queue);

		DBG(6, "API:%s(urb=%p) completed leaving %d queued requests\n",__FUNCTION__,urb,vmb->en_queue_size);

		return 0;

	} else {

		ERR("API:%s() queue of requests is FULL\n",__FUNCTION__);

		return -ENOMEM;
	}
}
static void vmb_endpoint_disable(struct usb_hcd *hcd, struct usb_host_endpoint *hep) // hc_driver API
{
	ASSERT(! in_atomic());

	DBG(3, "API:%s(hep=%p)\n",__FUNCTION__,hep);

	struct vmb		*vmb = hcd_to_vmb(hcd);

	UTL_DEBUG(0,vmb);   // calls VMB_DEBUG
	PTR_DEBUG(0,vmb);

	DBG(3, "API:%s(hep=%p) completed\n",__FUNCTION__,hep);
}

static int vmb_get_frame(struct usb_hcd *hcd) // hc_driver API
{
	ASSERT(! in_atomic());

	struct vmb *vmb = hcd_to_vmb(hcd);

	SW_LOCK(vmb);

	DBG(1, "API:%s()\n",__FUNCTION__);

	/* wrong except while periodic transfers are scheduled;
	 * never matches the on-the-wire frame;
	 * subject to overruns.
	 */
	int frame = vmb->__frame__;

	DBG(3, "API:%s() completed %d\n",__FUNCTION__,vmb->__frame__);

	SW_UNLOCK(vmb);

	return frame;
}
static void vmb_hub_descriptor (struct vmb*vmb,struct usb_hub_descriptor *desc)
{
	SW_ASSERT_LOCKED(vmb);

	u16	temp = 0;

	desc->bDescriptorType = 0x29;
	desc->bHubContrCurrent = 0;
	desc->bNbrPorts = 2;
	desc->bDescLength = 9;	// less than 8 ports

	desc->bPwrOn2PwrGood = 0; /* per-port power switching (gang of one!), or none */

	if (vmb->board->port_power) {
		desc->bPwrOn2PwrGood = vmb->board->potpg;
		if (!desc->bPwrOn2PwrGood)
			desc->bPwrOn2PwrGood = 10;
		temp = 0x0001;
	} else {
		temp = 0x0002;
	}

	temp |= 0x0010; /* no overcurrent errors detection/handling */

	desc->wHubCharacteristics = cpu_to_le16(temp);

	/* two bitmaps:  ports removable, and legacy PortPwrCtrlMask */
	desc->bitmap[0] = 1 << 1;
	desc->bitmap[1] = 0xff;
}
/* the virtual root hub timer IRQ checks for hub status */
static int vmb_hub_status_data(struct usb_hcd *hcd, char *buf) // hc_driver API
{
	ASSERT(in_interrupt());	// cannot use SW_LOCK

	struct vmb *vmb = hcd_to_vmb(hcd);

	DBG(7, "API:%s()\n",__FUNCTION__);

	*buf = 0;

	VMB_DEBUG(3,vmb);

	u8 pc = read_pcsr(vmb);

	if ((pc & 0x31) == 0x10)
	{
		vmb->port[0].device = 1;
	}
	if ((pc & 0x31) == 0x00)
	{
		vmb->port[0].device = 0;
	}
	if ((pc & 0xC2) == 0x40)
	{
		vmb->port[1].device = 1;
	}
	if ((pc & 0xC2) == 0x00)
	{
		vmb->port[1].device = 0;
	}

	if	((0 == vmb->port[0].device) && (VPS_enumerating == vmb->port[0].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(1 == vmb->current_port);

		*buf |= (1 << 1);

		vmb->port[0].Status |= (1 << USB_PORT_FEAT_C_CONNECTION);
		vmb->port[0].Status &= ~(1 << USB_PORT_FEAT_CONNECTION);

	} else if ((1 == vmb->port[0].device) && (VPS_enumerating == vmb->port[0].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(1 == vmb->current_port);

	} else if ((0 == vmb->port[1].device) && (VPS_enumerating == vmb->port[1].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(2 == vmb->current_port);

		*buf |= (1 << 2);

		vmb->port[1].Status |= (1 << USB_PORT_FEAT_C_CONNECTION);
		vmb->port[1].Status &= ~(1 << USB_PORT_FEAT_CONNECTION);

	} else if ((1 == vmb->port[1].device) && (VPS_enumerating == vmb->port[1].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(2 == vmb->current_port);

	} else if ((0 == vmb->port[0].device) && (VPS_detected == vmb->port[0].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(1 == vmb->current_port);

		// leave things as they are - let the core finish RESET sequence

	} else if ((1 == vmb->port[0].device) && (VPS_detected == vmb->port[0].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(1 == vmb->current_port);

	} else if ((0 == vmb->port[1].device) && (VPS_detected == vmb->port[1].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(2 == vmb->current_port);

		// leave things as they are - let the core finish RESET sequence

	} else if ((1 == vmb->port[1].device) && (VPS_detected == vmb->port[1].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(2 == vmb->current_port);

	} else if ((1 == vmb->port[0].device) && (VPS_monitoring == vmb->port[0].status)) {

		UTL_ASSERT(!vmb->enumerating);
		UTL_ASSERT(0 == vmb->current_port);

		vmb->port[0].status = VPS_detected;

		vmb->current_port = 1;
		vmb->enumerating = 1;

		*buf |= (1 << 1);

		vmb->port[0].Status |= (1 << USB_PORT_FEAT_C_CONNECTION);
		vmb->port[0].Status |= (1 << USB_PORT_FEAT_CONNECTION);

	} else if ((1 == vmb->port[1].device) && (VPS_monitoring == vmb->port[1].status)) {

		UTL_ASSERT(!vmb->enumerating);
		UTL_ASSERT(0 == vmb->current_port);

		vmb->port[1].status = VPS_detected;

		vmb->current_port = 2;
		vmb->enumerating = 1;

		*buf |= (1 << 2);

		vmb->port[1].Status |= (1 << USB_PORT_FEAT_CONNECTION);
		vmb->port[1].Status |= (1 << USB_PORT_FEAT_C_CONNECTION);

	} else if ((0 == vmb->port[0].device) && (VPS_resetting == vmb->port[0].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(1 == vmb->current_port);

	} else if ((1 == vmb->port[0].device) && (VPS_resetting == vmb->port[0].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(1 == vmb->current_port);

	} else if ((0 == vmb->port[1].device) && (VPS_resetting == vmb->port[1].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(2 == vmb->current_port);

	} else if ((1 == vmb->port[1].device) && (VPS_resetting == vmb->port[1].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(2 == vmb->current_port);

	} else if ((0 == vmb->port[0].device) && (VPS_enabled == vmb->port[0].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(1 == vmb->current_port);

	} else if ((1 == vmb->port[0].device) && (VPS_enabled == vmb->port[0].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(1 == vmb->current_port);

	} else if ((0 == vmb->port[1].device) && (VPS_enabled == vmb->port[1].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(2 == vmb->current_port);

	} else if ((1 == vmb->port[1].device) && (VPS_enabled == vmb->port[1].status)) {

		UTL_ASSERT(vmb->enumerating);
		UTL_ASSERT(2 == vmb->current_port);

	} else if ((1 == vmb->port[0].device) && (VPS_error == vmb->port[0].status)) {

		*buf |= (1 << 1);

		vmb->port[0].Status |= (1 << USB_PORT_FEAT_C_CONNECTION);
		vmb->port[0].Status &= ~(1 << USB_PORT_FEAT_CONNECTION);

	} else if ((1 == vmb->port[1].device) && (VPS_error == vmb->port[1].status)) {

		*buf |= (1 << 2);

		vmb->port[1].Status |= (1 << USB_PORT_FEAT_C_CONNECTION);
		vmb->port[1].Status &= ~(1 << USB_PORT_FEAT_CONNECTION);

	} else {

		UTL_ASSERT(0 == vmb->current_port);

	}

	int retval;

	if (*buf) {
		retval = 1;

		VMB_DEBUG(3,vmb);
		DBG(7, "%s(%02X) completed %d\n",__FUNCTION__,buf[0],retval);
		PRINTK(KERN_DEBUG "%d-V:status_data %02X\n",hcd->self.busnum,buf[0]);
	} else {
		retval = 0;
		DBG(7, "API:%s() completed %d\n",__FUNCTION__,retval);
	}
	KICK_SCHEDULER(vmb);

	return retval;
}
static int vmb_hub_control(struct usb_hcd *hcd, u16 typeReq, u16 wValue, u16 wIndex, char *buf, u16 wLength) // hc_driver API
{
	ASSERT(! in_atomic());

	struct vmb	*vmb = hcd_to_vmb(hcd);

	DBG(1, "API:%s()\n",__FUNCTION__);
	SW_LOCK(vmb);



	int		retval = 0;

	switch (typeReq) {
	case ClearHubFeature:
		DBG(3, "%s(ClearHubFeature,wValue=%d,wIndex=%d)\n",__FUNCTION__,wValue,wIndex);
		switch (wValue) {
		case C_HUB_OVER_CURRENT:
		case C_HUB_LOCAL_POWER:
			break;
		default:
			goto error;
		}
		break;
	case SetHubFeature:
		DBG(3, "%s(SetHubFeature,wValue=%d,wIndex=%d)\n",__FUNCTION__,wValue,wIndex);
		switch (wValue) {
		case C_HUB_OVER_CURRENT:
		case C_HUB_LOCAL_POWER:
			break;
		default:
			goto error;
		}
		break;
	case ClearPortFeature:
	{
		DBG(3, "%s(ClearPortFeature,wValue=%d,wIndex=%d)\n",__FUNCTION__,wValue,wIndex);
		if (wIndex < 1 || wIndex > 2 || wLength != 0) goto error;
		PRINTK(KERN_DEBUG "%d-V:ClearPortFeature~%d = %d ~ %s\n",hcd->self.busnum,wIndex,wValue,feature_string(wValue));
		struct vmb_port *port = &vmb->port[wIndex-1];

		switch (wValue) {
		case USB_PORT_FEAT_ENABLE:
			port_enable(vmb,wIndex,0);
			break;
		case USB_PORT_FEAT_SUSPEND:
		{
			int delay = msecs_to_jiffies(20);
			port->finish = jiffies + delay;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
			queue_delayed_work(workqueue,&port->status_work,delay);
#else
			queue_delayed_work(workqueue,&port->delayed_work,delay);
#endif
			break;
		}
		case USB_PORT_FEAT_POWER:
			port_power(vmb,wIndex,0);
			break;
		case USB_PORT_FEAT_C_RESET:

			if ((1 == vmb->enumerating) && (wIndex == vmb->current_port) && (VPS_enabled == port->status))
			{
				port_enable(vmb,wIndex,1);
				port->status = VPS_enumerating;
			}
	
			break;
		case USB_PORT_FEAT_C_ENABLE:
		case USB_PORT_FEAT_C_SUSPEND:
		case USB_PORT_FEAT_C_CONNECTION:
		case USB_PORT_FEAT_C_OVER_CURRENT:
			break;
		default:
			goto error;
		}
		port->Status &= ~(1 << wValue);
		break;
	}
	case GetHubDescriptor:
		DBG(3, "%s(GetHubDescriptor,wValue=%d,wIndex=%d)\n",__FUNCTION__,wValue,wIndex);
		vmb_hub_descriptor(vmb, (struct usb_hub_descriptor *) buf);
		break;
	case GetHubStatus:
		DBG(3, "%s(GetHubStatus,wValue=%d,wIndex=%d)\n",__FUNCTION__,wValue,wIndex);
		*(__le32 *) buf = cpu_to_le32(0);
		break;
	case GetPortStatus:
		DBG(3, "%s(GetPortStatus,wValue=%d,wIndex=%d)\n",__FUNCTION__,wValue,wIndex);
		if (wIndex < 1 || wIndex > 2) goto error;

		*(__le32 *) buf = cpu_to_le32(vmb->port[wIndex-1].Status);;

		PRINTK(KERN_DEBUG "%d-V:GetStatus~%d = %08X\n",hcd->self.busnum,wIndex,*(__le32 *)buf);
		break;
	case SetPortFeature:
	{
		DBG(3, "%s(SetPortFeature,wValue=%d,wIndex=%d)\n",__FUNCTION__,wValue,wIndex);
		if (wIndex < 1 || wIndex > 2 || wLength != 0) goto error;
		PRINTK(KERN_DEBUG "%d-V:SetPortFeature~%d = %d ~ %s\n",hcd->self.busnum,wIndex,wValue,feature_string(wValue));
		struct vmb_port *port = &vmb->port[wIndex-1];

		switch (wValue) {
		case USB_PORT_FEAT_ENABLE:
			port_enable(vmb,wIndex,1);
			break;
		case USB_PORT_FEAT_SUSPEND:
			break;
		case USB_PORT_FEAT_POWER:
			port_power(vmb,wIndex,1);
			break;
		case USB_PORT_FEAT_RESET:
			VMB_DEBUG(1,vmb);
			port_reset(vmb,wIndex,1);
			if ((1 == vmb->enumerating) && (wIndex == vmb->current_port) && (VPS_detected == vmb->port[wIndex-1].status))
			{
				vmb->port[wIndex-1].status = VPS_resetting;
			}
			if ((1 == vmb->enumerating) && (wIndex == vmb->current_port) && (VPS_enumerating == vmb->port[wIndex-1].status))
			{
				vmb->port[wIndex-1].status = VPS_resetting;
			}
			VMB_DEBUG(1,vmb);
			break;
		default:
			goto error;
		}
		port->Status |= 1 << wValue;

		break;
	}
	default:
		DBG(3, "%s(typeReq=%d,wValue=%d,wIndex=%d)\n",__FUNCTION__,typeReq,wValue,wIndex);
error:
		/* "protocol stall" on error */
		DBG(3, "%s() completed %d\n",__FUNCTION__,retval);
		retval = -EPIPE;
	}

	VMB_DEBUG(3,vmb);
	DBG(3, "%s() completed %d\n",__FUNCTION__,retval);

	SW_UNLOCK(vmb);

	return retval;
}
static int vmb_start_port_reset(struct usb_hcd *hcd, unsigned port_num)	// hc_driver API
{
	struct vmb	*vmb = hcd_to_vmb(hcd);
	SW_LOCK(vmb);

	DBG(1, "API:%s()\n",__FUNCTION__);

	SW_UNLOCK(vmb);

	return 0;
}
static void vmb_hub_irq_enable(struct usb_hcd *hcd)	// hc_driver API
{
	struct vmb	*vmb = hcd_to_vmb(hcd);
	SW_LOCK(vmb);
	DBG(1, "API:%s()\n",__FUNCTION__);
	SW_UNLOCK(vmb);
}
#ifdef  CONFIG_PM
static int vmb_hcd_suspend(struct usb_hcd *hcd,pm_message_t message) // hc_driver API
{
	struct vmb	*vmb = hcd_to_vmb(hcd);
	SW_LOCK(vmb);
	DBG(1, "API:%s()\n",__FUNCTION__);
	SW_UNLOCK(vmb);
	return 0;
}
static int vmb_hcd_resume(struct usb_hcd *hcd) // hc_driver API
{
	struct vmb	*vmb = hcd_to_vmb(hcd);
	SW_LOCK(vmb);
	DBG(1, "API:%s()\n",__FUNCTION__);
	SW_UNLOCK(vmb);
	return 0;
}
static int vmb_bus_suspend(struct usb_hcd *hcd) // hc_driver API
{
	struct vmb	*vmb = hcd_to_vmb(hcd);
	SW_LOCK(vmb);
	// sofs off
	int retval = 0;
	DBG(3, "API:%s()\n",__FUNCTION__);
	DBG(3, "API:%s() completed %d\n",__FUNCTION__,retval);
	SW_UNLOCK(vmb);
	return retval;
}
static int vmb_bus_resume(struct usb_hcd *hcd) // hc_driver API
{
	struct vmb	*vmb = hcd_to_vmb(hcd);
	SW_LOCK(vmb);
	// sofs on
	int retval = 0;
	DBG(3, "API:%s()\n",__FUNCTION__);
	DBG(3, "API:%s() completed %d\n",__FUNCTION__,retval);
	SW_UNLOCK(vmb);
	return retval;
}
#else
#define vmb_hcd_suspend      NULL
#define vmb_hcd_resume       NULL
#define vmb_bus_suspend      NULL
#define vmb_bus_resume       NULL
#endif
/* ----------------------------------------------------------------------------------------------- */
static struct hc_driver vmb_hc_driver = {
	.description =		vmb_name,
	.hcd_priv_size =	sizeof(struct vmb),

	/*
	 * generic hardware linkage
	 */
	.irq =			vmb_hcd_irq,
	.flags =		HCD_USB11 | HCD_MEMORY,

	/* basic lifecycle operations */
	.reset =		vmb_hcd_init,	// "reset" is misnamed; its role is now one-time init.
	.start =		vmb_hcd_start,
	.suspend =		vmb_hcd_suspend,
	.resume =		vmb_hcd_resume,
	.stop =			vmb_hcd_stop,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue =		vmb_urb_enqueue,
	.urb_dequeue =		vmb_urb_dequeue,
	.endpoint_disable =	vmb_endpoint_disable,

	/*
	 * periodic schedule support
	 */
	.get_frame_number =	vmb_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data =	vmb_hub_status_data,
	.hub_control =		vmb_hub_control,
	.bus_suspend =		vmb_bus_suspend,
	.bus_resume =		vmb_bus_resume,
	.start_port_reset = 	vmb_start_port_reset,
	.hub_irq_enable =	vmb_hub_irq_enable,
};

/*-------------------------------------------------------------------------*/
static int __devexit vmb_remove(struct platform_device *pdev) // device_driver API
{
	ASSERT(! in_atomic());
	ASSERT(!in_interrupt());	// required for usb_remove_hcd

	DBG(1, "API:%s(%p)\n",__FUNCTION__,pdev);

	int retval = 0;

	struct usb_hcd		*hcd = platform_get_drvdata(pdev);
	struct vmb		*vmb = hcd_to_vmb(hcd);

	SW_LOCK(vmb);

	cancel_delayed_work(&vmb->diagnostics);

	DBG(1,"%s(cancel_delayed_work for diagnostics)\n",__FUNCTION__);

	int ports = MAX_VMB_PORTS;
	while(ports-- > 0)
	{
		struct vmb_port *port = &vmb->port[ports];

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
		cancel_delayed_work(&port->status_work);
#else
		cancel_delayed_work(&port->delayed_work);
#endif
	}

	int slots = MAX_VMB_SLOTS;
	while(slots-- > 0)
	{
		struct vmb_slot *slot = &vmb->slot[slots];
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
		cancel_delayed_work(&slot->status_work);
#else
		cancel_delayed_work(&slot->delayed_work);
#endif
		DBG(1,"%s(cancel_delayed_work for slot)\n",__FUNCTION__);
	}

	SW_UNLOCK(vmb);
	usb_remove_hcd(hcd);
	SW_LOCK(vmb);	// we don't unlock it now
	DBG(1,"%s(done usb_remove_hcd(0x%p))\n",__FUNCTION__,hcd);

	iounmap(vmb->addr);

	spin_lock(&vmb_static_lock);
	list_del_init(&vmb->vmb_list);
	vmb_instances -= 1;
	spin_unlock(&vmb_static_lock);

	usb_put_hcd(hcd); // should now free the hcd (along with the contained vmb)

	DBG(1, "API:%s(%p) completed returns %d\n",__FUNCTION__,pdev,retval);

	return retval;
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void initialise_event(struct vmb *vmb,struct vmb_event*event,void(*function)(void*))
{
#else
static void initialise_event(struct vmb *vmb,struct vmb_event*event,void(*function)(struct work_struct*))
{
#endif
	event->vmb = vmb;
	event->kicked = 0;
	event->worked = 0;
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
	INIT_WORK(&event->work,function,(void*)event);
#else
	INIT_WORK(&event->work,function);
#endif
}
static int initialise_vmb(struct vmb *vmb, struct platform_device *pdev, void __iomem *base, void __iomem *addr)
{
	ASSERT(! in_atomic());
	ASSERT(addr > 0);
	ASSERT(base > 0);

	DBG(3, "%s(vmb=%p,pdev=%p,base=%p,addr=%p) STARTING\n",__FUNCTION__,vmb,pdev,base,addr);
	// note that the interrupt handler has not yet been hooked up

	vmb->board = pdev->dev.platform_data;

	if (vmb->board->ports > MAX_VMB_PORTS) {
		int retval = -EINVAL;
		DBG(3, "%s() completed returns %d\n",__FUNCTION__,retval);
		return retval;
	} else {
		int retval = 0;

		vmb->addr = addr;
		vmb->base = base;
		vmb->sr_mask = 0x1c;

		u8 cr = read_cr(vmb);
		cr &= ~0x01;	// clear RUN
		write_cr(vmb,cr);
		if (vmb->board->switch_on_required) {
			u8 tl = read_tl(vmb);
			INFO("going to pulse the Digital IO line TL=%02X\n",tl);	
			tl |= 0x80;
			write_tl(vmb,tl);
			ssleep(3);
			tl &= ~0x80;
			write_tl(vmb,tl);
			msleep(100);
			INFO("finished pulsing the Digital IO line TL=%02X\n",read_tl(vmb));	
			// it should now be switched on
		} else {
			// it should work as soon as it was inserted
		}
		vmb->power = 0;
		vmb->reset = 0;
		vmb->unplugged = 0;
		vmb->enumerating = 0;
		vmb->current_port = 0;
		vmb->slot_count = MAX_VMB_SLOTS;

		spin_lock_init(&vmb->hw_lock);
		init_MUTEX(&vmb->sw_lock);

#if LOCK_DEBUGGING_COMPILED_IN
		vmb->hw_flag = 0;
		vmb->sw_flag = 0;

		vmb->hw_line = 0;
		vmb->sw_line = 0;
#endif

		vmb->__frame__ = 0;
		vmb->diag_txit = 0;
		vmb->diag_txwp = 0;
		vmb->diag_txrp = 0;
		vmb->diag_rxwp = 0;
		vmb->diag_rxrp = 0;

		SW_LOCK(vmb);

		INIT_LIST_HEAD(&vmb->async);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
		INIT_WORK(&vmb->scheduler,vmb_scheduler_event,(void*)vmb);
		INIT_WORK(&vmb->diagnostics,vmb_diagnostics_event,(void*)vmb);
		INIT_WORK(&vmb->urb_en_queue,vmb_en_queue_event,(void*)vmb);
		INIT_WORK(&vmb->urb_de_queue,vmb_de_queue_event,(void*)vmb);
#else

		INIT_WORK(&vmb->scheduler,vmb_scheduler_event);
		INIT_DELAYED_WORK(&vmb->diagnostics,vmb_diagnostics_event);
		INIT_WORK(&vmb->urb_en_queue,vmb_en_queue_event);
		INIT_WORK(&vmb->urb_de_queue,vmb_de_queue_event);
#endif
		initialise_event(vmb,&vmb->rxthresh,vmb_rxthresh_event);
		initialise_event(vmb,&vmb->rxto,vmb_rxto_event);
		initialise_event(vmb,&vmb->txthresh,vmb_txthresh_event);
		initialise_event(vmb,&vmb->ovr,vmb_ovr_event);
		initialise_event(vmb,&vmb->eot,vmb_eot_event);
		initialise_event(vmb,&vmb->rxerr,vmb_rxerr_event);

		int ports = MAX_VMB_PORTS;
		while(ports-- > 0)
		{
			struct vmb_port *port = &vmb->port[ports];

			port->Status = ( 1<< USB_PORT_FEAT_C_CONNECTION)
						 | ( 1<< USB_PORT_FEAT_C_ENABLE)
						 | ( 1<< USB_PORT_FEAT_POWER)
						 ;
			port->number = 1+ports;
			port->device = 0;
			port->enable = 0;
			port->reset  = 0;
			port->reset  = 0;
			port->status = VPS_monitoring;
			port->udev   = NULL;
			port->vmb    = vmb;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
			INIT_WORK(&port->status_work,vmb_single_port,(void*)port);
#else
			INIT_DELAYED_WORK(&port->delayed_work,vmb_single_port);
#endif
		}

		vmb->en_queue_size = 0;
		vmb->en_queue_next = 0;
		
		vmb->de_queue_size = 0;
		vmb->de_queue_next = 0;
		
		vmb->giveback_size = 0;
		vmb->giveback_next = 0;
		
		HW_LOCK(vmb);
		u16 start_fc = read_fc(vmb);    // the frame currently or recently finished executing
		HW_UNLOCK(vmb);

		vmb->frame_tail = start_fc;
		vmb->frame_head = start_fc;
		u16 fifo_offset = read_txwp(vmb);

		int frames = UTL_FRAME_SIZE;

		while(frames-- > 0) {

			vmb->frame[frames].scheduled_bits = 0;
			vmb->frame[frames].requested_bits = 0;
			vmb->frame[frames].fifo_offset = fifo_offset;
		}
		vmb->frame[start_fc].fifo_offset = fifo_offset;

		vmb->next_slot = 0;

		int slots = MAX_VMB_SLOTS;
		while(slots-- > 0)
		{
			struct vmb_slot *slot = &vmb->slot[slots];

			slot->allocated = 0;
			slot->initialized = 0;
			slot->queue_size = 0;
			slot->queue_next = 0;
			slot->next_urb_count = 0;
			slot->next_urb_index = 0;
			slot->idref_done = 0;
			slot->idref_this = 0;
			slot->idref_pkts = 0;
			slot->td = slots;
			slot->ec = 0;
			slot->end_seen = 0;
			slot->retry_flag = 0;
			slot->device_error = 0;
			slot->actioned = 0;
			slot->running = 0;
			slot->retired = 1;
			slot->max_plen = 0;
			slot->usb_addr = 0;
			slot->usb_endp = 0;
			slot->vmb = vmb;
			slot->jiffies = jiffies;
			slot->extra = 0;	// overspill buffer

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
			INIT_WORK(&slot->status_work,vmb_slot_event,(void*)slot);
#else
			INIT_DELAYED_WORK(&slot->delayed_work,vmb_slot_event);
#endif
		}

		int udevs = MAX_DEV_ADDRS;
		while(udevs-- > 0)
		{
			struct vmb_udev*udev = &vmb->udev[udevs];

			udev->portnumber = 0;
			udev->devicefail = 0;
			udev->usb_device = NULL;
		}

		write_pcsr(vmb,0x00);
	
		DBG(3, "%s()\n",__FUNCTION__);

		vmb_power(vmb, 0);

		VMB_DEBUG(3,vmb);

		msleep(200);

		queue_delayed_work(workqueue,&vmb->diagnostics,msecs_to_jiffies(5000));

		retval = 0;
		DBG(3, "%s() completed returns %d\n",__FUNCTION__,retval);

		SW_UNLOCK(vmb);

		return retval;
	}
}
static int __devinit vmb_probe(struct platform_device *pdev) // device_driver API
{
	ASSERT(! in_atomic());

	DBG(1, "API:%s(%p)\n",__FUNCTION__,pdev);

	struct usb_hcd		*hcd;
	struct vmb		*vmb;
	struct resource		*addr_res, *base_res;
	int			irq;
	void __iomem		*base;
	void __iomem		*addr;
	int			retval = 0;

	/* basic sanity checks first.  board-specific init logic should
	 * have initialized these three resources and probably board
	 * specific platform_data.  we don't probe for IRQs, and do only
	 * minimal sanity checking.
	 */
	irq = platform_get_irq(pdev, 0);
	if (pdev->num_resources < 3 || irq < 0) {
		DBG(1,"no resources\n");
		retval = -EINVAL;
		goto completed;
	}

	/* refuse to confuse usbcore */
	if (pdev->dev.dma_mask) {
		DBG(1,"no we won't dma\n");
		retval = -EINVAL;
		goto completed;
	}

	base_res = platform_get_resource(pdev, IORESOURCE_IO, 0);

	if (!base_res) {
		DBG(1,"failed to get IO\n");
		retval = -EINVAL;
		goto completed;
	}

	addr_res = platform_get_resource(pdev, IORESOURCE_MEM, 0);

	if (!addr_res) {
		DBG(1,"failed to get MEM\n");
		retval = -EINVAL;
		goto completed;
	}

	base = (void __iomem *) base_res->start;

	addr = ioremap(addr_res->start, 1 + addr_res->end - addr_res->start);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
	DBG(1,"platform resources IO=%04lX MEM=%08lX-->%p\n",base_res->start,addr_res->start,addr);
#else	
	DBG(1,"platform resources IO=%04X MEM=%08X-->%p\n",base_res->start,addr_res->start,addr);
#endif

	if (addr == NULL) {
		retval = -ENOMEM;
		goto err1;
	}

	/* allocate and initialize hcd */
	hcd = usb_create_hcd(&vmb_hc_driver, &pdev->dev, pdev->dev.bus_id);
	if (!hcd) {
		retval = -ENOMEM;
		goto err2;
	}
	hcd->rsrc_start = addr_res->start;
	vmb = hcd_to_vmb(hcd);

	retval = initialise_vmb(vmb,pdev,base,addr);
	if (retval != 0) goto err3;

	spin_lock(&vmb_static_lock);
	list_add_tail(&vmb->vmb_list,&vmb_static_list);
	vmb_instances += 1;
	spin_unlock(&vmb_static_lock);

	hcd->product_desc = "ELAN VMB Host Controller";

	DBG(1,"%s hcd=%p adding product_desc = %s\n",__FUNCTION__,hcd, hcd->product_desc);

/*
 *
 *  IRQF_SHARED             Interrupt is shared
 *  IRQF_DISABLED           Disable local interrupts while processing
 *  IRQF_SAMPLE_RANDOM      The interrupt can be used for entropy
 *
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
	retval = usb_add_hcd(hcd, irq, SA_INTERRUPT | SA_SHIRQ);
#else
	retval = usb_add_hcd(hcd, irq, IRQF_DISABLED | IRQF_SHARED);
#endif	
	if (retval != 0) goto err4;
	
	goto completed;

 	err4: {
		DBG(1,"%s removing %s\n",__FUNCTION__,hcd->product_desc);
		list_del_init(&vmb->vmb_list);
		vmb_instances -= 1;
		usb_put_hcd(hcd);
	}
	err3: 
 	err2: iounmap(addr);
 	err1: DBG(3,"init error, %d\n", retval);
	goto completed;

	completed:

	DBG(1, "API:%s(%p) completed %d\n",__FUNCTION__,pdev,retval);
	
	return retval;
}

#ifdef	CONFIG_PM

/* for this device there's no useful distinction between the controller
 * and its root hub, except that the root hub only gets direct PM calls
 * when CONFIG_USB_SUSPEND is enabled.
 */

static int vmb_suspend(struct platform_device *pdev, pm_message_t state) // device_driver API
{
	DBG(1, "I=%d A=%d C=%p %s()\n",(int)in_interrupt(),in_atomic(),current,__FUNCTION__);

	DBG(1, "API:%s(%p)\n",__FUNCTION__,pdev);
	struct usb_hcd	*hcd = platform_get_drvdata(pdev);
	struct vmb	*vmb = hcd_to_vmb(hcd);
	int		retval = 0;

	if (state.event <= PM_SUSPEND_MEM)
		retval = vmb_bus_suspend(hcd);
	else
		vmb_power(vmb, 0);
	if (retval == 0)
		pdev->dev.power.power_state = state;
	DBG(1, "API:%s(%p) completed %d\n",__FUNCTION__,pdev,retval);
	return retval;
}

static int vmb_resume(struct platform_device *pdev) // device_driver API
{
	DBG(1, "I=%d A=%d C=%p %s()\n",(int)in_interrupt(),in_atomic(),current,__FUNCTION__);

	DBG(1, "API:%s(%p)\n",__FUNCTION__,pdev);
	struct usb_hcd	*hcd = platform_get_drvdata(pdev);
	struct vmb	*vmb = hcd_to_vmb(hcd);

	/* with no "check to see if VBUS is still powered" board hook,
	 * let's assume it'd only be powered to enable remote wakeup.
	 */
	if (pdev->dev.power.power_state.event > PM_SUSPEND_MEM) {
		vmb_power(vmb, 1);
		return 0;
	}

	pdev->dev.power.power_state = PMSG_ON;
	int retval = vmb_bus_resume(hcd);
	DBG(1, "API:%s(%p) completed %d\n",__FUNCTION__,pdev,retval);
	return retval;
}

#else

#define	vmb_suspend	NULL
#define	vmb_resume	NULL

#endif

/* this driver is exported so vmb_cs can depend on it */
struct platform_driver vmb_driver = {
	.driver = {
		.name = (char *) vmb_name,
		.owner = THIS_MODULE,
	},

	.probe =	vmb_probe,
	.remove =	__devexit_p(vmb_remove),

//	.shutdown =	__(struct platform_device *),
	.suspend =	vmb_suspend,
	.resume =	vmb_resume,
};

EXPORT_SYMBOL_GPL(vmb_driver);

/*-------------------------------------------------------------------------*/

static int __init init_vmb_hcd(void)
{
	spin_lock_init(&vmb_static_lock);
	INIT_LIST_HEAD(&vmb_static_list);
	vmb_instances = 0;

	DBG(1, "I=%d A=%d C=%p %s()\n",(int)in_interrupt(),in_atomic(),current,__FUNCTION__);

	if (usb_disabled())
		return -ENODEV;

	INFO("driver %s %s pc_retry=%d pc_debug=%d pc_trace=%08X built at %s on %s\n"
			,vmb_name, DRIVER_VERSION,pc_retry, pc_debug, pc_trace, __TIME__, __DATE__);

	workqueue = create_singlethread_workqueue("vmb");

	int retval = platform_driver_register(&vmb_driver);

	DBG(3, "%s() completed %d\n",__FUNCTION__,retval);
	return retval;
}
module_init(init_vmb_hcd);

static void __exit exit_vmb_hcd(void)
{
	DBG(1, "I=%d A=%d C=%p %s()\n",(int)in_interrupt(),in_atomic(),current,__FUNCTION__);

	platform_driver_unregister(&vmb_driver);

	flush_workqueue(workqueue);
	destroy_workqueue(workqueue);

	DBG(3, "%s() completed\n",__FUNCTION__);
}
module_exit(exit_vmb_hcd);
