/*
 * Author: Tony Olech - Elan Digital Systems Limited
 * $Revision: 1.82 $
 * $Date: 2007/12/20 15:06:18 $
 *
 * This VMB scheduler is event driven and is kicked by calling "queue_work(workqueue,&vmb->work)"
 *
 * Interfaces to kernel 2.6.19
 *
 */

// each urb controlled carries it's own state machine in urb->error_count

#define PRINTK(args...) do{}while(0)

#define isprint(c)      ((c > 0x1f) && (c < 0x7f))

static struct
{
	unsigned length:10;
	unsigned utle:4;
	unsigned frame:12;
	unsigned data01:1;
	unsigned crc:16;

	u8 data[1024];	// no way round this buffering because the UTLE is read last from the footer

} input_packet;

static struct
{
	unsigned utle:4;
	unsigned status:2;
	unsigned id:12;
	unsigned exterr:3;

} error_packet;
static u16 vmb_read_tler(const char*function,const unsigned line,struct vmb*vmb,u8 td)   // 0 =< n < 16
{
	u16 tler = 0xFFFF;
	u16 long_stop = 1000;

    if (td<16) {
		// correct
	} else {
		assertion_failure(function,__FILE__,line,"TLER index n >= 16");
	}

	do {
		int read_tries = 0;
		
		while(0xFFFF == tler && 0 == vmb->unplugged && read_tries++ < 256) {
		
			tler = readw((void*)(0x0360+td+td+vmb->addr));
		}

		if (vmb->unplugged) {
    		return tler;
		} else if (0xFFFF == tler) {
			u8 sr = read_sr(vmb);
			if (0xff == sr) {
				vmb->unplugged = 1;	
				ERR("%s(UNPLUGGED) called from %s:%d\n",__FUNCTION__,function,line);
    			return tler;
			} else {
				continue;	
			}
		} else if (read_tries > 1) {
			DBG(0,"TLER[%X] read as 0xFFFF %d times before giving %04X\n",td,read_tries,tler);
    		return tler;
		} else {
    		return tler;
		}
	} while(long_stop--);
	ERR("%s:%d TLER[%X] read consistantly as 0xFFFF\n",function,line,td);
	return tler;
}	
static void dump_slots(struct vmb*vmb)
{
	u8 n = 0;
	u8 slots = 16;

	while(slots-- > 0) {

		u8 td = n++;
		struct vmb_slot*slot = &vmb->slot[td];
		ASSERT(td == slot->td);
		ASSERT(vmb == slot->vmb);
		char slot_flags[64];
	
		snprintf(slot_flags,sizeof(slot_flags),"ID=%d NX=%02X QN=%02X QS=%02X flags=%c%c%c%c%c%c%c%c ec=%X"
				,slot->idref_done,slot->next_urb_index,slot->queue_next,slot->queue_size
				,slot->allocated?'A':'-'
				,slot->initialized?'I':'-'
				,slot->retired?'R':'-'
				,slot->running?'r':'-'
				,slot->actioned?'a':'-'
				,slot->end_seen?'E':'-'
				,slot->retry_flag?'R':'-'
				,slot->device_error?'D':'-'
				,slot->ec);
		if (0 == slot->initialized) {
			continue;
		} else if (PIPE_CONTROL == slot->pipetype) {
			PRINTK(KERN_DEBUG "SLOT[%d] CONTROL <-> %d:%d max=%d/%d flags=%s\n",td,slot->usb_addr,slot->usb_endp
							,slot->max_plen,slot->maxusbpktlen,slot_flags);
		} else if (PIPE_INTERRUPT == slot->pipetype) {
			PRINTK(KERN_DEBUG "SLOT[%d] INT     <-- %d:%d max=%d/%d flags=%s\n",td,slot->usb_addr,slot->usb_endp
							,slot->max_plen,slot->maxusbpktlen,slot_flags);
		} else if (slot->pipeout) {
			if (PIPE_ISOCHRONOUS == slot->pipetype) {
				PRINTK(KERN_DEBUG "SLOT[%d] ISO  OUT--> %d:%d max=%d/%d flags=%s\n",td,slot->usb_addr,slot->usb_endp
								,slot->max_plen,slot->maxusbpktlen,slot_flags);
			} else {
				PRINTK(KERN_DEBUG "SLOT[%d] BULK OUT--> %d:%d max=%d/%d flags=%s\n",td,slot->usb_addr,slot->usb_endp
								,slot->max_plen,slot->maxusbpktlen,slot_flags);
			}			
		} else {	
			if (PIPE_ISOCHRONOUS == slot->pipetype) {
				PRINTK(KERN_DEBUG "SLOT[%d] ISO  IN <-- %d:%d max=%d/%d flags=%s\n",td,slot->usb_addr,slot->usb_endp
								,slot->max_plen,slot->maxusbpktlen,slot_flags);
			} else {
				PRINTK(KERN_DEBUG "SLOT[%d] BULK IN <-- %d:%d max=%d/%d flags=%s\n",td,slot->usb_addr,slot->usb_endp
								,slot->max_plen,slot->maxusbpktlen,slot_flags);
			}			
		}
	}
}	
static const int max_size_xlat[4] = {8,16,32,64};
static void ptr_debug(const char*function,const unsigned line,struct vmb *vmb)
{
	HW_LOCK(vmb);
	u16 start_fc = read_fc(vmb);	// the frame currently or recently finished executing
	u16 txwp = read_txwp(vmb);
	u16 txrp = read_txrp(vmb);
	u8  txpc = read_txpc(vmb);	// might not be in step with txwp and txrp
	u16 rxwp = read_rxwp(vmb);
	u16 rxrp = read_rxrp(vmb);
	u16 txit = read_txit(vmb);
	HW_UNLOCK(vmb);
	if (vmb->diag_txwp == txwp
	&&  vmb->diag_txrp == txrp
	&&  vmb->diag_rxwp == rxwp
	&&  vmb->diag_rxrp == rxrp
	&&  vmb->diag_txit == txit
	) {
	} else {
		u16 fifo_bytes = (txwp - txrp) & VMB_FIFO_MASK;
		vmb->diag_txwp = txwp;
		vmb->diag_txrp = txrp;
		vmb->diag_rxwp = rxwp;
		vmb->diag_rxrp = rxrp;
		vmb->diag_txit = txit;

		PRINTK(KERN_DEBUG "%s:%d(FC=%04X T=%04X H=%04X rxR=%04X rxW=%04X txR=%04X txW=%04X txL=%04X txT=%04X txPC=%02X)\n"
				,function,line,start_fc,vmb->frame_tail,vmb->frame_head,rxrp,rxwp,txrp,txwp,fifo_bytes,txit,txpc);
	}
}
static void utl_debug(const char*function,const unsigned line,struct vmb *vmb)
{
	vmb_debug(function,line,vmb);

	int slots = 16;
	int index = 0;
	int count = 0;
	while(slots-- > 0) {
		u8 td = index++;
		struct vmb_slot*slot = &vmb->slot[td];
		ASSERT(td == slot->td);
		ASSERT(vmb == slot->vmb);

		if (0 == slot->allocated) {
			continue;
		} else if (0 == slot->initialized) {
			continue;
		} else {
			count += 1;
		}	
	}
	if (0 == count) {
	} else {
		int N = 0;
		int C = 16;
		u16 max_plen[2];
	
		max_plen[0] = read_mplr(vmb,0);
		max_plen[1] = read_mplr(vmb,1);
	
		u16 txrp = read_txrp(vmb);
		u16 txwp = read_txwp(vmb);
	
		PRINTK(KERN_DEBUG "N P -> D:EP TLER  type mx --urb--- --usb--- (slot_count=%X next_slot=%08X max=%04X%04X TxR=%04X TxW=%04X)\n",vmb->slot_count,vmb->next_slot,max_plen[1],max_plen[0],txrp,txwp);
	
	 	while(C-- > 0)
		{
			int n = N++;
			struct vmb_slot*slot = &vmb->slot[n];
	
			if (0 == slot->allocated) {
				continue;
			} else if (0 == slot->initialized) {
				continue;
			} else {
				char slot_flags[64];
		
				snprintf(slot_flags,sizeof(slot_flags),"ID=%d %02X:%02X QS=%d flags=%c%c%c%c%c%c%c%c ec=%X"
							,slot->idref_done,slot->next_urb_index,slot->queue_next,slot->queue_size
					,slot->allocated?'A':'-'
					,slot->initialized?'I':'-'
					,slot->retired?'R':'-'
					,slot->running?'r':'-'
					,slot->actioned?'a':'-'
					,slot->end_seen?'E':'-'
					,slot->retry_flag?'R':'-'
					,slot->device_error?'D':'-'
					,slot->ec
		  );
		
				u16 tler = READ_TLER(vmb,n);
		
				int size = max_size_xlat[0x03 & (max_plen[n >> 3] >> (0x0f & (n << 1)))];
		
				struct urb *urb = slot_to_head_urb(slot);
		
				if (urb) {
		
					URB_ASSERT(vmb == urb->hcpriv);
		
					u8 usb_addr = usb_pipedevice(urb->pipe);
					u8 usb_endp = usb_pipeendpoint(urb->pipe);
					u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
		
					//URB_ASSERT(-EINPROGRESS == urb->status || -ECONNRESET == urb->status || -ENOENT == urb->status || -ESTRPIPE == urb->status);
					URB_ASSERT(slot->pipetype == usb_pipetype(urb->pipe));
					URB_ASSERT(slot->usb_addr == usb_pipedevice(urb->pipe));
					URB_ASSERT(slot->usb_endp == usb_pipeendpoint(urb->pipe));
					URB_ASSERT(slot->pipeout == usb_pipeout(urb->pipe));	
		
					char urb_buffer[32];
	
					if (slot->maxusbpktlen == maxusbpktlen) {	
						snprintf(urb_buffer,sizeof(urb_buffer),"length(A=%d T=%d M=%d)"
									,urb->actual_length,urb->transfer_buffer_length,maxusbpktlen);
					} else {
						snprintf(urb_buffer,sizeof(urb_buffer),"length(A=%d T=%d M=%d/%d)"
									,urb->actual_length,urb->transfer_buffer_length,maxusbpktlen,slot->maxusbpktlen);
					}
		
					char*type = (0x0001  == (0x0039 & tler)) ? " OUT "
					          : (0x0009  == (0x0039 & tler)) ? "  IN "
					          : (0x0011  == (0x0039 & tler)) ? "SETUP"
					          : (0x0019  == (0x0039 & tler)) ? "I_ONE"
					          : (0x0021  == (0x0039 & tler)) ? "O_ISO"
					          : (0x0029  == (0x0039 & tler)) ? "I_ISO"
					          : (0x0031  == (0x0039 & tler)) ? "I_SPD"
					          : (0x0039  == (0x0039 & tler)) ? "*****"
					          : (0x0008  == (0x0039 & tler)) ? "  in "
					          : (0x0010  == (0x0039 & tler)) ? "setup"
					          : (0x0018  == (0x0039 & tler)) ? "i_one"
					          : (0x0020  == (0x0039 & tler)) ? "o_iso"
					          : (0x0028  == (0x0039 & tler)) ? "i_iso"
					          : (0x0030  == (0x0039 & tler)) ? "i_spd"
					          : (0x0038  == (0x0039 & tler)) ? "*****"
					          : (0x0000  != (0xFFC6 & tler)) ? " out "
		                      : "-----"
							  ;
					if (maxusbpktlen == size) {
		
					} else if (PIPE_INTERRUPT == slot->pipetype) {
		
						if (maxusbpktlen == 1) {
						} else if (maxusbpktlen == 2) {
						} else {
							
							PRINTK(KERN_DEBUG "%X **** %d:%02d %04X %s %2d %p %p E_C=%d %s ****ERROR**** usb_maxpacket=%d)\n",n,usb_addr,usb_endp,tler,type,size,urb,urb->dev,urb->error_count,slot_flags,maxusbpktlen);
						}
					} else {
						if (8 == size && maxusbpktlen <= 64) {
		
						} else {
							
							PRINTK(KERN_DEBUG "%X **** %d:%02d %04X %s %2d %p %p E_C=%d %s ****ERROR**** usb_maxpacket=%d)\n",n,usb_addr,usb_endp,tler,type,size,urb,urb->dev,urb->error_count,slot_flags,maxusbpktlen);
						}
					}
		
					unsigned long trigger = slot->jiffies;
					long waiting = trigger - jiffies;
		
					if (0 == usb_addr) {
		
						ASSERT(0 == usb_endp);
		
						if (waiting > 0) {
							PRINTK(KERN_DEBUG "%X ENUMERATE %04X %s %2d %p %p E_C=%d %s %s WAITING=%ld\n",n,tler,type,size,urb,urb->dev,urb->error_count,slot_flags,urb_buffer,waiting);
						} else {
							PRINTK(KERN_DEBUG "%X ENUMERATE %04X %s %2d %p %p E_C=%d %s %s\n",n,tler,type,size,urb,urb->dev,urb->error_count,slot_flags,urb_buffer);
						}
					} else {
		
						ASSERT(usb_addr < MAX_DEV_ADDRS);
			
						struct vmb_udev *udev = &vmb->udev[usb_addr];
			
						URB_ASSERT(udev->usb_device == urb->dev);
			
						if (0 < udev->portnumber) {
			
							struct vmb_port*port = &vmb->port[udev->portnumber - 1];
			
							if (waiting > 0) {
								PRINTK(KERN_DEBUG "%X %d -> %d:%02d %04X %s %2d %p %p E_C=%d %s %s status=%d WAITING=%ld\n",n,port->number,usb_addr,usb_endp,tler,type,size,urb,urb->dev,urb->error_count,slot_flags,urb_buffer,port->status,waiting);
							} else {
								PRINTK(KERN_DEBUG "%X %d -> %d:%02d %04X %s %2d %p %p E_C=%d %s %s status=%d\n",n,port->number,usb_addr,usb_endp,tler,type,size,urb,urb->dev,urb->error_count,slot_flags,urb_buffer,port->status);
							}
						} else {
			
							if (waiting > 0) {
								PRINTK(KERN_DEBUG "%X ? -> %d:%02d %04X %s %2d %p %p E_C=%d %s %s WAITING=%ld\n",n,usb_addr,usb_endp,tler,type,size,urb,urb->dev,urb->error_count,slot_flags,urb_buffer,waiting);
							} else {
								PRINTK(KERN_DEBUG "%X ? -> %d:%02d %04X %s %2d %p %p E_C=%d %s %s\n",n,usb_addr,usb_endp,tler,type,size,urb,urb->dev,urb->error_count,slot_flags,urb_buffer);
							}
						}
					}
				} else if (slot->usb_addr > 0 || slot->usb_endp > 0) {
		
					PRINTK(KERN_DEBUG "%X OVER %d:%02d %04X %2d extra=%d %s\n",n,slot->usb_addr,slot->usb_endp,tler,size,slot->extra,slot_flags);
				} else if ((tler & 0xFFFE) != 0x0000) {
					PRINTK(KERN_DEBUG "%X GONE AWAY %04X %2d extra=%d %s\n",n,tler,size,slot->extra,slot_flags);
				} else if (0 < slot->idref_done || 0 < slot->queue_size) {
					PRINTK(KERN_DEBUG "%X GONE AWAY %04X %2d extra=%d %s\n",n,tler,size,slot->extra,slot_flags);
				} else {
					if (pc_debug > 2) PRINTK(KERN_DEBUG "%X -         %04X %2d - no urb\n",n,tler,size);
				}
			}
		}
	}
}
static void urb_setup(const char*function,struct urb *urb)
{
	char msg[1+3*8];
	int len = (sizeof(msg)/sizeof(char) - 1)/3;
	char*m=msg;
	int M = len;
	u8*b = urb->setup_packet;

	while(M-- > 0)	m += sprintf(m," %02X",*b++);

	*m = 0;

	u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
	
	PRINTK(KERN_DEBUG "%s(max=%d len=%d %d~SETUP->%d%s)\n",function,maxusbpktlen,urb->transfer_buffer_length,usb_pipedevice(urb->pipe),usb_pipeendpoint(urb->pipe),msg);
}
static void pkt_error(const char*function,struct urb *urb)
{
	ASSERT(urb);

	PRINTK(KERN_DEBUG "%s(TD=%X urb=%p error_count=%d ERROR PACKET id=%04X status=%d exterr=%d)\n",function,error_packet.utle,urb,urb ? urb->error_count : -1,error_packet.id,error_packet.status,error_packet.exterr);

	struct vmb*vmb = urb_to_vmb(urb);
	ASSERT(vmb);

}
static void pkt_debug(const char*function,struct urb *urb)
{
		char msg[4+3*24];
		int len = (sizeof(msg)/sizeof(char) - 1)/3;
		char*m=msg;
		int M = len = input_packet.length > 24 ? 24 : input_packet.length;

		u8*b = input_packet.data;

		while(M-- > 0)	m += sprintf(m," %02X",*b++);

		if (input_packet.length > 24) { m += sprintf(m," .."); }

		*m = 0;

		PRINTK(KERN_DEBUG "%s(urb=%p error_count=%d INPUT PACKET crc=%04X utle=%X frame=%04X data01=%d)\n",function,urb,urb ? urb->error_count : -1,input_packet.crc,input_packet.utle,input_packet.frame,input_packet.data01);
	
		PRINTK(KERN_DEBUG "TD=%X urb=%p input_packet(%d bytes)=%s\n",input_packet.utle,urb,input_packet.length,msg);
}
static void slot_debug(const char*function,const unsigned line,struct vmb_slot *slot)
{
	PRINTK(KERN_DEBUG "%s:%d SLOT[%X] %d:%02d %s %s ID=%04X flags=%c%c%c%c%c%c%c%c ec=%X\n"
					,function,line
					,slot->td,slot->usb_addr,slot->usb_endp
					,slot->pipetype == PIPE_BULK?"BULK"
					:slot->pipetype == PIPE_ISOCHRONOUS?"ISO"
					:slot->pipetype == PIPE_INTERRUPT?"INTERRUPT"
					:"CONTROL"
					,slot->pipeout?"OUT":" IN"
					,slot->write__id
					,slot->allocated?'A':'-'
					,slot->initialized?'I':'-'
					,slot->retired?'R':'-'
					,slot->running?'r':'-'
					,slot->actioned?'a':'-'
					,slot->end_seen?'E':'-'
					,slot->retry_flag?'R':'-'
					,slot->device_error?'D':'-'
					,slot->ec
		  );
	{
		PRINTK(KERN_DEBUG "U_R_B next_urb_index=%02X next_urb_count=%04X queue_size=%02X queue_next=%04X\n"
					,slot->next_urb_index
					,slot->next_urb_count
					,slot->queue_size
					,slot->queue_next
			  );
		int i = 0;
		int I = SLOT_QUEUE_SIZE/8;
		while(I--) {
			char message[96]; // 85 should be enough
			int m = sprintf(message,"u_r_b[%02X..]=",i);
			int f = 0;
			int J = 8;
			while(J--) {
				struct urb *urb = slot->u_r_b[i++];
				m += sprintf(message+m," %p",urb);
				f += (urb != NULL);
			}	
			if (f > 0) PRINTK(KERN_DEBUG "%s\n",message);
		}
	}
	{
		PRINTK(KERN_DEBUG "IDREF done=%02X this=%04X pkts=%04X\n"
					,slot->idref_done
					,slot->idref_this
					,slot->idref_pkts
					);
		int i = 0;
		int I = SLOT_IDREF_SIZE;
		PRINTK(KERN_DEBUG "idref[]=");
		while(I--) {
			struct slot_idref *idref = &slot->idref[i++];
			PRINTK(" Q=%02X,B=%04X,L=%04X",idref->queue_this,idref->frame_base,idref->frame_last);
		}
		PRINTK("\n");
	}
	{
		u8 queue_size = slot->queue_size;
		u16 queue_item = slot->queue_next - slot->queue_size;

		PRINTK(KERN_DEBUG "slot[%X] %d:%02d %s %s has %d queued urbs\n"
					,slot->td,slot->usb_addr,slot->usb_endp
					,slot->pipetype == PIPE_BULK?"BULK"
					:slot->pipetype == PIPE_ISOCHRONOUS?"ISO"
					:slot->pipetype == PIPE_INTERRUPT?"INTERRUPT"
					:"CONTROL"
					,slot->pipeout?"OUT":" IN"
					,queue_size);

		while(queue_size-- > 0) {
			u8 offset = SLOT_QUEUE_MASK & queue_item++;	
			struct urb *urb_entry = slot->u_r_b[offset];
			if (usb_pipein(urb_entry->pipe)) {
		
				PRINTK(KERN_DEBUG "u_r_b[%02X]=%p e_c=%d xferlen=%d actlen=%d %d~INPUT<-%d\n"
					,offset,urb_entry,urb_entry->error_count,urb_entry->transfer_buffer_length,urb_entry->actual_length,usb_pipedevice(urb_entry->pipe),usb_pipeendpoint(urb_entry->pipe));
			} else {
		
				PRINTK(KERN_DEBUG "u_r_b[%02X]=%p e_c=%d xferlen=%d actlen=%d %d~OUTPUT->%d\n"
					,offset,urb_entry,urb_entry->error_count,urb_entry->transfer_buffer_length,urb_entry->actual_length,usb_pipedevice(urb_entry->pipe),usb_pipeendpoint(urb_entry->pipe));
			}
		}
	}
	if (!in_atomic()) msleep(2);
}
static void port_debug(struct vmb *vmb,struct vmb_slot *slot)
{
	u8 usb_addr = slot->usb_addr;
	struct vmb_udev*udev = &vmb->udev[usb_addr];

	UTL_DEBUG(0,vmb);

	u8 pcs[10];
try_next_frame: {
		int I = 10;
		int i = 0;
		u16 fc_initial = read_fc(vmb);
read_next_sample: if (I-- > 0) {
			pcs[i++] = read_pcsr(vmb);
			if (read_fc(vmb) == fc_initial) {
				goto read_next_sample;	
			} else {
				goto try_next_frame;
			}
		}	
	}	
	PRINTK(KERN_DEBUG "PCSR= %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
			pcs[0],pcs[1],pcs[2],pcs[3],pcs[4],pcs[5],pcs[6],pcs[7],pcs[8],pcs[9]);
	PRINTK(KERN_DEBUG "SLOT[%X] -- %d:%d rejected(%d active slots) port1 udev=%p status=%d number=%d %c%c%c%c %08X port1 udev=%p status=%d number=%d %c%c%c%c %08X\n"
				,slot->td,slot->usb_addr,slot->usb_endp,MAX_VMB_SLOTS-vmb->slot_count
				,vmb->port[0].udev,vmb->port[0].status,vmb->port[0].number
				,vmb->port[0].enable?'e':'-'
				,vmb->port[0].device?'d':'-'
				,vmb->port[0].reset?'r':'-'
				,vmb->port[0].power?'p':'-'
				,vmb->port[0].Status
				,vmb->port[1].udev,vmb->port[1].status,vmb->port[1].number
				,vmb->port[1].enable?'e':'-'
				,vmb->port[1].device?'d':'-'
				,vmb->port[1].reset?'r':'-'
				,vmb->port[1].power?'p':'-'
				,vmb->port[1].Status
				);
	u8 u = MAX_DEV_ADDRS;
	u8 count = 0;
	while(u-- > 0)
	{
		if (vmb->udev[u].usb_device) {
			PRINTK(KERN_DEBUG "UDEV[%d] usb_device=%p portnumber=%d devicefail=%d\n",u,vmb->udev[u].usb_device,vmb->udev[u].portnumber,vmb->udev[u].devicefail);
			count += 1;
		}
	}
	if (0 == count) {
		PRINTK(KERN_DEBUG "NO UDEV's found\n");	
	}
	if (1 == udev->portnumber) {
		int I = 10;
		int i = 0;
test_p1_sample: if (I-- > 0) {
			u8 pc = pcs[i++];
			if ((0x35&pc) == 0x01) goto test_p1_sample;
		} else {
			PRINTK(KERN_DEBUG "SLOT[%X] rejected probably because port 1 is in SE0 state\n",slot->td);
		}
	} else if (2 == udev->portnumber) {
		int I = 10;
		int i = 0;
test_p2_sample: if (I-- > 0) {
			u8 pc = pcs[i++];
			if ((0xCA&pc) == 0x02) goto test_p2_sample;
		} else {
			PRINTK(KERN_DEBUG "SLOT[%X] rejected probably because port 2 is in SE0 state\n",slot->td);
		}
	} else {
		PRINTK(KERN_DEBUG "SLOT[%X] rejected but it is not associated to a specific port - probably connected via a hub\n",slot->td);
	}
}			
static void urb_debug(const char*function,const unsigned line,struct urb *urb)
{
	char msg[1+3*20];
	int len = (sizeof(msg)/sizeof(char) - 1)/3;
	char*m=msg;
	int M = urb->actual_length < len ? urb->actual_length : len;
	u8*b = urb->transfer_buffer;
	struct vmb_slot *slot = urb_to_slot(urb);

	while(M-- > 0)	m += sprintf(m," %02X",*b++);

	*m = 0;

	if (slot) {
//		u8 queue_size = slot->queue_size;
//		u16 queue_item = slot->queue_next - slot->queue_size;

		int position = find_urb_position_in_slot(slot,urb);
		if (usb_pipein(urb->pipe)) {

			PRINTK(KERN_DEBUG "%s:%d(urb[%X]=%p error_count=%d len=%d %d~INPUT<-%d%s)\n",function,line,position,urb,urb->error_count,urb->actual_length,usb_pipedevice(urb->pipe),usb_pipeendpoint(urb->pipe),msg);
		} else {

			PRINTK(KERN_DEBUG "%s:%d(urb[%X]=%p error_count=%d len=%d %d~OUTPUT->%d%s)\n",function,line,position,urb,urb->error_count,urb->actual_length,usb_pipedevice(urb->pipe),usb_pipeendpoint(urb->pipe),msg);
		}
		slot_debug(function,line,slot);
//		PRINTK(KERN_DEBUG "slot[%X] %d:%02d %s %s has %d queued urbs\n"
//					,slot->td,slot->usb_addr,slot->usb_endp
//					,slot->pipetype == PIPE_BULK?"BULK"
//					:slot->pipetype == PIPE_ISOCHRONOUS?"ISO"
//					:slot->pipetype == PIPE_INTERRUPT?"INTERRUPT"
//					:"CONTROL"
//					,slot->pipeout?"OUT":" IN"
//					,queue_size);
//
//		while(queue_size-- > 0) {
//			u8 offset = SLOT_QUEUE_MASK & queue_item++;	
//			struct urb *urb_entry = slot->u_r_b[offset];
//			if (usb_pipein(urb_entry->pipe)) {
//		
//				PRINTK(KERN_DEBUG "u_r_b[%02X]=%p error_count=%d len=%d %d~INPUT<-%d\n",offset,urb_entry,urb_entry->error_count,urb_entry->actual_length,usb_pipedevice(urb_entry->pipe),usb_pipeendpoint(urb_entry->pipe));
//			} else {
//		
//				PRINTK(KERN_DEBUG "u_r_b[%02X]=%p error_count=%d len=%d %d~OUTPUT->%d\n",offset,urb_entry,urb_entry->error_count,urb_entry->transfer_buffer_length,usb_pipedevice(urb_entry->pipe),usb_pipeendpoint(urb_entry->pipe));
//			}
//		}
	} else {
		if (usb_pipein(urb->pipe)) {

			PRINTK(KERN_DEBUG "%s(urb=%p error_count=%d len=%d %d~INPUT<-%d%s)\n",function,urb,urb->error_count,urb->actual_length,usb_pipedevice(urb->pipe),usb_pipeendpoint(urb->pipe),msg);
		} else {

			PRINTK(KERN_DEBUG "%s(urb=%p error_count=%d len=%d %d~OUTPUT->%d%s)\n",function,urb,urb->error_count,urb->actual_length,usb_pipedevice(urb->pipe),usb_pipeendpoint(urb->pipe),msg);
		}
	}
}
static void queue_debug(const char*function,const unsigned line,struct vmb *vmb)
{
	utl_debug(function,line,vmb);

	int slots = 16;
	int index = 0;
	while(slots-- > 0) {
		u8 td = index++;
		struct vmb_slot*slot = &vmb->slot[td];
		ASSERT(td == slot->td);
		ASSERT(vmb == slot->vmb);

		if (0 == slot->allocated) {
			continue;
		} else if (0 == slot->initialized) {
			continue;
		} else {
			slot_debug(function,line,slot);
		}
	}
}
static void vmb_check_rx_fifo(struct vmb*vmb);
/*
 * vmb_urb_completed() is only called whilst holding the SW_LOCK
 *
 */
#define VMB_URB_COMPLETED(vmb,slot,urb,e,s)	vmb_urb_completed(__FUNCTION__,__LINE__,vmb,slot,urb,e,s)
static void vmb_urb_giveback(const char*function,const int line_number,struct vmb*vmb,struct urb*urb,int error_count,int status)
{
	urb->error_count = error_count;
	if (urb->status == -EINPROGRESS) urb->status = status;
	urb->hcpriv = NULL;

try_again:;
    if (vmb->giveback_size < URB_GIVEBACK_SIZE) {
		HW_LOCK(vmb);
        vmb->giveback[URB_GIVEBACK_MASK & vmb->giveback_next++] = urb;
		vmb->giveback_size += 1;
		HW_UNLOCK(vmb);

		URB_DBG(7,"%s(urb=%p successfully pushed onto giveback queue) queue size = %d\n",__FUNCTION__,urb,vmb->giveback_size);
		return;
	} else {
		HW_LOCK(vmb);
		struct urb*urb = vmb->giveback[URB_GIVEBACK_MASK & (vmb->giveback_next - vmb->giveback_size--)];
		HW_UNLOCK(vmb);
		ERR("%s(giveback queue overflowed handing back urb=%p)\n",__FUNCTION__,urb);
		SW_UNLOCK(vmb);
		struct usb_hcd  *hcd = vmb_to_hcd(vmb);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
		usb_hcd_giveback_urb(hcd,urb,NULL);
#else
		usb_hcd_giveback_urb(hcd,urb);
#endif
		SW_LOCK(vmb);
		goto try_again;	// we temporarily relinquished the SW_LOCK so we must retest
	}
}
static int vmb_urb_completed(const char*function,const int line_number,struct vmb*vmb,struct vmb_slot *slot,struct urb*urb,int error_count,int status)
{
	ASSERT(0 < in_vmb_scheduler);	// we hold the SW_LOCK

	ASSERT(vmb);
	ASSERT(slot);
	ASSERT(urb);

	UTL_ASSERT(slot == urb_to_slot(urb));
	URB_ASSERT(urb == slot_to_head_urb(slot));

	URB_DEBUG(4,urb);

	if (0 < slot->extra) {
		URB_DBG(2, "%s:%d(TD=%X usb_hcd_giveback_urb=%p E_C=%d error_count=%d) actual_length=%d overspill=%d result status=%d\n",function,line_number,slot->td,urb,urb->error_count,error_count,urb->actual_length,slot->extra,status);
	} else {
		URB_DBG(5, "%s:%d(TD=%X usb_hcd_giveback_urb=%p E_C=%d error_count=%d) actual_length=%d result status=%d\n",function,line_number,slot->td,urb,urb->error_count,error_count,urb->actual_length,status);
	}

	URB_DBG(0,"%s(%s:%d is stacking urb=%p from TD=%X QS=%d onto giveback queue size = %d)\n"
			,__FUNCTION__,function,line_number,urb,slot->td,slot->queue_size,vmb->giveback_size);
	int empty = remove_slot_head_urb(slot,urb);

	vmb_urb_giveback(function,line_number,vmb,urb,error_count,status);
	return empty;
}
static void vmb_urb_rejected(struct vmb*vmb,struct urb*urb)
{
	ASSERT(0 < in_vmb_scheduler);	// we hold the SW_LOCK

	ASSERT(vmb);
	ASSERT(urb);

	URB_ASSERT(VMB_URB_STATE_RETURNED_TO_CALLER == urb->error_count);

try_again:;
    if (vmb->giveback_size < URB_GIVEBACK_SIZE) {
		URB_DBG(1,"%s(stacking urb=%p onto giveback queue) queue size = %d\n",__FUNCTION__,urb,vmb->giveback_size);
		HW_LOCK(vmb);
        vmb->giveback[URB_GIVEBACK_MASK & vmb->giveback_next++] = urb;
		vmb->giveback_size += 1;
		HW_UNLOCK(vmb);

	} else {
		HW_LOCK(vmb);
		struct urb*urb = vmb->giveback[URB_GIVEBACK_MASK & (vmb->giveback_next - vmb->giveback_size--)];
		HW_UNLOCK(vmb);
		ERR("%s(giveback queue overflowed handing back urb=%p)\n",__FUNCTION__,urb);
		SW_UNLOCK(vmb);
		struct usb_hcd  *hcd = vmb_to_hcd(vmb);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
		usb_hcd_giveback_urb(hcd,urb,NULL);
#else
		usb_hcd_giveback_urb(hcd,urb);
#endif
		SW_LOCK(vmb);
		goto try_again;	// we temporarily relinquished the SW_LOCK so we must retest
	}
}
/*
 * vmb_giveback_urbs() is only called whilst NOT holding the SW_LOCK
 *
 *
 */
static void vmb_giveback_urbs(struct vmb*vmb)
{
	HW_LOCK(vmb);
	while(vmb->giveback_size > 0) {
		struct urb*urb = vmb->giveback[URB_GIVEBACK_MASK & (vmb->giveback_next - vmb->giveback_size--)];
		HW_UNLOCK(vmb);

		ASSERT(urb);

		if (usb_pipein(urb->pipe) && (PIPE_BULK == usb_pipetype(urb->pipe))) {
			char msg[2+4*16];
			int n = (sizeof(msg)/sizeof(char) - 2)/4;
			int len = min(urb->actual_length,n);
			char*m=msg;
			int B = len;
			int A = len;
			u8*b = urb->transfer_buffer;
			u8*a = urb->transfer_buffer;

			while(B-- > 0) {
				m += sprintf(m," %02X",*b++);
			}	
			m += sprintf(m," ");
			while(A-- > 0) {
				char c = *a++;	
				m += sprintf(m,"%c",isprint(c)?c:'?');
			}	
			*m = 0;

			//INFO("%d bytes BULK IN =%s\n",urb->actual_length,msg);
		}

		struct usb_hcd  *hcd = vmb_to_hcd(vmb);

		URB_DBG(0,"%s(removing urb=%p len=%d from giveback queue) queue size = %d\n",__FUNCTION__,urb,urb->actual_length,vmb->giveback_size);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
		usb_hcd_giveback_urb(hcd,urb,NULL);
#else
		usb_hcd_giveback_urb(hcd,urb);
#endif

		HW_LOCK(vmb);
	}
	HW_UNLOCK(vmb);
}
//-----------------------------------------------------------------------------------------
//01		txrp > txwp && fifo > txwp && fifo > txrp	wait
//02		txrp > txwp && fifo > txwp && fifo = txrp	wait
//03		txrp > txwp && fifo > txwp && fifo < txrp	done
//04		txrp > txwp && fifo = txwp && fifo > txrp	xxxx
//05		txrp > txwp && fifo = txwp && fifo = txrp	xxxx
//06		txrp > txwp && fifo = txwp && fifo < txrp	done
//07		txrp > txwp && fifo < txwp && fifo > txrp	xxxx
//08		txrp > txwp && fifo < txwp && fifo = txrp	xxxx
//09		txrp > txwp && fifo < txwp && fifo < txrp	wait
//10		txrp = txwp && fifo > txwp && fifo > txrp	done
//11		txrp = txwp && fifo > txwp && fifo = txrp	xxxx
//12		txrp = txwp && fifo > txwp && fifo < txrp	xxxx
//13		txrp = txwp && fifo = txwp && fifo > txrp	xxxx
//14		txrp = txwp && fifo = txwp && fifo = txrp	done
//15		txrp = txwp && fifo = txwp && fifo < txrp	xxxx
//16		txrp = txwp && fifo < txwp && fifo > txrp	xxxx
//17		txrp = txwp && fifo < txwp && fifo = txrp	xxxx
//18		txrp = txwp && fifo < txwp && fifo < txrp	done
//19		txrp < txwp && fifo > txwp && fifo > txrp	done
//20		txrp < txwp && fifo > txwp && fifo = txrp	xxxx
//21		txrp < txwp && fifo > txwp && fifo < txrp	xxxx
//22		txrp < txwp && fifo = txwp && fifo > txrp	done
//23		txrp < txwp && fifo = txwp && fifo = txrp	xxxx
//24		txrp < txwp && fifo = txwp && fifo < txrp	xxxx
//25		txrp < txwp && fifo < txwp && fifo > txrp	wait
//26		txrp < txwp && fifo < txwp && fifo = txrp	wait
//27		txrp < txwp && fifo < txwp && fifo < txrp	done
//-----------------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------------
//04		txrp > txwp && fifo = txwp && fifo > txrp	xxxx
//05		txrp > txwp && fifo = txwp && fifo = txrp	xxxx
//06		txrp > txwp && fifo = txwp && fifo < txrp	done
//13		txrp = txwp && fifo = txwp && fifo > txrp	xxxx
//14		txrp = txwp && fifo = txwp && fifo = txrp	done
//15		txrp = txwp && fifo = txwp && fifo < txrp	xxxx
//22		txrp < txwp && fifo = txwp && fifo > txrp	done
//23		txrp < txwp && fifo = txwp && fifo = txrp	xxxx
//24		txrp < txwp && fifo = txwp && fifo < txrp	xxxx
//..........................................................
//01		txrp > txwp && fifo > txwp && fifo > txrp	wait
//02		txrp > txwp && fifo > txwp && fifo = txrp	wait
//03		txrp > txwp && fifo > txwp && fifo < txrp	done
//07		txrp > txwp && fifo < txwp && fifo > txrp	xxxx
//08		txrp > txwp && fifo < txwp && fifo = txrp	xxxx
//09		txrp > txwp && fifo < txwp && fifo < txrp	wait
//10		txrp = txwp && fifo > txwp && fifo > txrp	done
//11		txrp = txwp && fifo > txwp && fifo = txrp	xxxx
//12		txrp = txwp && fifo > txwp && fifo < txrp	xxxx
//16		txrp = txwp && fifo < txwp && fifo > txrp	xxxx
//17		txrp = txwp && fifo < txwp && fifo = txrp	xxxx
//18		txrp = txwp && fifo < txwp && fifo < txrp	done
//19		txrp < txwp && fifo > txwp && fifo > txrp	done
//20		txrp < txwp && fifo > txwp && fifo = txrp	xxxx
//21		txrp < txwp && fifo > txwp && fifo < txrp	xxxx
//25		txrp < txwp && fifo < txwp && fifo > txrp	wait
//26		txrp < txwp && fifo < txwp && fifo = txrp	wait
//27		txrp < txwp && fifo < txwp && fifo < txrp	done
//-----------------------------------------------------------------------------------------
//                     scheduler support functions
//-----------------------------------------------------------------------------------------
//
// packets sent out are assumed to have been actually sent when
//         the frame in which they were placed has become history
// 
// it is possible that the frames as being dealt with by the TLE
//       out of sync with this drivers best guess, so it is
//       important that this driver is a bit cautious/conservative        
//
// basically this subroutine is and can only be concerned with
//           packets being written to the outward bound FIFO, ie
//           OUT's      
//
//-----------------------------------------------------------------------------------------
static void done_frame(struct vmb*vmb,u16 frame_tail)
{
	SW_ASSERT_LOCKED(vmb);

	ASSERT(0 == (~UTL_FRAME_MASK & frame_tail));

	u8 n = 0;
	u8 slots = 16;
	
	while(slots-- > 0) {
	
		u8 td = n++;
		struct vmb_slot*slot = &vmb->slot[td];
		ASSERT(td == slot->td);
		ASSERT(vmb == slot->vmb);

		if (0 == slot->initialized) {
			continue;	
		} else while(0 < slot->idref_done) {

			SLOT_ASSERT(slot->allocated);

			DBG(3,"%s(TD=%X idref=%d:%d queue=%d running=%d actioned=%d frame_tail=%04X)\n",__FUNCTION__,td,slot->idref_this,slot->idref_done,slot->queue_size,slot->running,slot->actioned,frame_tail);
			SLOT_ASSERT(0 < slot->queue_size);

			u16 idref_head = slot->idref_this - slot->idref_done;

			struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & idref_head];

			if (frame_tail == idref->frame_last) {

				slot->idref_done -= 1;

				struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & idref->queue_this];
				u16 queue_neck = SLOT_QUEUE_MASK & (1+idref->queue_this);

				SLOT_ASSERT((SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)) == (SLOT_QUEUE_MASK & idref->queue_this));
			
				SLOT_ASSERT(urb);
				URB_ASSERT(vmb == urb->hcpriv);
				URB_ASSERT(slot->pipetype == usb_pipetype(urb->pipe));
				URB_ASSERT(slot->usb_addr == usb_pipedevice(urb->pipe));
				URB_ASSERT(slot->usb_endp == usb_pipeendpoint(urb->pipe));
				URB_ASSERT(slot->pipeout == usb_pipeout(urb->pipe));	
				URB_ASSERT(slot->maxusbpktlen == usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe)));

				if (VMB_URB_STATE_CONTROL_SETUP == urb->error_count) {

					URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));
					SLOT_ASSERT(1 == slot->actioned);	// CONTROL SETUP
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT((SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)) == (SLOT_QUEUE_MASK & slot->next_urb_index));

					SLOT_ASSERT(1 == slot->running);	// CONTROL SETUP
					u16 tler = READ_TLER(vmb,td);
					TLER_ASSERT(0 == (tler & 1));	// it should have had EOT set so auto retires
					WRITE_TLER(vmb,td,0x00);	// clear out the TD
					slot->running = 0;	// CONTROL SETUP stopped

					slot->actioned = 0;
					urb->actual_length = 0;
					update_max_plen(vmb,slot->td,urb);

					if (0 == slot->pipeout || 0 == urb->transfer_buffer_length) {
						urb->error_count = VMB_URB_STATE_CONTROL_IN;
						DBG(1, "%s(TD=%X CONTROL_IN urb=%p len=%d idref=%d queue=%d)\n",__FUNCTION__,td,urb,urb->transfer_buffer_length,slot->idref_done,slot->queue_size);
					} else {
						urb->error_count = VMB_URB_STATE_CONTROL_OUT;	
						DBG(1, "%s(TD=%X CONTROL_OUT urb=%p len=%d idref=%d queue=%d)\n",__FUNCTION__,td,urb,urb->transfer_buffer_length,slot->idref_done,slot->queue_size);
					}
				} else if (VMB_URB_STATE_CONTROL_OUT == urb->error_count) {

					URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));
					SLOT_ASSERT(1 == slot->pipeout);
					URB_ASSERT(0 < urb->transfer_buffer_length);
					SLOT_ASSERT(1 == slot->running);	// CONTROL OUT
					u16 tler = READ_TLER(vmb,td);
					TLER_ASSERT(0 == (tler & 1));	// it should have had EOT set so auto retires
					WRITE_TLER(vmb,td,0x00);	// clear out the TD
					slot->running = 0;	// CONTROL OUT stopped

					SLOT_ASSERT(1 == slot->actioned);
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->next_urb_count);
					URB_ASSERT(urb->actual_length == urb->transfer_buffer_length);

					slot->actioned = 0;
					update_max_plen(vmb,slot->td,urb);	// redundant

					urb->error_count = VMB_URB_STATE_CONTROL_ACK_IN;
					DBG(1, "%s(TD=%X CONTROL_ACK IN urb=%p len=%d idref=%d queue=%d)\n",__FUNCTION__,td,urb,urb->transfer_buffer_length,slot->idref_done,slot->queue_size);

					// should we set urb->transfer_buffer_length to ZERO ???
				} else if (VMB_URB_STATE_CONTROL_ACK_OUT == urb->error_count) {

					URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));
					SLOT_ASSERT(1 == slot->actioned);	// CONTROL ACK OUT
					SLOT_ASSERT(1 == slot->queue_size);
					SLOT_ASSERT(0 == slot->next_urb_count);
/* 08-NOV */					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->queue_next) == (SLOT_QUEUE_MASK & slot->next_urb_index));

					SLOT_ASSERT(1 == slot->running);	// CONTROL ACK OUT
					u16 tler = READ_TLER(vmb,td);
					TLER_ASSERT(0 == (tler & 1));
					WRITE_TLER(vmb,td,0x00);	// clear out the TD
					slot->running = 0;	// CONTROL ACK OUT tler stopped

					UTL_DEBUG(4,vmb);
	
					if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {

						SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
						slot->retired = 1;
						slot->usb_addr = 0;
						slot->usb_endp = 0;
						slot->jiffies = jiffies;

						free_vmb_slot(vmb,slot);

					} else {
						ERR("%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) CONTROL urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);

						SLOT_ASSERT(slot->queue_size > 0);
						slot->jiffies = jiffies;

						slot->actioned = 0;

						SLOT_ASSERT(slot->u_r_b[queue_neck]);
						SLOT_ASSERT(PIPE_CONTROL == usb_pipetype(slot->u_r_b[queue_neck]->pipe));
						SLOT_ASSERT(VMB_URB_STATE_CONTROL_SETUP == slot->u_r_b[queue_neck]->error_count);
					}

					UTL_DEBUG(4,vmb);

				} else if (VMB_URB_STATE_INT_WAIT == urb->error_count) {

					URB_ASSERT(PIPE_INTERRUPT == usb_pipetype(urb->pipe));
					SLOT_ASSERT(1 == slot->actioned);	// INTERRUPT WAIT
					SLOT_ASSERT(0 == slot->running);	// INTERRUPT WAIT

					urb->error_count = VMB_URB_STATE_INTERRUPT;
					slot->actioned = 0;	// INTERRUPT

					update_max_plen(vmb,slot->td,urb);

				} else if (VMB_URB_STATE_INTERRUPT == urb->error_count) {

					URB_ASSERT(PIPE_INTERRUPT == usb_pipetype(urb->pipe));
					SLOT_ASSERT(1 == slot->actioned);	// INTERRUPT NAK
					SLOT_ASSERT(1 == slot->running);	// INTERRUPT NAK
					SLOT_ASSERT(0 == slot->retired);	// INTERRUPT NAK

					u16 tler = READ_TLER(vmb,td);
					TLER_ASSERT(0x0018 == (tler & 0x0039));
					TLER_ASSERT(tler & 0x0006);	// it MUST have failed on the previous attempt
					WRITE_TLER(vmb,td,0x0000);	// clear out the TD

					slot->actioned = 0; // INTERRUPT RETRY

					HW_LOCK(vmb);
					u16 start_fc = read_fc(vmb);    // the frame currently or recently finished executing
					HW_UNLOCK(vmb);

					DBG(1,"%s(INTERRUPT %d:%02d urb=%p interval=%d start_fc=%04X frame_last=%04X)\n",__FUNCTION__,slot->usb_addr,slot->usb_endp,urb,urb->interval,start_fc,idref->frame_last);

				} else if (VMB_URB_STATE_BULK_OUT == urb->error_count) {

					URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
					URB_ASSERT(usb_pipeout(urb->pipe));

					DBG(4, "%s(urb=%p) error_count=%d\n",__FUNCTION__,urb,urb->error_count);

					SLOT_ASSERT(0 == slot->actioned);	// BULK OUT
					UTL_DEBUG(4,vmb);
	
					URB_ASSERT(urb->actual_length == urb->transfer_buffer_length);

					DBG(1,"%s(urb=%p) error_count=%d urb_index=%02X queue_next=%02X queue_size=%02X\n",__FUNCTION__,urb,urb->error_count,slot->next_urb_index,slot->queue_next,slot->queue_size);

					if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {

						slot->actioned = 1;
						SLOT_ASSERT(1 == slot->running);	// just finished BULK OUT
						u16 tler = READ_TLER(vmb,td);
						TLER_ASSERT(1 == (tler & 1));	// because the EOT bit was not set
						WRITE_TLER(vmb,td,0x00);	// clear out the TD
						slot->running = 0;	// freeing slot used by BULK OUT
						slot->retired = 1;

						SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
						SLOT_ASSERT(0 == slot->next_urb_count);

						slot->jiffies = jiffies;
						slot->usb_addr = 0;
						slot->usb_endp = 0;

						free_vmb_slot(vmb,slot);

					} else {
						SLOT_ASSERT(slot->queue_size > 0);
						slot->jiffies = jiffies;

						DBG(1, "%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) BULK OUT urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);

						SLOT_ASSERT(1 == slot->running);	// just finished BULK OUT
						SLOT_ASSERT(0 == slot->retired);
						SLOT_ASSERT(slot->u_r_b[queue_neck]);
						SLOT_ASSERT(usb_pipeout(slot->u_r_b[queue_neck]->pipe));
						SLOT_ASSERT(PIPE_BULK == usb_pipetype(slot->u_r_b[queue_neck]->pipe));
						SLOT_ASSERT(VMB_URB_STATE_BULK_OUT == slot->u_r_b[queue_neck]->error_count);
					}

					UTL_DEBUG(4,vmb);

				} else if (VMB_URB_STATE_ISO_OUT == urb->error_count) {

					URB_ASSERT(PIPE_ISOCHRONOUS == usb_pipetype(urb->pipe));
					URB_ASSERT(usb_pipeout(urb->pipe));

					DBG(4, "%s(urb=%p) error_count=%d\n",__FUNCTION__,urb,urb->error_count);

					UTL_DEBUG(4,vmb);

					int error_count = 0;	// number of errors to be reported for ISO transfer

					int packets = urb->number_of_packets;
					u16 p = 0;
					while(packets-- > 0) {

						if (urb->iso_frame_desc[p++].status) {

							error_count += 1;	// number of errors to be reported for ISO transfer
						}
					}
					// error_count is set to number of errors and not to VMB_URB_STATE_RETURNED_TO_CALLER

					if (VMB_URB_COMPLETED(vmb,slot,urb,error_count,0)) {

						SLOT_ASSERT(1 == slot->running);	// ISOCHRONOUS OUT
						u16 tler = READ_TLER(vmb,td);
						TLER_ASSERT(0 == (tler & 1));
						WRITE_TLER(vmb,td,0x00);	// clear out the TD
						slot->running = 0;	// ISOCHRONOUS OUT tler stopped
						slot->retired = 1;

						SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
						SLOT_ASSERT(0 == slot->next_urb_count);

						slot->jiffies = jiffies;
						slot->usb_addr = 0;
						slot->usb_endp = 0;

						free_vmb_slot(vmb,slot);

					} else {
						DBG(1, "%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) ISO OUT urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
						SLOT_ASSERT(slot->queue_size > 0);
						slot->jiffies = jiffies;

						SLOT_ASSERT(slot->u_r_b[queue_neck]);
						SLOT_ASSERT(usb_pipeout(slot->u_r_b[queue_neck]->pipe));
						SLOT_ASSERT(PIPE_ISOCHRONOUS == usb_pipetype(slot->u_r_b[queue_neck]->pipe));
						SLOT_ASSERT(VMB_URB_STATE_ISO_OUT == slot->u_r_b[queue_neck]->error_count);
					}

					UTL_DEBUG(4,vmb);
				} else {

					SLOT_ASSERT(0);	// TODO - INTERRUPT OUT ????

					URB_DBG(1,"%s(TD=%X idref=%d:%d queue=%d running=%d actioned=%d frame_tail=%04X urb=%p error_count=%d)\n",__FUNCTION__,td,slot->idref_this,slot->idref_done,slot->queue_size,slot->running,slot->actioned,frame_tail,urb,urb->error_count);
				}
				continue; // try next idref
			} else {

				break; // still possibly processing this one

			}
		}
	}
	//ASSERT(0);
}
//------------------------------------------------------------------------------------
//
//  plan_frame calculates the maximum number of input bits that might
//             be expected in the currently executing frame or in a
//             subsequent one - it is strictly a worst case estimate.
//
//------------------------------------------------------------------------------------
static int plan_frame(struct vmb*vmb)
{
	SW_ASSERT_LOCKED(vmb);

	int bulk_in_bits = 0;	// BULK IN
	int must_do_bits = 0;	// INT, CONTROL, ISO

	u8 n = 0;
	u8 slots = 16;

	while(slots-- > 0) {

		u8 td = n++;
		struct vmb_slot*slot = &vmb->slot[td];
		ASSERT(td == slot->td);
		ASSERT(vmb == slot->vmb);

		if (0 == slot->initialized) {
			continue;
		} else if (0 == slot->queue_size && slot->usb_endp > 0) {

			SLOT_ASSERT(1 == slot->allocated);
			SLOT_ASSERT(PIPE_BULK == slot->pipetype);

			if (slot->pipeout) {
				/*
				 * we had an error doing a BULK OUT
				 *    the tx fifo is in the process of being purged
				 */	
				SLOT_ASSERT(0 == slot->idref_done);
				SLOT_ASSERT(1 == slot->device_error);
				SLOT_ASSERT(1 == slot->actioned);
				SLOT_ASSERT(0 == slot->retry_flag);
				SLOT_ASSERT(0 == slot->retired);
				SLOT_ASSERT(0 == slot->end_seen);
				SLOT_ASSERT(1 == slot->running);

			} else if (slot->retired) {

				SLOT_ASSERT(0 == slot->running);
				SLOT_ASSERT(1 == slot->actioned);

			} else {

				SLOT_ASSERT(1 == slot->running);
				SLOT_ASSERT(1 == slot->actioned);

				u16 bytes = slot->maxusbpktlen;  // overspilling BULK IN
				u16 pkt_bits = (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);

				bulk_in_bits += slot->ec*pkt_bits;
				must_do_bits += pkt_bits;
			}
		} else if (0 < slot->queue_size && slot->usb_endp > 0) {

			SLOT_ASSERT(slot->allocated);
			SLOT_ASSERT(slot->queue_size < SLOT_QUEUE_SIZE);	// this is why we cannot fill up the queue completely
			if (PIPE_INTERRUPT == slot->pipetype) {

				if (slot->pipeout) {

					SLOT_ASSERT(0);	// I don't think this can happen

				} else {	// uses IN ONCE with slot->ec == 0
					SLOT_ASSERT(0 == slot->ec);
					u16 bytes = slot->maxusbpktlen;
					u16 pkt_bits = (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);

					must_do_bits += pkt_bits;
				}
			} else if (PIPE_ISOCHRONOUS == slot->pipetype) {
		
				if (slot->pipeout) {
				} else {

					//u16 urb_index = slot->next_urb_index;
					//u16 urb_count = slot->next_urb_count;

					//struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];
					//ASSERT(urb);
		
					u32 bytes = slot->maxusbpktlen;
					u16 pkt_bits = (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);

					must_do_bits += pkt_bits;
				}
			} else { // BULK IN and OUT

				SLOT_ASSERT(PIPE_BULK == slot->pipetype);

				if (slot->pipeout) {
				} else {

					u16 bytes = slot->maxusbpktlen;
					u16 pkt_bits = (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);

					bulk_in_bits += slot->ec*pkt_bits;
					must_do_bits += pkt_bits;
				}
			}
		} else if (0 < slot->queue_size && 0 == slot->usb_endp) {
	
			SLOT_ASSERT(PIPE_CONTROL == slot->pipetype);

			SLOT_ASSERT(slot->allocated);
			SLOT_ASSERT(slot->queue_size < SLOT_QUEUE_SIZE);	// this is why we cannot fill up the queue completely

			struct urb *urb = slot_to_head_urb(slot);
			SLOT_ASSERT(urb);

			if (VMB_URB_STATE_CONTROL_SETUP == urb->error_count) {
				SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
				SLOT_ASSERT(0 == slot->next_urb_count);
				SLOT_ASSERT((0 == slot->actioned && 0 == slot->idref_done) || (1 == slot->actioned && 1 == slot->idref_done));	// CONTROL OUT

			} else if (VMB_URB_STATE_CONTROL_IN == urb->error_count) {
/*-IN-USE-*/	SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
				SLOT_ASSERT(0 == slot->idref_done);

				u16 urb_count = slot->next_urb_count;
				u8 urb_shift = vmb_maxpacket_shift(slot->maxusbpktlen);
				u16 left = urb->transfer_buffer_length - (urb_count << urb_shift);
				u16 bytes = left < slot->maxusbpktlen ? left : slot->maxusbpktlen;
				u16 pkt_bits = (32+3+4) + (32+8*bytes+3+4) + (16+3+4);

				must_do_bits += pkt_bits;
			} else if (VMB_URB_STATE_CONTROL_OUT == urb->error_count) {
				if (slot->actioned) {
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT(1 == slot->idref_done);
				} else {
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT(0 == slot->idref_done);
				}

			} else if (VMB_URB_STATE_CONTROL_ACK_IN == urb->error_count) {
				SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));
				SLOT_ASSERT(0 == slot->next_urb_count);
				SLOT_ASSERT(0 == slot->idref_done);

				UTL_DEBUG(5,vmb);
				PTR_DEBUG(5,vmb);
				URB_DEBUG(5,urb);

			} else if (VMB_URB_STATE_CONTROL_ACK_OUT == urb->error_count) {
				SLOT_ASSERT(0 == slot->next_urb_count);
				SLOT_ASSERT(1 == slot->queue_size);
				SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
				if (1 == slot->actioned) {
					SLOT_ASSERT(1 == slot->idref_done);	// CONTROL ACK OUT actioned already
				} else {
					SLOT_ASSERT(0 == slot->idref_done);	// CONTROL ACK OUT not yet actioned
				}

			} else {
				UTL_ASSERT(0);
			}
		} else {
			SLOT_ASSERT(0);	// I don't think this case is possible because then the slot ushould be free
			SLOT_ASSERT(slot->allocated);
			SLOT_ASSERT(0 == slot->queue_size && 0 == slot->usb_endp);
			SLOT_ASSERT(slot->queue_size < SLOT_QUEUE_SIZE);	// this is why we cannot fill up the queue completely
		}
	}
	u16 input_bits = bulk_in_bits + must_do_bits;
	if (must_do_bits > 11960) {
		DBG(0,"must_do_bits = %d bulk_in_bits = %d input_bits = %d exceeds limit of 11960\n",must_do_bits,bulk_in_bits,input_bits);
		dump_slots(vmb);
		UTL_ASSERT(0); // frame overfilled
	} else {
	}
	return must_do_bits;
}
//-------------------------------------------------------------------------------
//
// IN's are the lowest priority when filling a frame
//
// the remaining space after SETUP's, ISOCHRONOUS's, INTERRUPT's and OUT's
//     is used for BULK IN's
//
//-------------------------------------------------------------------------------
static int fill_input(struct vmb*vmb,int budget_bits)
{
	SW_ASSERT_LOCKED(vmb);

	int count_filled = 0;

	int input_pkts[16];
	int input_size[16];
	int available[16];
	int scheduled[16];
	int control_in[16];
	int control_ack[16];
	int interrupt[16];
	int isochronous[16];
	int int_bits = 0;
	int iso_bits = 0;
	int bulk_in_bits = 0;
	int must_do_bits = 0;
	int control_bits = 0;

	{
		u8 n = 0;
		u8 slots = 16;
	
		while(slots-- > 0) {
	
			u8 td = n++;
			struct vmb_slot*slot = &vmb->slot[td];
			ASSERT(td == slot->td);
			ASSERT(vmb == slot->vmb);
	
			input_pkts[td] = 0;
			input_size[td] = 0;
			control_in[td] = 0;
			control_ack[td] = 0;
			available[td] = 0;
			scheduled[td] = 0;
			interrupt[td] = 0;
			isochronous[td] = 0;
	
			if (0 == slot->initialized) {
				continue;
			} else if (0 == slot->queue_size && slot->usb_endp > 0) {

				SLOT_ASSERT(slot->allocated);
				SLOT_ASSERT(PIPE_BULK == slot->pipetype);

				if (slot->pipeout) {
					/*
				 	* we had an error doing a BULK OUT
				 	*    the tx fifo is in the process of being purged
				 	*/	
					SLOT_ASSERT(0 == slot->idref_done);
					SLOT_ASSERT(1 == slot->device_error);
					SLOT_ASSERT(1 == slot->actioned);
					SLOT_ASSERT(0 == slot->retry_flag);
					SLOT_ASSERT(0 == slot->retired);
					SLOT_ASSERT(0 == slot->end_seen);
					SLOT_ASSERT(1 == slot->running);

				} else if (slot->retired) {

					SLOT_ASSERT(0 == slot->running);
					SLOT_ASSERT(1 == slot->actioned);

				} else {	

					SLOT_ASSERT(1 == slot->running);	// BULK IN
					SLOT_ASSERT(1 == slot->actioned);	// BULK IN
		
					u16 bytes = slot->maxusbpktlen;  // overspilling BULK IN
	
					available[td] = 8;	// maximum number of IN packets
					input_pkts[td] = 1+slot->ec;
					input_size[td] = (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);
					bulk_in_bits += slot->ec * input_size[td] ;
					must_do_bits += input_size[td] ;
				}	
			} else if (0 < slot->queue_size && slot->usb_endp > 0) {
	
				SLOT_ASSERT(slot->allocated);
				if (PIPE_INTERRUPT == slot->pipetype) {
	
					SLOT_ASSERT(1 == slot->queue_size);
	
					if (slot->pipeout) {
					} else {
	
						struct urb *urb = slot_to_head_urb(slot);
						SLOT_ASSERT(urb);
	
						if (VMB_URB_STATE_INT_WAIT == urb->error_count) {
	
							SLOT_ASSERT(1 == slot->actioned);	// INTERRUPT WAIT 
							SLOT_ASSERT(0 == slot->running);	// INTERRUPT WAIT
							SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
							SLOT_ASSERT(0 == slot->next_urb_count);
							SLOT_ASSERT(1 == slot->idref_done);
	
						} else if (VMB_URB_STATE_INTERRUPT == urb->error_count) {
	
							u16 bytes = slot->maxusbpktlen;
	
							interrupt[td] += 1;
							int_bits += (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);
	
						} else {
	
							URB_ASSERT(0);
						}
					}
				} else if (PIPE_ISOCHRONOUS == slot->pipetype) {
			
					if (slot->pipeout) {
					} else {
	
						//u16 urb_index = slot->next_urb_index;
						//u16 urb_count = slot->next_urb_count;
	
						//struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];
						//ASSERT(urb);
			
						u32 bytes = slot->maxusbpktlen;
	
						isochronous[td] += 1;	
						iso_bits += (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);
					}
				} else {	// PIPE_BULK == slot->pipetype

					SLOT_ASSERT(PIPE_BULK == slot->pipetype);

					if (slot->pipeout) {
					} else {
						DBG(1,"%s:%d(TD=%X actioned=%d running=%d)\n",__FUNCTION__,__LINE__,slot->td,slot->actioned,slot->running);
	
						u16 bytes = slot->maxusbpktlen;
	
						available[td] = 8;	// maximum number of IN packets
						input_size[td] = (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);

						if (slot->actioned) {
							SLOT_ASSERT(1 == slot->running);
							input_pkts[td] = 1+slot->ec ;
							bulk_in_bits += slot->ec * input_size[td] ;
							must_do_bits += input_size[td] ;
						} else {
							SLOT_ASSERT(0 == slot->running);
							input_pkts[td] = 0 ;
						}	
					}
				}
			} else if (0 < slot->queue_size && 0 == slot->usb_endp) {
		
				SLOT_ASSERT(slot->allocated);
				SLOT_ASSERT(PIPE_CONTROL == slot->pipetype);
	
				struct urb *urb = slot_to_head_urb(slot);
				SLOT_ASSERT(urb);
	
				if (VMB_URB_STATE_CONTROL_SETUP == urb->error_count) {
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT(1 == slot->idref_done);
	
				} else if (VMB_URB_STATE_CONTROL_IN == urb->error_count) {
/*-IN-USE-*/		SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->idref_done);
	
					u16 urb_count = slot->next_urb_count;
					u8 urb_shift = vmb_maxpacket_shift(slot->maxusbpktlen);
					u16 left = urb->transfer_buffer_length - (urb_count << urb_shift);
					u16 bytes = left < slot->maxusbpktlen ? left : slot->maxusbpktlen;
					control_bits += (32+3+4) + (32+8*bytes+3+4) + (16+3+4);
					control_in[td] += 1;
				} else if (VMB_URB_STATE_CONTROL_OUT == urb->error_count) {
					if (slot->actioned) {
						SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));
						SLOT_ASSERT(0 == slot->next_urb_count);
						SLOT_ASSERT(1 == slot->idref_done);
					} else {
						SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
						SLOT_ASSERT(0 < slot->next_urb_count);
						SLOT_ASSERT(0 == slot->idref_done);
					}
				} else if (VMB_URB_STATE_CONTROL_ACK_IN == urb->error_count) {
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT(0 == slot->idref_done);
					control_bits += (32+3+4) + (32+3+4) + (16+3+4);
					control_ack[td] += 1;
	
				} else if (VMB_URB_STATE_CONTROL_ACK_OUT == urb->error_count) {
					SLOT_ASSERT(1 == slot->queue_size);
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
					SLOT_ASSERT(1 == slot->idref_done);
				} else {
					UTL_ASSERT(0);
				}
			} else {
				SLOT_ASSERT(0);
				SLOT_ASSERT(slot->allocated);
				SLOT_ASSERT(0 == slot->queue_size && 0 == slot->usb_endp);
			}
		}
	}

	int available_bits = budget_bits - (iso_bits + int_bits + control_bits);

	if ((int_bits + iso_bits + must_do_bits + control_bits) <= budget_bits) {
		UTL_ASSERT(available_bits >= 0);
	} else {
		u8 n = 0;
		u8 slots = 16;
		dump_slots(vmb);
		DBG(0,"available_bits=%d budget_bits=%d int_bits=%d iso_bits=%d must_do_bits=%d bulk_in_bits=%d control_bits=%d\n"
					,available_bits,budget_bits,int_bits,iso_bits,must_do_bits,bulk_in_bits,control_bits);	
		while(slots-- > 0) {
		
			u8 td = n++;
			if((input_pkts[td]  > 0)
			|| (input_size[td]  > 0)
			|| (available[td]  > 0)
			|| (scheduled[td]  > 0)
			|| (control_in[td]  > 0)
			|| (control_ack[td]  > 0)
			|| (interrupt[td]  > 0)
			|| (isochronous[td]  > 0)
			) {
				DBG(0,"SLOT[%X] input_pkts=%d control_in=%d control_ack=%d interrupt=%d isochronous=%d input_size=%d available=%d scheduled=%d\n"
					,td,input_pkts[td],control_in[td],control_ack[td],interrupt[td],isochronous[td],input_size[td],available[td],scheduled[td]);

			} else {
			}
		}	
	}	
	UTL_ASSERT((int_bits + iso_bits + must_do_bits + control_bits) <= budget_bits); // frame overfilled

	scan_again:	// schedule the actual BULK IN's
	{
		int found = 0;
		u8 n = 0;
		u8 slots = 16;
	
		while(slots-- > 0) {
		
			u8 td = n++;
			if (available[td] > 0) {

				ASSERT(vmb->slot[td].allocated);
				ASSERT(vmb->slot[td].initialized);

				if (input_size[td] < available_bits) {

					available[td] -= 1;
					scheduled[td] += 1;
					available_bits -= input_size[td];

					found += 1;

				} else {
				}

			} else {
			}
		}
		if (found > 0) goto scan_again;
	}
	int changes = 0;
	{
		u8 n = 0;
		u8 slots = 16;
	
		while(slots-- > 0) {
		
			u8 td = n++;
			if (input_pkts[td] == scheduled[td]) {
			} else {

				changes += 1;
			}
		}
	}	

	int scheduled_bits = control_bits + iso_bits;

	// first pass reduce input packets per frame - ie EC
	{
		u8 n = 0;
		u8 slots = 16;
	
		while(slots-- > 0) {
		
			u8 td = n++;
			struct vmb_slot*slot = &vmb->slot[td];
			ASSERT(td == slot->td);
			ASSERT(vmb == slot->vmb);

			if (0 == input_pkts[td]) {

			} else if (0 == scheduled[td]) {	// start up the input flow

				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(0 == slot->device_error);
				SLOT_ASSERT(PIPE_BULK == slot->pipetype);
	
				u16 urb_index = slot->next_urb_index;

				struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];

				SLOT_ASSERT(urb);
				URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
				SLOT_ASSERT(0 == slot->pipeout);
				URB_ASSERT(VMB_URB_STATE_BULK_IN == urb->error_count);

				SLOT_ASSERT(0 == slot->running);	// BULK IN
				slot->ec = 0;
				u16 tler = READ_TLER(vmb,td);
				TLER_ASSERT(0 == (tler & 1));
				u16 new_tle = set_slot_tle(slot,0x6,0);	// IN SPD
				WRITE_TLER(vmb,td,new_tle);	// IN SPD
				slot->running = 1;	// start up the BULK IN

				slot->retired = 0;
				slot->actioned = 1;	// BULK IN

			} else if (scheduled[td] < input_pkts[td]) {	// reduce the input flow

				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(0 == slot->device_error);
				SLOT_ASSERT(1 == slot->running);
				SLOT_ASSERT(1 == slot->actioned);
				SLOT_ASSERT(0 == slot->retired);
				SLOT_ASSERT(PIPE_BULK == slot->pipetype);
				SLOT_ASSERT(0 == slot->pipeout);	// BULK IN

				/*
				 * might be overruning - so no urb is possible
				 *
				 */	

				u8 ec = scheduled[td] - 1 ;	

				SLOT_ASSERT(8 > ec);

				slot->ec = ec;
				u16 tler = READ_TLER(vmb,td);
				TLER_ASSERT(1 == (tler & 1));
				TLER_ASSERT(0x6 == ((tler>>3)&0x7));	// IN SPD
				u16 new_tle = set_slot_tle(slot,0x6,1);	// IN SPD
	
				WRITE_TLER(vmb,td,new_tle);	// start up the BULK IN as IN SPD
	
				HW_LOCK(vmb);
				vmb->sr_mask &= ~ 0x08;	// RX THRESH
				write_sr(vmb,vmb->sr_mask);
				HW_UNLOCK(vmb);
	
				scheduled_bits += scheduled[td] * input_size[td];

			} else {
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(0 == slot->device_error);
			}
		}
	}
	// second pass - do the required packet types
	{
		u8 n = 0;
		u8 slots = 16;
	
		while(slots-- > 0) {
		
			u8 td = n++;
			struct vmb_slot*slot = &vmb->slot[td];
			ASSERT(td == slot->td);
			ASSERT(vmb == slot->vmb);

			if (control_in[td]) {

				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(0 == slot->device_error);
				SLOT_ASSERT(PIPE_CONTROL == slot->pipetype);
				SLOT_ASSERT(0  < slot->queue_size);
				SLOT_ASSERT(0 == slot->idref_done);

/*-IN-USE-*/	SLOT_ASSERT((SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)) == (SLOT_QUEUE_MASK & slot->next_urb_index));
	
				struct urb *urb = slot_to_head_urb(slot);
				SLOT_ASSERT(urb);
				URB_ASSERT(VMB_URB_STATE_CONTROL_IN == urb->error_count);

				if (slot->actioned) {	// CONTROL IN
				} else if (changes) {
					SLOT_ASSERT(0 == slot->next_urb_count);
				} else {
					SLOT_ASSERT(0 == slot->next_urb_count);

					DBG(1,"%s(urb=%p slot=%X CONTROL_IN actioned=%d changes=%d)\n",__FUNCTION__,urb,td,slot->actioned,changes);

					SLOT_ASSERT(0 == slot->running);	// CONTROL IN
					slot->ec = 0;
					u16 tler = READ_TLER(vmb,td);
					TLER_ASSERT(0 == (tler & 1));

					HW_LOCK(vmb);
					vmb->sr_mask &= ~ 0x08;	// RX THRESH
					write_sr(vmb,vmb->sr_mask);
					HW_UNLOCK(vmb);

					u16 new_tle = set_slot_tle(slot,0x6,1);	// IN SPD
					WRITE_TLER(vmb,td,new_tle);	// IN SPD
					slot->running = 1;	// start up the CONTROL IN

					slot->retired = 0;
					slot->actioned = 1;	// CONTROL IN

					count_filled += 1; // diagnostics
				}
			} else if (control_ack[td]) {

				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(0 == slot->device_error);
				SLOT_ASSERT(PIPE_CONTROL == slot->pipetype);
				SLOT_ASSERT(0 <  slot->queue_size);
				SLOT_ASSERT(0 == slot->idref_done);
				SLOT_ASSERT(0 == slot->next_urb_count);
				SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));

				struct urb *urb = slot_to_head_urb(slot);
				SLOT_ASSERT(urb);
				URB_ASSERT(VMB_URB_STATE_CONTROL_ACK_IN == urb->error_count);

				if (slot->actioned) {	// CONTROL ACK IN
				} else if (changes) {
					SLOT_ASSERT(0 == slot->next_urb_count);
				} else {
					SLOT_ASSERT(0 == slot->next_urb_count);

					DBG(1,"%s(urb=%p slot=%X CONTROL_IN actioned=%d changes=%d)\n",__FUNCTION__,urb,td,slot->actioned,changes);

					SLOT_ASSERT(0 == slot->running);	// CONTROL ACK IN
					slot->ec = 0;
					u16 tler = READ_TLER(vmb,td);
					TLER_ASSERT(0 == (tler & 1));

					HW_LOCK(vmb);
					vmb->sr_mask &= ~ 0x08;	// RX THRESH
					write_sr(vmb,vmb->sr_mask);
					HW_UNLOCK(vmb);

					u16 new_tle = set_slot_tle(slot,0x6,1);	// IN SPD
					WRITE_TLER(vmb,td,new_tle);	// IN SPD
					slot->running = 1;	// start up the CONTROL ACK IN

					slot->retired = 0;
					slot->actioned = 1;	// CONTROL ACK IN

					count_filled += 1; // diagnostics
				}
			} else if (interrupt[td]) {	// INTERRUPT IN

				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(0 == slot->device_error);
				SLOT_ASSERT(PIPE_INTERRUPT == slot->pipetype);
				SLOT_ASSERT(1 == slot->queue_size);
	
				u16 urb_index = slot->next_urb_index;
				struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];

				SLOT_ASSERT(urb);
				URB_ASSERT(VMB_URB_STATE_INTERRUPT == urb->error_count);
				URB_ASSERT(PIPE_INTERRUPT == usb_pipetype(urb->pipe));

				SLOT_ASSERT(0 == slot->retired);
				SLOT_ASSERT(0 == slot->idref_pkts);

				if (slot->actioned) {
					SLOT_ASSERT(1 == slot->idref_done);
					SLOT_ASSERT(1 == slot->running);	// INTERRUPT IN

					struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & (slot->idref_this - slot->idref_done)];
					SLOT_ASSERT((SLOT_QUEUE_MASK & idref->queue_this) == (SLOT_QUEUE_MASK & urb_index));

				} else if (changes) {
				} else {

					SLOT_ASSERT(0 == slot->idref_done);

					SLOT_ASSERT(0 == slot->running);	// INTERRUPT IN
					slot->ec = 0;
					u16 tler = READ_TLER(vmb,td);
					TLER_ASSERT(0 == (tler & 1));
					u16 new_tle = set_slot_tle(slot,0x3,1);	// IN ONCE
				
					HW_LOCK(vmb);
					vmb->sr_mask &= ~ 0x08;	// RX THRESH
					write_sr(vmb,vmb->sr_mask);
					u16 start_fc = read_fc(vmb);    // the frame currently or recently finished executing
					HW_UNLOCK(vmb);
				
					WRITE_TLER(vmb,td,new_tle);	// IN ONCE
					slot->running = 1;	// start up the INTERRUPT IN
					slot->retired = 0;
					slot->actioned = 1;	// INTERRUPT IN

					struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & slot->idref_this++];
					slot->idref_done += 1;

					idref->frame_last = UTL_FRAME_MASK & (start_fc + 100+urb->interval);	// effectively the timeout
					idref->queue_this = urb_index;

					count_filled += 1; // diagnostics
				}
			} else if (isochronous[td]) {

				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(0 == slot->device_error);
				SLOT_ASSERT(PIPE_ISOCHRONOUS== slot->pipetype);
	
				u16 urb_index = slot->next_urb_index;

				struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];

				SLOT_ASSERT(urb);
				URB_ASSERT(VMB_URB_STATE_ISO_IN == urb->error_count);
				URB_ASSERT(PIPE_ISOCHRONOUS == usb_pipetype(urb->pipe));
				SLOT_ASSERT(0 == slot->pipeout);

				if (slot->actioned) {	// ISOCHRONOUS IN

				} else {

				}
			} else if (scheduled[td] == input_pkts[td]) {

				scheduled_bits += scheduled[td] * input_size[td];

			} else {
			}
		}
	}
	// third pass - do increments of EC
	{
		u8 n = 0;
		u8 slots = 16;
	
		while(slots-- > 0) {
		
			u8 td = n++;
			struct vmb_slot*slot = &vmb->slot[td];
			ASSERT(td == slot->td);
			ASSERT(vmb == slot->vmb);

			if (scheduled[td] == input_pkts[td]) {
				// no change OR it's not a BULK IN
			} else if (scheduled[td] < input_pkts[td]) {
				// already reduced flow
			} else if (0 == input_pkts[td]) {	// start up flow
				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(0 == slot->device_error);
				SLOT_ASSERT(PIPE_BULK == slot->pipetype);
				SLOT_ASSERT(0 == slot->pipeout);
				SLOT_ASSERT(0 == slot->retired);
				SLOT_ASSERT(0 == slot->running);
				SLOT_ASSERT(0 == slot->actioned);
				u16 urb_index = slot->next_urb_index;
				struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];

				SLOT_ASSERT(urb);
				URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
				URB_ASSERT(VMB_URB_STATE_BULK_IN == urb->error_count);
				SLOT_ASSERT(usb_pipein(urb->pipe));

				u8 ec = scheduled[td] - 1 ;	

				ASSERT(8 > ec);

				slot->ec = ec;
				u16 tler = READ_TLER(vmb,td);
				if ((tler & 1) || ((tler>>3)&0x7)) {
					DBG(0,"SLOT[%X] <-- %d:%d tler=%04X seems corrupt\n",slot->td,slot->usb_addr,slot->usb_endp,tler);
					utl_debug(__FUNCTION__,__LINE__,vmb);
				}
				u16 new_tle = set_slot_tle(slot,0x6,1);	// IN SPD
	
				HW_LOCK(vmb);
				vmb->sr_mask &= ~ 0x08;	// RX THRESH
				write_sr(vmb,vmb->sr_mask);
				HW_UNLOCK(vmb);
	
				WRITE_TLER(vmb,td,new_tle);	// IN SPD
	
				scheduled_bits += scheduled[td] * input_size[td];

				slot->running = 1;	// start up the BULK IN
				slot->actioned = 1;

				UTL_DEBUG(4,vmb);

			} else {	// increase flow
				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(0 == slot->device_error);
				SLOT_ASSERT(PIPE_BULK == slot->pipetype);
				SLOT_ASSERT(0 == slot->pipeout);
				SLOT_ASSERT(0 == slot->retired);
				SLOT_ASSERT(1 == slot->running);
				SLOT_ASSERT(1 == slot->actioned);

				/*
				 * might be overruning - so no urb is possible
				 *
				 */

				u8 ec = scheduled[td] - 1 ;	

				ASSERT(8 > ec);

				slot->ec = ec;
				u16 tler = READ_TLER(vmb,td);
				TLER_ASSERT(1 == (tler & 1));
				TLER_ASSERT(0x6 == ((tler>>3)&0x7));	// IN SPD
				u16 new_tle = set_slot_tle(slot,0x6,1);	// IN SPD
	
				HW_LOCK(vmb);
				vmb->sr_mask &= ~ 0x08;	// RX THRESH
				write_sr(vmb,vmb->sr_mask);
				HW_UNLOCK(vmb);
	
				WRITE_TLER(vmb,td,new_tle);	// IN SPD
	
				scheduled_bits += scheduled[td] * input_size[td];

				UTL_DEBUG(4,vmb);
			}
		}
	}
	if (0 < count_filled) UTL_DEBUG(1,vmb);
	return scheduled_bits;
}
//-------------------------------------------------------------------------------
//
// BULK OUT's are the second lowest priority when filling a frame,
//            only BULK IN's are lower priority
//
// the remaining space after SETUP's, ISOCHRONOUS's and INTERRUPT's
//     is used for BULK OUT's
//
//-------------------------------------------------------------------------------
static int fill_frame(struct vmb*vmb,u16 frame_head,u16 budget_bits)
{
	SW_ASSERT_LOCKED(vmb);

	struct vmb_frame*frame = &vmb->frame[UTL_FRAME_MASK & frame_head];

	int control_bits = 0;
	int bulk_in_bits = 0;
	int must_do_bits = 0;
	int output_bits = 0;
	int int_bits = 0;
	int iso_bits = 0;

	int output_size[16];
	int scheduled[16];
	int isochronous[16];
	int control_setup[16];
	int control_out[16];
	int control_ack[16];
	int output_bytes[16];
	int available[16];

	int fifo_bytes = 0;
	int max_pkts = 0;	// total packets in all queued BULK OUT urbs not yet transmitted
	int bulk_pkts = 0;	// for diagnostics
	int iso_pkts = 0;	// for diagnostics
	int ack_pkts = 0;	// for diagnostics
	int out_pkts = 0;	// for diagnostics

	DBG(4, "%s(vmb=%p frame_head=%04X budget_bits=%d)\n",__FUNCTION__,vmb,frame_head,budget_bits);
	{
		u8 n = 0;
		u8 slots = 16;
	
		while(slots-- > 0) {
	
			u8 td = n++;
			struct vmb_slot*slot = &vmb->slot[td];
			ASSERT(td == slot->td);
			ASSERT(vmb == slot->vmb);

			SLOT_ASSERT(slot->queue_size < SLOT_QUEUE_SIZE);	// this is why we cannot fill up the queue completely
	
			output_size[td] = 0;
			output_bytes[td] = 0;
			scheduled[td] = 0;
			available[td] = 0;
			isochronous[td] = 0;
			control_setup[td] = 0;
			control_out[td] = 0;
			control_ack[td] = 0;

			if (0 == slot->initialized) {
				continue;
			} else if (0 == slot->queue_size && slot->usb_endp > 0) {
	
				SLOT_ASSERT(slot->allocated);
				SLOT_ASSERT(PIPE_BULK == slot->pipetype);

				if (slot->pipeout) {
					/*
				 	* we had an error doing a BULK OUT
				 	*    the tx fifo is in the process of being purged
				 	*/	
					SLOT_ASSERT(0 == slot->idref_done);
					SLOT_ASSERT(1 == slot->device_error);
					SLOT_ASSERT(1 == slot->actioned);
					SLOT_ASSERT(0 == slot->retry_flag);
					SLOT_ASSERT(0 == slot->retired);
					SLOT_ASSERT(0 == slot->end_seen);
					SLOT_ASSERT(1 == slot->running);	// ???????????

				} else if (1 == slot->retired) {

					SLOT_ASSERT(0 == slot->running);
					SLOT_ASSERT(1 == slot->actioned);

				} else {	
	
					SLOT_ASSERT(1 == slot->running);
					SLOT_ASSERT(1 == slot->actioned);

					u16 bytes = slot->maxusbpktlen;  // overspilling BULK IN
					u16 pkt_bits = (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);

					bulk_in_bits += slot->ec * pkt_bits;
					must_do_bits += pkt_bits;
				}
			} else if (0 < slot->queue_size && slot->usb_endp > 0) {

				SLOT_ASSERT(slot->allocated);
				if (PIPE_INTERRUPT == slot->pipetype) {

					if (slot->pipeout) {

						SLOT_ASSERT(0);	// TODO

					} else {

						SLOT_ASSERT(0 == slot->ec);
						u16 bytes = slot->maxusbpktlen;
						u16 pkt_bits = (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);

						int_bits += pkt_bits;
					}
				} else if (PIPE_ISOCHRONOUS == slot->pipetype) {
			
					u16 urb_index = slot->next_urb_index;
					u16 urb_count = slot->next_urb_count;
	
					struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];
			
					ASSERT(urb);
					URB_ASSERT(vmb == urb->hcpriv);
					URB_ASSERT(slot->pipetype == usb_pipetype(urb->pipe));
					URB_ASSERT(slot->usb_addr == usb_pipedevice(urb->pipe));
					URB_ASSERT(slot->usb_endp == usb_pipeendpoint(urb->pipe));
					URB_ASSERT(slot->pipeout == usb_pipeout(urb->pipe));	
					URB_ASSERT(slot->maxusbpktlen == usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe)));
					DBG(4, "%s(urb=%p slot=%X vmb=%p dev=%p)\n",__FUNCTION__,urb,td,vmb,urb->dev);
	
					if (slot->pipeout) {

						u32 bytes = urb->iso_frame_desc[urb_count].length;
		
						isochronous[td] += 1;
						fifo_bytes += bytes + 4;
	
						iso_bits += (32+3+4) + (32+8*bytes*7/6+3+4) + (16+3+4);
	
					} else {

						u32 bytes = slot->maxusbpktlen;
	
						iso_bits += (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);
					}
				} else {	// PIPE_BULK == slot->pipetype

					SLOT_ASSERT(PIPE_BULK == slot->pipetype);

					if (slot->pipeout) {	// BULK OUT
		
						if (0 == slot->running) {	// BULK OUT
	
							DBG(4, "%s # %d(slot=%X vmb=%p) setting slot running doing OUT's to %d:%02d\n",__FUNCTION__,__LINE__,td,vmb,slot->usb_addr,slot->usb_endp);
							u16 new_tle = set_slot_tle(slot,0x0,1); // OUT
							WRITE_TLER(vmb,td,new_tle); // OUT
							slot->running = 1;	// start up the BULK OUT
	
						} else {
						}
	
						u16 urb_index = slot->next_urb_index;
						u16 urb_count = slot->next_urb_count;
						u8 idref_done = slot->idref_done;
		
						u16 bytes = slot->maxusbpktlen;
						output_bytes[td] = bytes;
	
						while((idref_done < SLOT_IDREF_SIZE) && (SLOT_QUEUE_MASK & (urb_index ^ slot->queue_next))) {
				
							struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];
				
							ASSERT(urb);
							URB_ASSERT(vmb == urb->hcpriv);
							URB_ASSERT(slot->pipetype == usb_pipetype(urb->pipe));
							URB_ASSERT(slot->usb_addr == usb_pipedevice(urb->pipe));
							URB_ASSERT(slot->usb_endp == usb_pipeendpoint(urb->pipe));
							URB_ASSERT(slot->pipeout == usb_pipeout(urb->pipe));	
							URB_ASSERT(slot->maxusbpktlen == usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe)));
							URB_ASSERT(urb_count <= urb->transfer_buffer_length);
				
							u8 urb_shift = vmb_maxpacket_shift(bytes);
							u16 remaining = urb->transfer_buffer_length - urb_count;
							u16 packets = remaining >> urb_shift;
		
							u16 pkt_bits = (32+3+4) + (32+8*bytes*7/6+3+4) + (16+3+4);	
							output_bits += packets*pkt_bits;
	
							available[td] += packets;
							output_size[td] = pkt_bits;
	
							max_pkts += packets;
	
							u16 fraction = remaining - (packets << urb_shift);

							if (fraction > 0) {
								max_pkts += 1;
								available[td] += 1;
								output_bits += (32+3+4) + (32+8*fraction*7/6+3+4) + (16+3+4);
							}

							DBG(4, "%s(BULK OUT urb=%p slot=%X done=%d len=%d off=%d packets=%d)\n",__FUNCTION__ ,urb,td,idref_done,urb->transfer_buffer_length,urb_count,packets);
	
							urb_count = 0;
							urb_index += 1;
							idref_done += 1;
						}
						DBG(4, "%s(BULK OUT slot=%X output_bytes=%d output_size=%d available=%d)\n",__FUNCTION__ ,td,output_bytes[td],output_size[td],available[td]);
					} else {	// BULK IN
		
						u16 bytes = slot->maxusbpktlen;
						u16 pkt_bits = (32+3+3) + (32+8*bytes*7/6+3+9) + (16+3+4);

						bulk_in_bits += slot->ec * pkt_bits;
						must_do_bits += pkt_bits;
					}
				}
			} else if (0 < slot->queue_size && 0 == slot->usb_endp) {
		
				SLOT_ASSERT(slot->allocated);
				SLOT_ASSERT(PIPE_CONTROL == slot->pipetype);

				struct urb *urb = slot_to_head_urb(slot);
				SLOT_ASSERT(urb);

				if (slot->actioned) {	// CONTROL ???
				} else if (VMB_URB_STATE_CONTROL_SETUP == urb->error_count) {
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT(0 == slot->idref_done);

					control_bits += (32+3+4) + (32+8*8+3+4) + (16+3+4);
					control_setup[td] += 1;
					fifo_bytes += 8 + 4;
					DBG(1, "%s(CONTROL_SETUP td=%X idref=%d queue=%d)\n",__FUNCTION__,td,slot->idref_done,slot->queue_size);
				} else if (VMB_URB_STATE_CONTROL_IN == urb->error_count) {
/*-IN-USE-*/		SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->idref_done);
				} else if (VMB_URB_STATE_CONTROL_OUT == urb->error_count) {
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT(0 == slot->idref_done);
					URB_ASSERT(urb->transfer_buffer_length <= slot->maxusbpktlen);
					u16 bytes = urb->transfer_buffer_length;
					control_bits += (32+3+4) + (32+8*bytes+3+4) + (16+3+4);
					control_out[td] += 1;
					fifo_bytes += bytes + 4;
				} else if (VMB_URB_STATE_CONTROL_ACK_IN == urb->error_count) {
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT(0 == slot->idref_done);
				} else if (VMB_URB_STATE_CONTROL_ACK_OUT == urb->error_count) {
					SLOT_ASSERT(1 == slot->queue_size);
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT(0 == slot->idref_done);

					control_bits += (32+3+4) + (32+3+4) + (16+3+4);
					control_ack[td] += 1;
					fifo_bytes += 0 + 4;
				} else if (VMB_URB_STATE_CONTROL_STALL == urb->error_count) {
				} else {
					UTL_ASSERT(0);
				}
			} else {
	
				SLOT_ASSERT(slot->allocated);
				SLOT_ASSERT(0 == slot->queue_size && 0 == slot->usb_endp);
			}
		}
	}

	UTL_ASSERT((iso_bits + int_bits + must_do_bits + control_bits) < 11960); // frame overfilled

	int available_bits = budget_bits - (iso_bits + int_bits + must_do_bits + control_bits);

	if (max_pkts > 0) {
		DBG(1, "%s(%d max_pkts) available bits=%d slots = %X %X %X %X %X %X %X %X %X %X %X %X %X %X %X %X\n",__FUNCTION__,max_pkts,available_bits,available[0],available[1],available[2],available[3],available[4],available[5],available[6],available[7],available[8],available[9],available[10],available[11],available[12],available[13],available[14],available[15]);
	}

	scan_again:	// schedule the actual BULK OUT's
	{
		int found = 0;
		u8 n = 0;
		u8 slots = 16;
	
		while(slots-- > 0) {

			u8 td = n++;
			if (available[td] > 0) {

				if (output_size[td] < available_bits) {

					u16 bytes = output_bytes[td];

					available[td] -= 1;
					scheduled[td] += 1;
					available_bits -= output_size[td];
					fifo_bytes += bytes + 4;

					found += 1;

				} else {
					DBG(4, "%s(TD=%d) output_size=%d available_bits=%d\n",__FUNCTION__,td,output_size[td],available_bits);	
				}

			} else {
			}
		}
		if (found > 0) goto scan_again;
	}
	if (max_pkts > 0) {
		DBG(1, "%s(%d max_pkts) scheduled %X %X %X %X %X %X %X %X %X %X %X %X %X %X %X %X in slots\n",__FUNCTION__,max_pkts,scheduled[0],scheduled[1],scheduled[2],scheduled[3],scheduled[4],scheduled[5],scheduled[6],scheduled[7],scheduled[8],scheduled[9],scheduled[10],scheduled[11],scheduled[12],scheduled[13],scheduled[14],scheduled[15]);
	}

	int scheduled_bits = 0;
	frame->txfifo_bytes = 0;

	if (fifo_bytes < fifo_space(vmb)) { // now do the write packets for this frame

		u8 n = 0;
		u8 slots = 16;
	
		while(slots-- > 0) {
		
			u8 td = n++;
			struct vmb_slot*slot = &vmb->slot[td];
			ASSERT(td == slot->td);
			ASSERT(vmb == slot->vmb);

			if (0 < control_setup[td]) {

				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(0 == slot->device_error);

				SLOT_ASSERT(0 == slot->running);	// about to do CONTROL SETUP

				u16 tler = READ_TLER(vmb,td);
				TLER_ASSERT(0 == (tler & 1));
				u16 new_tle = set_slot_tle(slot,0x2,1);	// SETUP
				WRITE_TLER(vmb,td,new_tle);	// SETUP
				slot->running = 1;	// start up the CONTROL SETUP

				SLOT_ASSERT(0 == slot->idref_done);
				SLOT_ASSERT(0 == slot->actioned);	// CONTROL SETUP
				SLOT_ASSERT(0 <  slot->queue_size);
				SLOT_ASSERT(0 == slot->next_urb_count);
				SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
				struct urb *urb = slot_to_head_urb(slot);
				SLOT_ASSERT(urb);

				HW_LOCK(vmb);
				u16 start_fc = read_fc(vmb);	// the frame currently or recently finished executing
				u16 txwp = read_txwp(vmb);
				u16 txrp = read_txrp(vmb);
				HW_UNLOCK(vmb);

				DBG(1, "%s(control_setup FC=%04X R=%04X W=%04X td=%X idref=%d queue=%d)\n",__FUNCTION__,start_fc,txrp,txwp,td,slot->idref_done,slot->queue_size);
				SLOT_ASSERT(0 == slot->idref_done);

				u16 urb_index = slot->next_urb_index;
	
				int last_pkt = 1;
				int flag_eot = 0x80 ;
				u16 packet_id = slot->write__id = (SLOT_IDREF_MASK & slot->idref_this) | (1 << SLOT_IDREF_SHIFT);

				struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & slot->idref_this++];
				slot->idref_done += 1;
				slot->idref_pkts = 0;
				idref->queue_this = urb_index;
				idref->frame_base = frame_head;
				idref->frame_last = frame_head;

				slot->actioned = 1;	// CONTROL SETUP
		
				urb->actual_length = 0;
		
				scheduled_bits += (32+3+4) + (32+8*8+3+4) + (16+3+4);
				frame->txfifo_bytes += 12;
				out_pkts += 1;

				u8*b = urb->setup_packet;
		
				write_fifo(vmb,(td << 4)|(last_pkt<<3)|(0<<2));	// data toggle is always 0 for SETUP packets
				write_fifo(vmb,0x08);
				write_fifo(vmb,b[0]);
				write_fifo(vmb,b[1]);
				write_fifo(vmb,b[2]);
				write_fifo(vmb,b[3]);
				write_fifo(vmb,b[4]);
				write_fifo(vmb,b[5]);
				write_fifo(vmb,b[6]);
				write_fifo(vmb,b[7]);
				write_fifo(vmb,packet_id & 0xFF);	// ID
				write_fifo(vmb,flag_eot | (packet_id>>8));	// 1==EOT
		
				write_txpc(vmb,0);

				slot->retired = 0;
		
				usb_settoggle(urb->dev, slot->usb_endp, 1, 1);
		
				UTL_DEBUG(4,vmb);
			} else if (0 < control_out[td]) {

				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(0 == slot->device_error);
				SLOT_ASSERT(0 < slot->usb_addr);
				SLOT_ASSERT(slot->idref_done < SLOT_IDREF_SIZE);
				SLOT_ASSERT(0 == slot->running);	// about to do CONTROL OUT

				u16 tler = READ_TLER(vmb,td);
				if (tler & 1) {
					utl_debug(__FUNCTION__,__LINE__,vmb);
				}

				u16 new_tle = set_slot_tle(slot,0x0,1);	// OUT
				WRITE_TLER(vmb,td,new_tle);	// OUT
				slot->running = 1;	// start up the CONTROL OUT

				SLOT_ASSERT(0 == slot->idref_done);
				SLOT_ASSERT(0 == slot->actioned);	// CONTROL OUT
				SLOT_ASSERT(0 == slot->retired);
				SLOT_ASSERT(0 == slot->next_urb_count);
				SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));

				slot->retired = 0;
		
				HW_LOCK(vmb);
				u16 start_fc = read_fc(vmb);	// the frame currently or recently finished executing
				u16 txwp = read_txwp(vmb);
				u16 txrp = read_txrp(vmb);
				HW_UNLOCK(vmb);

				DBG(1, "%s(control_out FC=%04X R=%04X W=%04X td=%X idref=%d queue=%d)\n",__FUNCTION__,start_fc,txrp,txwp,td,slot->idref_done,slot->queue_size);
				SLOT_ASSERT(0 == slot->idref_done);

				u8 urb_index = slot->next_urb_index;
				u16 urb_count = slot->next_urb_count;

				SLOT_ASSERT(SLOT_QUEUE_MASK & (urb_index ^ slot->queue_next));

				struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];
				SLOT_ASSERT(urb);
				SLOT_ASSERT(urb == slot_to_head_urb(slot));

				struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & slot->idref_this];

				if (0 == urb_count) idref->frame_base = frame_head;

				idref->frame_last = frame_head;
				idref->queue_this = urb_index;

				u16 packet_no = slot->idref_pkts++;

				URB_ASSERT(urb->actual_length == slot->next_urb_count);

				int transfer_length = urb->transfer_buffer_length - urb->actual_length;

				URB_ASSERT(transfer_length > 0);

				URB_ASSERT(slot->usb_addr == usb_pipedevice(urb->pipe));
				URB_ASSERT(slot->usb_endp == usb_pipeendpoint(urb->pipe));
				URB_ASSERT(slot->maxusbpktlen == usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe)));

				int last_pkt = 1;	// only send one CONTROL OUT packet at a time
				int flag_eot = 0x80 ;
				if (transfer_length > slot->maxusbpktlen) {

					u16 block_offset = slot->next_urb_count >> vmb_maxpacket_shift(slot->maxusbpktlen);
					SLOT_ASSERT(packet_no == block_offset);
					u16 packet_id = slot->write__id = (SLOT_IDREF_MASK & slot->idref_this) | (block_offset << SLOT_IDREF_SHIFT);
			
					DBG(5,"CONTROL OUT %d:%02d FULL PACKET TD=%X idref=%d:%d urb=%p transfer_buffer_length=%d packet_id=%04X packet_size=%d packet_no=%d toggle=%d\n",slot->usb_addr,slot->usb_endp,td,slot->idref_this,slot->idref_done,urb,urb->transfer_buffer_length,packet_id,slot->maxusbpktlen,packet_no,usb_gettoggle(urb->dev, slot->usb_endp,1));

					slot->next_urb_count += slot->maxusbpktlen;
			
					int len = slot->maxusbpktlen;
			
					scheduled_bits += (32+3+4) + (32+8*len*7/6+3+4) + (16+3+4);
					frame->txfifo_bytes += 4+len;
					out_pkts += 1;

					u8 h0 = len >> 8;
					u8 h1 = len;
				
					ASSERT(0 == h0);
			
					u8*b = urb->actual_length + (u8*)urb->transfer_buffer;
				
					if (slot->retry_flag) {
						write_fifo(vmb,(td << 4)|(last_pkt<<3)|(usb_gettoggle(urb->dev, slot->usb_endp,1)<<2)|0x02|(0x01 & h0));
						slot->retry_flag = 0;
					} else {
						write_fifo(vmb,(td << 4)|(last_pkt<<3)|(usb_gettoggle(urb->dev, slot->usb_endp,1)<<2)|(0x03 & h0));
					}
					write_fifo(vmb,h1);
			
					while(len-- > 0) {
						write_fifo(vmb,*b++);
					}
					write_fifo(vmb,packet_id & 0xFF);	// ID
					write_fifo(vmb,flag_eot | (packet_id>>8));	// 1==EOT
				
					urb->actual_length += slot->maxusbpktlen;
			
					write_txpc(vmb,0);
				
					usb_dotoggle(urb->dev, slot->usb_endp, 1);
			
					URB_ASSERT(urb->actual_length < urb->transfer_buffer_length);
	
				} else {
			
					URB_ASSERT(urb->transfer_buffer_length == (transfer_length+urb->actual_length));
			
					u16 block_offset = slot->next_urb_count >> vmb_maxpacket_shift(slot->maxusbpktlen);
					u32 packet_id = slot->write__id = (SLOT_IDREF_MASK & slot->idref_this) | (block_offset << SLOT_IDREF_SHIFT);
			
					DBG(5,"CONTROL OUT %d:%02d LAST PACKET TD=%X idref=%d:%d urb=%p transfer_buffer_length=%d packet_id=%04X packet_size=%d packet_no=%d toggle=%d\n",slot->usb_addr,slot->usb_endp,td,slot->idref_this,slot->idref_done,urb,urb->transfer_buffer_length,packet_id,transfer_length,packet_no,usb_gettoggle(urb->dev, slot->usb_endp,1));

					slot->idref_this += 1;
					slot->idref_done += 1;
					slot->idref_pkts = 0;
					slot->next_urb_count = 0;
					slot->next_urb_index += 1;
			
					int len = transfer_length;
					ASSERT(0 < len);
					u8 h0 = len >> 8;
					u8 h1 = len;
				
					scheduled_bits += (32+3+4) + (32+8*len*7/6+3+4) + (16+3+4);
					frame->txfifo_bytes += 4+len;
					out_pkts += 1;

					ASSERT(0 == h0);
			
					u8*b = urb->actual_length + (u8*)urb->transfer_buffer;
			
					if (slot->retry_flag) {
						write_fifo(vmb,(td << 4)|(last_pkt<<3)|(usb_gettoggle(urb->dev, slot->usb_endp,1)<<2)|0x02|(0x01 & h0));
						slot->retry_flag = 0;
					} else {
						write_fifo(vmb,(td << 4)|(last_pkt<<3)|(usb_gettoggle(urb->dev, slot->usb_endp,1)<<2)|(0x03 & h0));
					}
					write_fifo(vmb,h1);
				
					while(len-- > 0) {
						write_fifo(vmb,*b++);
					}
					write_fifo(vmb,packet_id & 0xFF);	// ID
					write_fifo(vmb,flag_eot | (packet_id>>8));	// 1==EOT
			
					urb->actual_length += transfer_length;
			
					slot->actioned = 1;	// CONTROL OUT

					write_txpc(vmb,0);
			
					usb_dotoggle(urb->dev, slot->usb_endp, 1);

					slot->toggle[SLOT_QUEUE_MASK & slot->next_urb_index] = usb_gettoggle(urb->dev, slot->usb_endp,1);
					URB_ASSERT(urb->actual_length == urb->transfer_buffer_length);
				}
				UTL_DEBUG(4,vmb);
			} else if (control_ack[td]) {

				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(0 == slot->device_error);
				if (0 == slot->running) {	// CONTROL ACK

					u16 new_tle = set_slot_tle(slot,0x0,1);	// OUT
					WRITE_TLER(vmb,td,new_tle);	// OUT
					slot->running = 1;	// start up the CONTROL ACK

					slot->retired = 0;

				} else {
				}

				if (1 == slot->running) {	// CONTROL ACK
		
					SLOT_ASSERT(0 == slot->idref_done);
					SLOT_ASSERT(0 == slot->actioned);	// CONTROL ACK
					SLOT_ASSERT(1 == slot->queue_size);
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
					struct urb *urb = slot_to_head_urb(slot);
					SLOT_ASSERT(urb);
	
					int last_pkt = 1;
					int flag_eot = 0x80 ;
					u16 packet_id = slot->write__id = (SLOT_IDREF_MASK & slot->idref_this) | (2 << SLOT_IDREF_SHIFT);
	
					struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & slot->idref_this++];
					slot->idref_done += 1;
					slot->idref_pkts = 0;
					idref->queue_this = SLOT_QUEUE_MASK & (slot->next_urb_index - 1);
					idref->frame_base = frame_head;
					idref->frame_last = frame_head;
	
					clear_max_plen(vmb,td);
			
					slot->actioned = 1;	// CONTROL ACK
		
					scheduled_bits += (32+3+4) + (32+3+4) + (16+3+4);
					frame->txfifo_bytes += 4;
					ack_pkts += 1;
		
					write_fifo(vmb,(td << 4)|(last_pkt<<3)|(usb_gettoggle(urb->dev,slot->usb_endp,1)<<2));
					write_fifo(vmb,0x00);
					write_fifo(vmb,packet_id & 0xFF);	// ID
					write_fifo(vmb,flag_eot | (packet_id>>8));	// 1==EOT
	
					write_txpc(vmb,0);
		
					usb_dotoggle(urb->dev,slot->usb_endp,1);
	
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
	
					UTL_DEBUG(4,vmb);

				} else {
				}
			} else if (isochronous[td]) {

				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(0 == slot->device_error);

				if (0 == slot->running) {	// ISOCHRONOUS OUT

					u16 tler = READ_TLER(vmb,td);
					TLER_ASSERT(0 == (tler & 1));
					u16 new_tle = set_slot_tle(slot,0x4,1); // OUT ISO

					WRITE_TLER(vmb,td,new_tle); // OUT ISO
					slot->running = 1;	// start up the ISOCHRONOUS OUT

				} else {
				}

				if (1 == slot->running) {	// ISOCHRONOUS OUT
		
					SLOT_ASSERT(SLOT_IDREF_SIZE > slot->idref_done);

					u8  urb_index = slot->next_urb_index;
					u16 urb_count = slot->next_urb_count;
		
					struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];
					SLOT_ASSERT(urb);
	
					struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & slot->idref_this];
	
					if (0 == urb_count) idref->frame_base = frame_head;
	
					idref->frame_last = frame_head;
					idref->queue_this = urb_index;
	
					int last_pkt = 1;
					int flag_eot = 0x00 ; // this is not the end of the urb
			
					u16 packet_id = slot->write__id = (SLOT_IDREF_MASK & slot->idref_this) | (1 << SLOT_IDREF_SHIFT);
	
					int len = urb->iso_frame_desc[urb_count].length;
					urb->actual_length += len;
					urb->iso_frame_desc[urb_count].actual_length = len;
	
					scheduled_bits += (32+3+4) + (32+8*len*7/6+3+4) + (16+3+4);
					frame->txfifo_bytes += 4+len;
					iso_pkts += 1;
		
					DBG(1,"ISO OUT %d:%02d FULL PACKET urb=%p transfer_buffer_length=%d packet_id=%04X packet_size=%d\n",slot->usb_addr,slot->usb_endp,urb,urb->transfer_buffer_length,packet_id,len);
	
					u8 h0 = len >> 8;
					u8 h1 = len;
				
					ASSERT(0 == h0);	// not true for ISO - just testing OUT's at the moment
			
					u8*b = urb->iso_frame_desc[urb_count].offset + (u8*)urb->transfer_buffer;
				
					write_fifo(vmb,(td << 4)|(last_pkt<<3)|(usb_gettoggle(urb->dev, slot->usb_endp,1)<<2)|(0x03 & h0));
	
					write_fifo(vmb,h1);
			
					while(len-- > 0) {
						write_fifo(vmb,*b++);
					}
					write_fifo(vmb,packet_id & 0xFF);	// ID
					write_fifo(vmb,flag_eot | (packet_id>>8));	// 0==EOT
					
					write_txpc(vmb,0);
				
					usb_dotoggle(urb->dev, slot->usb_endp, 1);
			
					if (++slot->next_urb_count == urb->number_of_packets) {
	
						slot->idref_this += 1;
						slot->idref_done += 1;
						slot->idref_pkts = 0;
						slot->next_urb_index += 1;
						slot->next_urb_count = 0;
					}
		
					UTL_DEBUG(4,vmb);

				} else {
				}
			} else while(scheduled[td]-- > 0) {	// BULK OUT

				SLOT_ASSERT(1 == slot->allocated);
				SLOT_ASSERT(1 == slot->initialized);
				SLOT_ASSERT(0 == slot->device_error);
				SLOT_ASSERT(0 < slot->usb_addr);
				SLOT_ASSERT(slot->idref_done < SLOT_IDREF_SIZE);

				if (0 == slot->running) {	// BULK OUT

					DBG(4, "%s # %d(slot=%X vmb=%p) setting slot running doing OUT's to %d:%02d\n",__FUNCTION__,__LINE__,td,vmb,slot->usb_addr,slot->usb_endp);
					u16 tler = READ_TLER(vmb,td);
					TLER_ASSERT(0 == (tler & 1));
					u16 new_tle = set_slot_tle(slot,0x0,1); // OUT
					WRITE_TLER(vmb,td,new_tle); // OUT
					slot->running = 1;	// start up the BULK OUT
				} else {
				}
	
				SLOT_ASSERT(0 == slot->actioned);	// BULK OUT

				u8 urb_index = slot->next_urb_index;
				u16 urb_count = slot->next_urb_count;

				SLOT_ASSERT(SLOT_QUEUE_MASK & (urb_index ^ slot->queue_next));

				struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];
				SLOT_ASSERT(urb);

				struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & slot->idref_this];

				if (0 == urb_count) idref->frame_base = frame_head;

				idref->frame_last = frame_head;
				idref->queue_this = urb_index;

				u16 packet_no = slot->idref_pkts++;

				URB_ASSERT(urb->actual_length == slot->next_urb_count);

				int transfer_length = urb->transfer_buffer_length - urb->actual_length;

				URB_ASSERT(transfer_length > 0);

				URB_ASSERT(slot->usb_addr == usb_pipedevice(urb->pipe));
				URB_ASSERT(slot->usb_endp == usb_pipeendpoint(urb->pipe));
				URB_ASSERT(slot->maxusbpktlen == usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe)));

				int last_pkt = (0 == scheduled[td]);
			
				if (transfer_length > slot->maxusbpktlen) {

					int flag_eot = 0x00 ; // this is not the end of the urb
			
					u16 block_offset = slot->next_urb_count >> vmb_maxpacket_shift(slot->maxusbpktlen);
					ASSERT(packet_no == block_offset);
					u16 packet_id = slot->write__id = (SLOT_IDREF_MASK & slot->idref_this) | (block_offset << SLOT_IDREF_SHIFT);
			
					SCHED_DBG(0,"BULK OUT %d:%02d FULL PACKET TD=%X idref=%d:%d L=%04X Q=%02X urb=%p len=%d packet_id=%04X packet_size=%d packet_no=%d toggle=%d\n",slot->usb_addr,slot->usb_endp,td,slot->idref_this,slot->idref_done,idref->frame_last,idref->queue_this,urb,urb->transfer_buffer_length,packet_id,slot->maxusbpktlen,packet_no,usb_gettoggle(urb->dev, slot->usb_endp,1));

					slot->next_urb_count += slot->maxusbpktlen;
			
					int len = slot->maxusbpktlen;
			
					scheduled_bits += (32+3+4) + (32+8*len*7/6+3+4) + (16+3+4);
					frame->txfifo_bytes += 4+len;
					bulk_pkts += 1;

					u8 h0 = len >> 8;
					u8 h1 = len;
				
					ASSERT(0 == h0);	// not true for ISO - just testing OUT's at the moment
			
					u8*b = urb->actual_length + (u8*)urb->transfer_buffer;
				
					if (slot->retry_flag) {
						write_fifo(vmb,(td << 4)|(last_pkt<<3)|(usb_gettoggle(urb->dev, slot->usb_endp,1)<<2)|0x02|(0x01 & h0));
						slot->retry_flag = 0;
					} else {
						write_fifo(vmb,(td << 4)|(last_pkt<<3)|(usb_gettoggle(urb->dev, slot->usb_endp,1)<<2)|(0x03 & h0));
					}
					write_fifo(vmb,h1);
			
					while(len-- > 0) {
						write_fifo(vmb,*b++);
					}
					write_fifo(vmb,packet_id & 0xFF);	// ID
					write_fifo(vmb,flag_eot | (packet_id>>8));	// 0==EOT
				
					urb->actual_length += slot->maxusbpktlen;
			
					write_txpc(vmb,0);
				
					usb_dotoggle(urb->dev, slot->usb_endp, 1);
			
					URB_ASSERT(urb->actual_length < urb->transfer_buffer_length);
	
				} else {
			
					URB_ASSERT(urb->transfer_buffer_length == (transfer_length+urb->actual_length));
			
					int flag_eot = 0x00 ; // do not set EOT
			
					u16 block_offset = slot->next_urb_count >> vmb_maxpacket_shift(slot->maxusbpktlen);
					u32 packet_id = slot->write__id = (SLOT_IDREF_MASK & slot->idref_this) | (block_offset << SLOT_IDREF_SHIFT);
			
					SCHED_DBG(0,"BULK OUT %d:%02d LAST PACKET TD=%X idref=%d:%d L=%04X Q=%02X urb=%p len=%d packet_id=%04X packet_size=%d packet_no=%d toggle=%d\n",slot->usb_addr,slot->usb_endp,td,slot->idref_this,slot->idref_done,idref->frame_last,idref->queue_this,urb,urb->transfer_buffer_length,packet_id,transfer_length,packet_no,usb_gettoggle(urb->dev, slot->usb_endp,1));

					slot->idref_this += 1;
					slot->idref_done += 1;
					slot->idref_pkts = 0;
					slot->next_urb_count = 0;
					slot->next_urb_index += 1;
			
					int len = transfer_length;
					u8 h0 = len >> 8;
					u8 h1 = len;
				
					scheduled_bits += (32+3+4) + (32+8*len*7/6+3+4) + (16+3+4);
					frame->txfifo_bytes += 4+len;
					bulk_pkts += 1;

					ASSERT(0 == h0);	// not true for ISO - just testing OUT's at the moment
			
					u8*b = urb->actual_length + (u8*)urb->transfer_buffer;
			
					if (slot->retry_flag) {
						write_fifo(vmb,(td << 4)|(last_pkt<<3)|(usb_gettoggle(urb->dev, slot->usb_endp,1)<<2)|0x02|(0x01 & h0));
						slot->retry_flag = 0;
					} else {
						write_fifo(vmb,(td << 4)|(last_pkt<<3)|(usb_gettoggle(urb->dev, slot->usb_endp,1)<<2)|(0x03 & h0));
					}
					write_fifo(vmb,h1);
				
					while(len-- > 0) {
						write_fifo(vmb,*b++);
					}
					write_fifo(vmb,packet_id & 0xFF);	// ID
					write_fifo(vmb,flag_eot | (packet_id>>8));	// 0==EOT
			
					urb->actual_length += transfer_length;
			
					write_txpc(vmb,0);
			
					usb_dotoggle(urb->dev, slot->usb_endp, 1);

					slot->toggle[SLOT_QUEUE_MASK & slot->next_urb_index] = usb_gettoggle(urb->dev, slot->usb_endp,1);
					URB_ASSERT(urb->actual_length == urb->transfer_buffer_length);
				}
			}
		}
		UTL_DEBUG(3,vmb);

		if (0 < scheduled_bits) {
			SCHED_DBG(0,"Frame=%04X bits=%5d bytes=%04X packets(O=%d A=%d I=%d B=%d)\n",frame_head,scheduled_bits,frame->txfifo_bytes,out_pkts,ack_pkts,iso_pkts,bulk_pkts);
		} else {
			SCHED_DBG(6,"Frame=%04X bits=%5d bytes=%04X packets(O=%d A=%d I=%d B=%d)\n",frame_head,scheduled_bits,frame->txfifo_bytes,out_pkts,ack_pkts,iso_pkts,bulk_pkts);
		}

		return scheduled_bits;
	} else { // not enough space for a whole frame
		ASSERT(0); // TODO
		return 0;
	}
}
static int available_bytes(struct vmb *vmb)
{
	SW_ASSERT_LOCKED(vmb);
	u8 n = 0;
	u8 slots = MAX_VMB_SLOTS;

	int bytes_waiting[MAX_VMB_SLOTS];
	int	bytes_out_waiting = 0;
	int	bytes_in_waiting = 0;

	while(slots-- > 0) {

		u8 td = n++;
		bytes_waiting[td] = 0;

		struct vmb_slot*slot = &vmb->slot[td];
		SLOT_ASSERT(vmb == slot->vmb);
		SLOT_ASSERT(td == slot->td);
		SLOT_ASSERT(slot->queue_size < SLOT_QUEUE_SIZE);	// this is why we cannot fill up the queue completely

		SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next) || slot->queue_size > 0);

		if (1 == slot->device_error) {
			SLOT_ASSERT(1 == slot->initialized);
			SLOT_ASSERT(1 == slot->allocated);
			continue;
		} else if (0 == slot->allocated) {
			SLOT_ASSERT(0 == slot->initialized);
			continue;
		} else {
			SLOT_ASSERT(1 == slot->initialized);

			u16 urb_index = slot->next_urb_index;
			u16 urb_count = slot->next_urb_count;

			if (0 == slot->queue_size && slot->usb_endp > 0) {

				SLOT_ASSERT(PIPE_BULK == slot->pipetype);
				if (slot->pipeout) {
					/*
					 *  BULK OUT's begin purged
					 */
					SLOT_ASSERT(1 == slot->device_error);
				} else {	
					/*
					 *  overspilling BULK IN's
					 */
					u32 overspill = (1+slot->ec)*slot->maxusbpktlen;
					bytes_waiting[td] += overspill;
					bytes_in_waiting += overspill;
				}
			} else if (0 < slot->queue_size && slot->usb_endp > 0) {
	
				SLOT_ASSERT(slot->allocated);
				while(SLOT_QUEUE_MASK & (urb_index ^ slot->queue_next)) {
		
					struct urb* urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];
					SLOT_ASSERT(urb);
		
					if (0 == urb_count) {
	
						u32 buffer = urb->transfer_buffer_length;
	
						bytes_waiting[td] += buffer;
	
						if (slot->pipeout) {
							bytes_out_waiting += buffer;
						} else {
							bytes_in_waiting += buffer;
						}
		
					} else if (PIPE_ISOCHRONOUS == slot->pipetype) {
		
						u32 buffer = urb->transfer_buffer_length - urb->iso_frame_desc[urb_count].offset;
	
						bytes_waiting[td] += buffer;
	
						if (slot->pipeout) {
							bytes_out_waiting += buffer;
						} else {
							bytes_in_waiting += buffer;
						}
		
					} else {
					
						u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
						u8 urb_shift = vmb_maxpacket_shift(maxusbpktlen);
		
						u32 buffer = urb->transfer_buffer_length - (urb_count << urb_shift);
	
						bytes_waiting[td] += buffer;
	
						if (slot->pipeout) {
							bytes_out_waiting += buffer;
						} else {
							bytes_in_waiting += buffer;
						}
					}
		
					urb_index += 1;
					urb_count = 0;
				}
	
			} else if (0 < slot->queue_size && 0 == slot->usb_endp) {
		
				SLOT_ASSERT(slot->allocated);
				SLOT_ASSERT(PIPE_CONTROL == slot->pipetype);
	
				struct urb *urb = slot_to_head_urb(slot);
				SLOT_ASSERT(urb);
	
				if (VMB_URB_STATE_CONTROL_SETUP == urb->error_count) {
	
					SLOT_ASSERT(0 == slot->next_urb_count);
	
					if (slot->actioned) {	// CONTROL SETUP
						if (slot->running) {	
							SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
						} else {
							SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));
						}
						SLOT_ASSERT(1 == slot->idref_done);
	
					} else {
	
						if (slot->running) {	
							SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
						} else if (slot->end_seen) {
							SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));
						} else {
							SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
						}
						SLOT_ASSERT(0 == slot->idref_done);
	
						bytes_out_waiting += 8;
					}	
					if (slot->pipeout) {
						bytes_in_waiting += urb->transfer_buffer_length;
					} else {
						bytes_out_waiting += urb->transfer_buffer_length;
					}
	
				} else if (VMB_URB_STATE_CONTROL_OUT == urb->error_count) {
	
					if (slot->actioned) {
	/*-IN-USE-*/		SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));
						SLOT_ASSERT(0 == slot->next_urb_count);
						SLOT_ASSERT(1 == slot->idref_done);
	
					} else {
	
	/*-IN-USE-*/		SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
						SLOT_ASSERT(0 < slot->next_urb_count);
						SLOT_ASSERT(0 == slot->idref_done);
	
						bytes_out_waiting += urb->transfer_buffer_length;
					}	
				} else if (VMB_URB_STATE_CONTROL_IN == urb->error_count) {
	
	/*-IN-USE-*/	SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->idref_done);
	
					if (slot->pipeout) {
	
						u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
						u8 urb_shift = vmb_maxpacket_shift(maxusbpktlen);
		
						u32 buffer = urb->transfer_buffer_length - (urb_count << urb_shift);
	
						bytes_in_waiting += buffer;
	
					} else {
					}
	
				} else if (VMB_URB_STATE_CONTROL_ACK_OUT == urb->error_count) {
					SLOT_ASSERT(1 == slot->queue_size);
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
					SLOT_ASSERT(0 == slot->next_urb_count);
					if (1 == slot->actioned) {
						SLOT_ASSERT(1 == slot->idref_done);	// CONTROL ACK OUT actioned already
					} else {
						SLOT_ASSERT(0 == slot->idref_done);	// CONTROL ACK OUT not yet actioned
					}
				} else if (VMB_URB_STATE_CONTROL_ACK_IN == urb->error_count) {
	
					SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));
					SLOT_ASSERT(0 == slot->next_urb_count);
					SLOT_ASSERT(0 == slot->idref_done);
					//SLOT_ASSERT((0 == slot->actioned && 0 == slot->idref_done) || (1 == slot->actioned && 1 == slot->idref_done));

					UTL_DEBUG(5,vmb);
					PTR_DEBUG(5,vmb);
					URB_DEBUG(5,urb);

					bytes_in_waiting += 0;
				} else {
	
					UTL_ASSERT(0);
				}
	
			} else {
	
				SLOT_ASSERT(0 == slot->queue_size && 0 == slot->usb_endp);
			}
		}
	}
	return bytes_out_waiting + bytes_in_waiting ;
}
//-----------------------------------------------------------------------------------------
//
//                               the vmb scheduler
//
// must be called AFTER the input fifo has been emptied
//
// allocate priority order = ISOCHRONOUS
//                           INTERRUPT
//                           CONTROL
//                           BULK - short
//                           BULK - long
//                           
//-----------------------------------------------------------------------------------------
static void vmb_scheduler(struct vmb *vmb)
{
	SW_ASSERT_LOCKED(vmb);

	UTL_DEBUG(1,vmb);

	vmb_check_rx_fifo(vmb);

	HW_LOCK(vmb);
	u16 start_fc = read_fc(vmb);	// the frame currently or recently finished executing
	u16 txwp = read_txwp(vmb);
	u16 txrp = read_txrp(vmb);
	//u8 txpc = read_txpc(vmb);	// might not be in step with txwp and txrp
	HW_UNLOCK(vmb);

//	ASSERT(0 < txpc && txwp != txrp || txwp == txrp);
//	ASSERT(0 == txpc && txwp == txrp || txwp != txrp);

	u16 txfifospace = txrp > txwp ? txwp + VMB_FIFO_SIZE - txrp : txwp - txrp;
	ASSERT(txfifospace == (VMB_FIFO_MASK & (txwp - txrp)));

	ASSERT(0 == (~VMB_FIFO_MASK & txfifospace));

	u16 frame_head = vmb->frame_head;
	u16 frame_tail = vmb->frame_tail;

	ASSERT(0 == (~UTL_FRAME_MASK & frame_head));
	ASSERT(0 == (~UTL_FRAME_MASK & frame_tail));

	u16 frame_size = UTL_FRAME_MASK & (frame_head - frame_tail);
	u16 frame_done = UTL_FRAME_MASK & (start_fc   - frame_tail);

	u8 n = 0;
	u8 slots = MAX_VMB_SLOTS;
	int error_found = 0;

	while(slots-- > 0) {

		u8 td = n++;
		struct vmb_slot*slot = &vmb->slot[td];
		ASSERT(vmb == slot->vmb);
		ASSERT(td == slot->td);
		SLOT_ASSERT(slot->queue_size < SLOT_QUEUE_SIZE);	// this is why we cannot fill up the queue completely

		/*
		 * slot->initialized now follows slot->allocated
		 * and because allocation and initialization is now
		 * performed under a single SW_LOCK to SW_UNLOCK
		 * call it makes  slot->initialized effectively
		 * redundant.
		 *
		 */
		if (1 == slot->device_error) {
			SLOT_ASSERT(1 == slot->initialized);
			SLOT_ASSERT(1 == slot->allocated);
			continue;
		} else if (0 == slot->allocated) {
			SLOT_ASSERT(0 == slot->initialized);
			continue;
		} else {
			SLOT_ASSERT(1 == slot->initialized);
			if (0 < slot->queue_size) {
			} else {
				continue;
			}
		}

		u16 urb_index = slot->next_urb_index;
		u16 urb_count = slot->next_urb_count;

		if ((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next) || slot->queue_size > 0) {
		} else {
			error_found += 1;
			SCHED_DBG(1,"%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) ****ERROR****\n",__FUNCTION__,td,urb_count,urb_index,slot->queue_next,slot->queue_size);
		}
	}
	if (error_found > 0) UTL_DEBUG(1,vmb);

	int bytes_waiting = available_bytes(vmb);
	int sched = 0 < bytes_waiting ? 0 : 2;

	SCHED_DBG(sched, "%s:0(start FC=%04X T=%04X H=%04X R=%04X W=%04X) size=%d done=%d bytes_waiting=%d\n",__FUNCTION__,start_fc,vmb->frame_tail,vmb->frame_head,txrp,txwp,frame_size,frame_done,bytes_waiting);


	if (frame_done > frame_size) {

		u16 fifo = vmb->frame[frame_head].fifo_offset;

		while (frame_done > frame_size) {
			frame_size += 1;
			frame_head = UTL_FRAME_MASK & (1+frame_head);
			vmb->frame[frame_head].fifo_offset = fifo;
		}
		SCHED_DBG(2+sched, "%s:1(frame FC=%04X T=%04X H=%04X-->%04X F=%04X) size=%d done=%d\n",__FUNCTION__,start_fc,frame_tail,vmb->frame_head,frame_head,fifo,frame_size,frame_done);

		vmb->frame_head = frame_head;

		ASSERT(frame_head == start_fc);
		ASSERT(frame_done == frame_size);

	} else {

		SCHED_DBG(2+sched, "%s:2(frame FC=%04X T=%04X H=%04X) size=%d done=%d \n",__FUNCTION__,start_fc,frame_tail,frame_head,frame_size,frame_done);
	}

	while(frame_done > 1) {

		struct vmb_frame*frame = &vmb->frame[frame_tail];

		u16 fifo = frame->fifo_offset;

		ASSERT(!(fifo == txwp && fifo  > txrp && txrp  > txwp));	//04
		ASSERT(!(fifo == txwp && fifo == txrp && txrp  > txwp));	//05
		ASSERT(!(fifo <  txwp && fifo  > txrp && txrp  > txwp));	//07
		ASSERT(!(fifo <  txwp && fifo == txrp && txrp  > txwp));	//08
		ASSERT(!(fifo  > txwp && fifo == txrp && txrp == txwp));	//11
		ASSERT(!(fifo  > txwp && fifo <  txrp && txrp == txwp));	//12
		ASSERT(!(fifo == txwp && fifo  > txrp && txrp == txwp));	//13
		ASSERT(!(fifo == txwp && fifo <  txrp && txrp == txwp));	//15
		ASSERT(!(fifo <  txwp && fifo  > txrp && txrp == txwp));	//16
		ASSERT(!(fifo <  txwp && fifo == txrp && txrp == txwp));	//17
		ASSERT(!(fifo  > txwp && fifo == txrp && txrp <  txwp));	//20
		ASSERT(!(fifo  > txwp && fifo <  txrp && txrp <  txwp));	//21
		ASSERT(!(fifo == txwp && fifo == txrp && txrp <  txwp));	//23
		ASSERT(!(fifo == txwp && fifo <  txrp && txrp <  txwp));	//24

		if (fifo > txwp) {

			if (fifo < txrp) {
				//03 fifo > txwp AA fifo < txrp AA txrp > txwp	done
				ASSERT(txrp > txwp);

				done_frame(vmb,frame_tail);
			} else if (txrp > txwp) {
				//01 fifo > txwp AA fifo > txrp AA txrp > txwp	wait
				//02 fifo > txwp AA fifo = txrp AA txrp > txwp	wait
				ASSERT(!(fifo < txrp));
				break;	// wait for UTLE to process this frame
			} else {
				//10 fifo > txwp AA fifo > txrp AA txrp = txwp	done
				//19 fifo > txwp AA fifo > txrp AA txrp < txwp	done
				ASSERT(fifo > txrp);

				done_frame(vmb,frame_tail);
			}
		} else if (fifo < txwp) {

			if (fifo > txrp) {
				//25 fifo < txwp AA fifo > txrp AA txrp < txwp	wait
				ASSERT(txrp < txwp);
				break;	// wait for UTLE to process this frame
			} else if (txrp > txwp) {
				//09 fifo < txwp AA fifo < txrp AA txrp > txwp	wait
				ASSERT(fifo < txrp);
				break;	// wait for UTLE to process this frame
			} else if (fifo < txrp) {
				//18 fifo < txwp AA fifo < txrp AA txrp = txwp	done
				//27 fifo < txwp AA fifo < txrp AA txrp < txwp	done
				ASSERT(txrp == txwp || txrp < txwp);

				done_frame(vmb,frame_tail);
			} else {
				//26 fifo < txwp AA fifo = txrp AA txrp < txwp	wait
				ASSERT(txrp < txwp);
				break;	// wait for UTLE to process this frame
			}
		} else {

			//06 fifo = txwp AA fifo < txrp AA txrp > txwp	done
			//22 fifo = txwp AA fifo > txrp AA txrp < txwp	done
			//14 fifo = txwp AA fifo = txrp AA txrp = txwp	done
			ASSERT((fifo < txrp && txrp > txwp) || (fifo > txrp && txrp < txwp) || (fifo == txrp && txrp == txwp));

			done_frame(vmb,frame_tail);
		}
		frame_tail = vmb->frame_tail = UTL_FRAME_MASK & (1+frame_tail);
		frame_size -= 1;
		frame_done -= 1;
	}
	ASSERT(1 == frame_done);
	ASSERT(frame_head == vmb->frame_head);
	ASSERT(frame_tail == vmb->frame_tail);
	ASSERT(0 == (~UTL_FRAME_MASK & frame_head));
	ASSERT(0 == (~UTL_FRAME_MASK & frame_tail));
	ASSERT(frame_size == (UTL_FRAME_MASK & (frame_head - frame_tail)));
	ASSERT(frame_done == (UTL_FRAME_MASK & (start_fc   - frame_tail)));

	SCHED_DBG(2+sched, "%s:3(frame FC=%04X T=%04X H=%04X) size=%d done=%d \n",__FUNCTION__,start_fc,vmb->frame_tail,vmb->frame_head,frame_size,frame_done);

	u16 must_do_input_bits = plan_frame(vmb);

	u16 budget_bits = 11960 - must_do_input_bits;

	u16 ahead_size = UTL_FRAME_MASK & (frame_head - start_fc);
#define FRAME_QUEUE_LIMIT 8

	ASSERT(FRAME_QUEUE_LIMIT < UTL_FRAME_SIZE);

	SCHED_DBG(2+sched, "%s:4(frame FC=%04X T=%04X H=%04X) done=%d ahead=%d requested=%d budget=%d\n",__FUNCTION__,start_fc,vmb->frame_tail,vmb->frame_head,frame_done,ahead_size,must_do_input_bits,budget_bits);

	while(ahead_size < FRAME_QUEUE_LIMIT) {

		int scheduled_bits = fill_frame(vmb,frame_head,budget_bits);

		if (scheduled_bits > 0) {

			vmb->frame[frame_head].scheduled_bits = scheduled_bits;

			frame_head = vmb->frame_head = UTL_FRAME_MASK & (1+frame_head);

			vmb->frame[frame_head].fifo_offset = txwp = read_txwp(vmb);

			ahead_size += 1;
			frame_size += 1;

		} else {

			break;
		}
	}
	int scheduled_bits = 0;

	u16 frame_plan = frame_tail;
	while(frame_size-- > 0) {
		struct vmb_frame*frame = &vmb->frame[UTL_FRAME_MASK & frame_plan++];

		frame->requested_bits = must_do_input_bits;

		if (scheduled_bits < frame->scheduled_bits) scheduled_bits = frame->scheduled_bits;
	}
	fill_input(vmb,11960 - scheduled_bits);

	if (1 < pc_debug) {
		HW_LOCK(vmb);
		vmb->sr_mask |= 0x04;	// mask TX THRESH to slow things down for diagnostics
		write_sr(vmb,vmb->sr_mask);
		HW_UNLOCK(vmb);
	} else {
		u16 txit = 0;	
		HW_LOCK(vmb);
		u16 txwp = read_txwp(vmb);
		u16 txrp = read_txrp(vmb);
		u16 fifo_bytes = (txwp - txrp) & VMB_FIFO_MASK;
		if (fifo_bytes > 0) {
			vmb->sr_mask &= ~ 0x04;	// allow TX THRESH to get better performance
			write_sr(vmb,vmb->sr_mask);
			write_txit(vmb,txit);
			HW_UNLOCK(vmb);
			SCHED_DBG(0, "%s(txL=%04X TX THRESH --> %04X)\n",__FUNCTION__,fifo_bytes,txit);
		} else {
			HW_UNLOCK(vmb);
		}	
	}
	PTR_DEBUG(0,vmb);
}
//------------------------------
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_scheduler_event(void*data)	// workqueue API triggered from slot_enqueue_new() and slot_enqueue_add()
{
	struct vmb *vmb = data;
#else
static void vmb_scheduler_event(struct work_struct*work)	// workqueue API triggered from slot_enqueue_new() and slot_enqueue_add()
{
	struct vmb *vmb = scheduler_to_vmb(work);
#endif
#if DEBUGGING_COMPILED_IN
	if (vmb_is_still_valid(vmb)) {
	} else {
		ERR("%s(vmb=%p not valid in scheduler kicked from line %d)\n",__FUNCTION__,vmb,vmb->last_scheduled_from_line);
		ASSERT(0);
	}
#endif
	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	DBG(3, "API:%s(starting)\n",__FUNCTION__);

	vmb_scheduler(vmb);

	DBG(3, "API:%s(completed)\n",__FUNCTION__);

	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);

	vmb_giveback_urbs(vmb);
}
//-----------------------------------------------------------------------------------------
static void urb_overspill(struct vmb *vmb,struct vmb_slot*slot,struct urb *urb)
{
	ASSERT(vmb);
	SW_ASSERT_LOCKED(vmb);

	URB_ASSERT(0 == urb->actual_length);

	ASSERT(vmb == slot->vmb);

	SLOT_ASSERT(1 == slot->allocated);
	SLOT_ASSERT(1 == slot->initialized);
	SLOT_ASSERT(0 == slot->device_error);	// an error should have cleared out the slot
	SLOT_ASSERT(slot->queue_size > 0);
	SLOT_ASSERT(urb == slot_to_head_urb(slot));

	if ((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next) || slot->queue_size > 0) {
	} else {
		DBG(1,"%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) ****ERROR****\n",__FUNCTION__,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size);
	}

	SLOT_ASSERT(0 == slot->next_urb_count);

	u8 usb_addr = usb_pipedevice(urb->pipe);
	u8 usb_endp = usb_pipeendpoint(urb->pipe);

	ASSERT(usb_addr > 0);
	ASSERT(usb_endp > 0);

	SLOT_ASSERT(usb_addr == slot->usb_addr);
	SLOT_ASSERT(usb_endp == slot->usb_endp);

	struct vmb_udev*udev = &vmb->udev[usb_addr];

	UTL_ASSERT(udev);

	DBG(1, "%s(TD=%X urb=%p extra=%d buffer=%d)\n",__FUNCTION__,slot->td,urb,slot->extra,urb->transfer_buffer_length);
	UTL_DEBUG(1,vmb);

	URB_ASSERT(VMB_URB_STATE_BULK_OVERSPILL == urb->error_count);

	u16 spill = slot->extra;
	int space = urb->transfer_buffer_length;	
	u8*tb = (u8*)urb->transfer_buffer;
	u8*os = slot->overspill;

store_data:

	while(spill > 0 && space > 0) {
		space -= 1;	
		spill -= 1;
		*tb++ = *os++ ;
		urb->actual_length += 1;
	}

	slot->extra = 0;

	if (spill > 0) {

		URB_ASSERT(urb->transfer_buffer_length == urb->actual_length);
		UTL_ASSERT(0 == space);

		u8*eb = slot->overspill;

		while(spill-- > 0) {
   	        *eb++ = *os++ ;
			slot->extra += 1;
		}

		if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {
			/*
			 * Leave the slot active to mop up the data
			 *
			 * It might also still be running
			 *
			 */	
			slot->jiffies = jiffies;
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
	
			URB_DBG(0, "%s(TD=%X usb_hcd_giveback_urb=%p) BULK IN actual_length=%d overspill=%d status=%d\n",__FUNCTION__,slot->td,urb,urb->actual_length,slot->extra,urb->status);
		} else {
			DBG(1, "%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) BULK IN urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
			SLOT_ASSERT(slot->queue_size > 0);
			slot->jiffies = jiffies;
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
	
			URB_DBG(0, "%s(TD=%X usb_hcd_giveback_urb=%p) BULK IN actual_length=%d overspill=%d status=%d\n",__FUNCTION__,slot->td,urb,urb->actual_length,slot->extra,urb->status);
	
			urb = slot_to_head_urb(slot);
	
			SLOT_ASSERT(urb);
			URB_ASSERT(VMB_URB_STATE_BULK_IN == urb->error_count);
	
			URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
			URB_ASSERT(usb_pipein(urb->pipe));
	
			slot->jiffies = jiffies;
			slot->line_number = __LINE__;

			URB_ASSERT(0 == urb->actual_length);
			tb = urb->transfer_buffer;
			space = urb->transfer_buffer_length;
			goto store_data;
		}
	
		SLOT_ASSERT(usb_addr == slot->usb_addr);
		SLOT_ASSERT(usb_endp == slot->usb_endp);

		UTL_DEBUG(4,vmb);

	} else if (urb->transfer_buffer_length == urb->actual_length) {

		UTL_ASSERT(0 == space);
		SLOT_ASSERT(0 == spill);
		SLOT_ASSERT(slot->queue_size > 0);
	
		if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {
			slot->jiffies = jiffies;
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
		

			if (slot->retired) {

				SLOT_ASSERT(0 == slot->running);	// overspilling BULK IN

				u16 tler = READ_TLER(vmb,slot->td);
				ASSERT(0 == (0x0001 & tler));

				WRITE_TLER(vmb,slot->td,0x00);        // clear out the TD

				slot->usb_addr = 0;
				slot->usb_endp = 0;

				free_vmb_slot(vmb,slot);

			} else {

				SLOT_ASSERT(1 == slot->running);	// overspilling BULK IN
				SLOT_ASSERT(usb_addr == slot->usb_addr);
				SLOT_ASSERT(usb_endp == slot->usb_endp);
			}
		} else {
			DBG(1, "%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) BULK IN urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
			SLOT_ASSERT(slot->queue_size > 0);
			slot->jiffies = jiffies;
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
		
			urb = slot_to_head_urb(slot);

			SLOT_ASSERT(urb);
			SLOT_ASSERT(usb_addr == slot->usb_addr);
			SLOT_ASSERT(usb_endp == slot->usb_endp);
			URB_ASSERT(VMB_URB_STATE_BULK_IN == urb->error_count);
			URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
			URB_ASSERT(usb_pipein(urb->pipe));

			slot->jiffies = jiffies;

			SLOT_DEBUG(0,slot); // Do we need to restart this BULK IN ????
		}

		UTL_DEBUG(1,vmb);
	} else if (slot->retired) {

		u16 tler = READ_TLER(vmb,slot->td);
		SLOT_ASSERT(0 == spill);
		TLER_ASSERT(0 == (0x0001 & tler));
		SLOT_ASSERT(0 == slot->running);	// overspilling BULK IN retired

		if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
			slot->usb_addr = 0;
			slot->usb_endp = 0;
			slot->jiffies = jiffies;

			free_vmb_slot(vmb,slot);
		} else {
			DBG(1, "%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) BULK IN urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
			SLOT_ASSERT(slot->queue_size > 0);
			slot->jiffies = jiffies;
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
	
			urb = slot_to_head_urb(slot);
			SLOT_ASSERT(urb);
			URB_ASSERT(VMB_URB_STATE_BULK_IN == urb->error_count);
			SLOT_DEBUG(0,slot); // Do we need to restart this BULK IN ????
		}
	
		UTL_DEBUG(4,vmb);

		SLOT_ASSERT(0 == slot->queue_size);	// TO DO
	} else {

		SLOT_ASSERT(1 == slot->running);
		urb->error_count = VMB_URB_STATE_BULK_IN;
		slot->next_urb_count = urb->actual_length;// ??????????????????????????????????

		DBG(1,"%s(transfer_length=%d actual_length=%d)\n",__FUNCTION__,urb->transfer_buffer_length,urb->actual_length);
		VMB_DEBUG(1,vmb);
	}
	return;
}
static void td_stalled(struct vmb *vmb,unsigned td,struct urb *urb)
{
	ASSERT(vmb);
	SW_ASSERT_LOCKED(vmb);
	ASSERT(td < 16);
	struct vmb_slot *slot = &vmb->slot[td];
	ASSERT(td == slot->td);
	ASSERT(vmb == slot->vmb);

	SLOT_ASSERT(1 == slot->allocated);
	SLOT_ASSERT(1 == slot->initialized);
	SLOT_ASSERT(0 == slot->device_error);
	SLOT_ASSERT(urb == slot_to_head_urb(slot));

	if (VMB_URB_STATE_CONTROL_IN == urb->error_count) {

		DBG(4, "%s(urb=%p) error_count=%d\n",__FUNCTION__,urb,urb->error_count);

		SLOT_ASSERT(1 == slot->actioned);	// CONTROL IN
		URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));

		urb->error_count = VMB_URB_STATE_CONTROL_STALL;

		SLOT_ASSERT(1 == slot->running);	// CONTROL IN
		u16 tler = READ_TLER(vmb,td);
		TLER_ASSERT(1 == (tler & 1));
		tler &= ~ 0x0001;

		HW_LOCK(vmb);
		vmb->sr_mask &= ~ 0x08;	// RX THRESH
		write_sr(vmb,vmb->sr_mask);
		HW_UNLOCK(vmb);

		WRITE_TLER(vmb,td,tler);	// clear ACTIVE bit
		slot->running = 0;	// CONTROL IN stalled

		UTL_DEBUG(1,vmb);

		slot->line_number = __LINE__;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
		queue_delayed_work(workqueue,&slot->status_work,2);	// timer for clearing STALL'ed CONTROL_IN
#else
		queue_delayed_work(workqueue,&slot->delayed_work,2);	// timer for clearing STALL'ed CONTROL_IN
#endif

	} else if (VMB_URB_STATE_BULK_IN == urb->error_count) {

		DBG(4, "%s(urb=%p) error_count=%d\n",__FUNCTION__,urb,urb->error_count);

		URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
		URB_ASSERT(0 < usb_pipedevice(urb->pipe));

		urb->error_count = VMB_URB_STATE_BULK_IN_STALL;

		SLOT_ASSERT(1 == slot->running);	// BULK IN
		u16 tler = READ_TLER(vmb,td);
		TLER_ASSERT(1 == (tler & 1));
		tler &= ~ 0x0001;

		HW_LOCK(vmb);
		vmb->sr_mask &= ~ 0x08;	// RX THRESH
		write_sr(vmb,vmb->sr_mask);
		HW_UNLOCK(vmb);

		WRITE_TLER(vmb,td,tler);	// clear ACTIVE bit
		slot->running = 0;	// BULK IN stalled

		UTL_DEBUG(1,vmb);

		slot->line_number = __LINE__;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
		queue_delayed_work(workqueue,&slot->status_work,2);	// VMB_URB_STATE_BULK_IN_STALL
#else
		queue_delayed_work(workqueue,&slot->delayed_work,2);	// VMB_URB_STATE_BULK_IN_STALL
#endif

	} else {

		URB_DBG(0, "%s(urb=%p) error_count=%d\n",__FUNCTION__,urb,urb->error_count);

		UTL_ASSERT(0);	// FIX ME
	}
}
static void td_failure(struct vmb *vmb,unsigned td,struct urb *urb)
{
	ASSERT(vmb);
	SW_ASSERT_LOCKED(vmb);
	ASSERT(td < 16);
	struct vmb_slot *slot = &vmb->slot[td];
	ASSERT(td == slot->td);
	ASSERT(vmb == slot->vmb);

	SLOT_ASSERT(1 == slot->allocated);
	SLOT_ASSERT(1 == slot->initialized);
	SLOT_ASSERT(urb == slot_to_head_urb(slot));

	if (1 == slot->device_error) {
	} else if (VMB_URB_STATE_INTERRUPT == urb->error_count) {

		DBG(4, "%s(urb=%p) error_count=%d\n",__FUNCTION__,urb,urb->error_count);

		SLOT_ASSERT(1 == slot->actioned);	// INTERRUPT IN
		URB_ASSERT(PIPE_INTERRUPT == usb_pipetype(urb->pipe));

		urb->error_count = VMB_URB_STATE_INT_FAILED;

		SLOT_ASSERT(1 == slot->running);	// INTERRUPT IN
		u16 tler = READ_TLER(vmb,td);
		TLER_ASSERT(1 == (tler & 1));
		tler &= ~ 0x0001;

		HW_LOCK(vmb);
		vmb->sr_mask &= ~ 0x08;	// RX THRESH
		write_sr(vmb,vmb->sr_mask);
		HW_UNLOCK(vmb);

		WRITE_TLER(vmb,td,tler);	// clear ACTIVE bit
		slot->running = 0;	// INTERRUPT IN FAILED

		UTL_DEBUG(1,vmb);

		slot->line_number = __LINE__;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
		queue_delayed_work(workqueue,&slot->status_work,1);	// timer for clearing FAILED INTERRUPT IN
#else
		queue_delayed_work(workqueue,&slot->delayed_work,1);	// timer for clearing FAILED INTERRUPT IN
#endif

	} else if (VMB_URB_STATE_BULK_IN == urb->error_count) {

		DBG(4, "%s(urb=%p) error_count=%d\n",__FUNCTION__,urb,urb->error_count);

		SLOT_ASSERT(1 == slot->actioned);	// BULK IN
		URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));

		urb->error_count = VMB_URB_STATE_BULK_IN_FAILED;

		SLOT_ASSERT(1 == slot->running);	// BULK IN
		u16 tler = READ_TLER(vmb,td);
		TLER_ASSERT(1 == (tler & 1));
		tler &= ~ 0x0001;

		HW_LOCK(vmb);
		vmb->sr_mask &= ~ 0x08;	// RX THRESH
		write_sr(vmb,vmb->sr_mask);
		HW_UNLOCK(vmb);

		WRITE_TLER(vmb,td,tler);	// clear ACTIVE bit
		slot->running = 0;	// BULK IN FAILED

		UTL_DEBUG(1,vmb);

		slot->line_number = __LINE__;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
		queue_delayed_work(workqueue,&slot->status_work,1);	// timer for clearing FAILED BULK IN
#else
		queue_delayed_work(workqueue,&slot->delayed_work,1);	// timer for clearing FAILED BULK IN
#endif
	} else if (VMB_URB_STATE_CONTROL_IN == urb->error_count) {

		SLOT_ASSERT(1 == slot->running);
		SLOT_ASSERT(1 == slot->actioned);
		SLOT_ASSERT(0 == slot->retired);
		URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));

		urb->error_count = VMB_URB_STATE_CONTROL_IN_FAILED;

		u16 tler = READ_TLER(vmb,td);
		TLER_ASSERT(1 == (tler & 1));
		tler &= ~ 0x0001;

		WRITE_TLER(vmb,td,tler);	// clear ACTIVE bit
		slot->running = 0;
		slot->device_error = 1;	

		UTL_DEBUG(1,vmb);

		slot->line_number = __LINE__;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
		queue_delayed_work(workqueue,&slot->status_work,1);	// timer for clearing FAILED BULK IN
#else
		queue_delayed_work(workqueue,&slot->delayed_work,1);	// timer for clearing FAILED BULK IN
#endif
	} else {

		UTL_ASSERT(0);	// FIX ME
		URB_DBG(0, "%s(urb=%p) error_count=%d\n",__FUNCTION__,urb,urb->error_count);
	}
}
static void rx_packet_contined(struct vmb *vmb,unsigned td)	// IN SPD packet
{
	ASSERT(vmb);
	SW_ASSERT_LOCKED(vmb);
	ASSERT(td < 16);
	ASSERT(td == input_packet.utle);

	struct vmb_slot*slot = &vmb->slot[td];
	ASSERT(td == slot->td);
	ASSERT(vmb == slot->vmb);

	SLOT_ASSERT(1 == slot->allocated);
	SLOT_ASSERT(1 == slot->initialized);
	SLOT_ASSERT(0 == slot->device_error);

	SLOT_ASSERT(PIPE_BULK == slot->pipetype);
	SLOT_ASSERT(0 == slot->pipeout);

	slot->recieved += 1;

	u16 len = input_packet.length;

	u16 spill = sizeof(slot->overspill) - slot->extra;

	u8*eb = slot->extra + slot->overspill;
	u8*ip = input_packet.data;

	while(len > 0 && spill > 0) {
		len -= 1;
		spill -= 1;
		*eb++ = *ip++ ;
		slot->extra += 1;
	}
	UTL_ASSERT(0 == len);	// the overspill buffer is TOO SMALL

	u16 tler = READ_TLER(vmb,td);

	u8 usb_addr = (tler >> 6) & 0x0007;
	u8 usb_endp = (tler >> 9) & 0x000F;

	SLOT_ASSERT(usb_addr == slot->usb_addr);
	SLOT_ASSERT(usb_endp == slot->usb_endp);

	if (input_packet.length < slot->max_plen) {	// this is a SHORT PACKET

		SLOT_ASSERT(1 == slot->running);
		TLER_ASSERT(0 == (tler & 1));
		WRITE_TLER(vmb,td,0x00);
		slot->running = 0;
		slot->retired = 1;

	}
	if (0x0000 == (0x007F & slot->recieved)) {
		UTL_DEBUG(1,vmb);
	}

	DBG(1,"%s(packet.length=%d overspill_length=%d)\n",__FUNCTION__,input_packet.length,slot->extra);
}
static void retry_rejected_bulk_packet(struct vmb *vmb,struct vmb_slot*slot,u16 error_packet_id)
{
	u16 idref_fail = SLOT_IDREF_MASK & error_packet_id;
	u16 packet_no = error_packet_id >> SLOT_IDREF_SHIFT;	// ZERO is the 1st in the urb
	u16 urb_count = packet_no << vmb_maxpacket_shift(slot->maxusbpktlen);
	u16 err_index = slot->idref[idref_fail].queue_this;

	URB_DBG(0, "%s(retrying %s -->%d:%d)\n",__FUNCTION__
			, PIPE_INTERRUPT == slot->pipetype ? "INTERRUPT"
			: PIPE_ISOCHRONOUS == slot->pipetype ? "ISOCHRONOUS"
			: PIPE_BULK == slot->pipetype  ? "BULK"
			: "CONTROL",slot->usb_addr,slot->usb_endp);

	struct vmb_udev*udev = &vmb->udev[slot->usb_addr];

	slot->retry_flag = 1;

	if (0 == slot->idref_pkts) {
		slot->idref_this -= 1;
		slot->idref_done -= 1;
	}

	while(SLOT_IDREF_MASK & (slot->idref_this ^ idref_fail)) {
		struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & slot->idref_this];
		u16 urb_index = idref->queue_this;
		SLOT_ASSERT(SLOT_QUEUE_MASK & (urb_index ^ slot->queue_next));
		struct urb*urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];
		SLOT_ASSERT(find_urb_position_in_slot(slot,urb));
		URB_ASSERT(udev->usb_device == urb->dev);
		URB_ASSERT(slot->usb_addr == usb_pipedevice(urb->pipe));
		URB_ASSERT(slot->usb_endp == usb_pipeendpoint(urb->pipe));
		URB_ASSERT(slot->maxusbpktlen == usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe)));
		URB_DBG(0,"%s(rolling back through urb[%d]=%p and %d done)\n",__FUNCTION__
				,urb_index,urb,slot->idref_done);

		urb->actual_length = 0;
		slot->idref_this -= 1;
		slot->idref_done -= 1;
	}
	SLOT_ASSERT(idref_fail == (SLOT_IDREF_MASK & slot->idref_this));
	// now slot->idref_this is referencing the NAK'ed urb
	{
		struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & slot->idref_this];
		u16 urb_index = idref->queue_this;
		SLOT_ASSERT(SLOT_QUEUE_MASK & (urb_index ^ slot->queue_next));	// failed urb_index <> slot->queue_next
		struct urb*urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];
		SLOT_ASSERT(find_urb_position_in_slot(slot,urb));
		URB_ASSERT(udev->usb_device == urb->dev);
		URB_ASSERT(slot->usb_addr == usb_pipedevice(urb->pipe));
		URB_ASSERT(slot->usb_endp == usb_pipeendpoint(urb->pipe));
		URB_ASSERT(slot->maxusbpktlen == usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe)));

		URB_ASSERT(urb_count < urb->actual_length);	// we don't send empty BULK/INTERRUPT/ISOCHRONOUS packets

		PKT_ERROR(2,urb);

		urb->actual_length = urb_count;
		slot->next_urb_index = urb_index;
		slot->next_urb_count = urb_count;

		u8 toggle = 1 & (packet_no ^ slot->toggle[SLOT_QUEUE_MASK & err_index]);
		usb_settoggle(udev->usb_device,slot->usb_endp,1,toggle);

		slot->idref_pkts = packet_no;

		URB_DBG(0, "%s(rolled back to urb[%d]=%p with %d done len=%d pkts=%d toggle=%d)\n",__FUNCTION__
				,urb_index,urb,slot->idref_done,urb_count,packet_no,toggle);
	}
	return;
}
static void retry_rejected_control_packet(struct vmb *vmb,struct vmb_slot*slot,u16 error_packet_id)
{
	if (0 < slot->idref_done && 0 == slot->idref_pkts) {
		slot->idref_this -= 1;
		slot->idref_done -= 1;
	}
	slot->next_urb_index -= slot->idref_done;
	slot->idref_done = 0;

	struct urb *urb = slot_to_head_urb(slot);
	SLOT_ASSERT(urb);

	ERR("%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) CONTROL urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
	URB_SETUP(0,urb);
	URB_DEBUG(0,urb);

	urb->actual_length = 0;

	URB_DBG(0, "%s(retrying %s -->%d:%d)\n",__FUNCTION__
			, PIPE_INTERRUPT == slot->pipetype ? "INTERRUPT"
			: PIPE_ISOCHRONOUS == slot->pipetype ? "ISOCHRONOUS"
			: PIPE_BULK == slot->pipetype  ? "BULK"
			: "CONTROL",slot->usb_addr,slot->usb_endp);

	if (VMB_URB_STATE_CONTROL_ACK_OUT == urb->error_count) {
		urb->error_count = VMB_URB_STATE_CONTROL_SETUP;
	}	
	SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next - slot->queue_size)));
	SLOT_ASSERT(0 == slot->next_urb_count);

	WRITE_TLER(vmb,slot->td,0x00);        // clear out the TD
	slot->running = 0;
	slot->actioned = 0;
	SLOT_ASSERT(1 == slot->allocated);
	SLOT_ASSERT(1 == slot->initialized);
	SLOT_ASSERT(0 == slot->retired);
	SLOT_ASSERT(0 == slot->end_seen);
	SLOT_ASSERT(0 == slot->retry_flag);
	SLOT_ASSERT(0 == slot->running);
	SLOT_ASSERT(0 == slot->device_error);
	SLOT_ASSERT(0 == slot->actioned);

	return;
}
static void fail_rejected_bulk_packet(struct vmb *vmb,struct vmb_slot*slot,u8 error_packet_exterr)
{
	slot->device_error = 1;
	if (0 < slot->idref_done && 0 == slot->idref_pkts) {
		slot->idref_this -= 1;
		slot->idref_done -= 1;
	}
	slot->next_urb_index -= slot->idref_done;
	slot->idref_done = 0;

giveback_next_urb:;

	struct urb *urb = slot_to_head_urb(slot);
	SLOT_ASSERT(urb);

	URB_DBG(0, "%s(failing %s -->%d:%d) urb=%p\n",__FUNCTION__
			,PIPE_INTERRUPT == slot->pipetype ? "INTERRUPT"
			: PIPE_ISOCHRONOUS == slot->pipetype ? "ISOCHRONOUS"
			: PIPE_BULK == slot->pipetype  ? "BULK"
			: "CONTROL",slot->usb_addr,slot->usb_endp,urb);

	URB_DBG(3,"%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);

	urb->actual_length = 0;

	if ((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next - slot->queue_size))) {
		slot->next_urb_index += 1;
		slot->next_urb_count = 0;
	}

	/*
	 * what error to return? maybe one of
	 *
	 *     -EAGAIN
	 *     -EBUSY
	 *     -ECONNRESET
	 *     -EEXIST
	 *     -EFAULT
	 *     -EHOSTUNREACH
	 *     -EINTR
	 *     -EIO
	 *     -EMSGSIZE
	 *     -ENODATA
	 *     -ENODEV
	 *     -ENOENT
	 *     -ENOIOCTLCMD
	 *     -ENOMEM
	 *     -ENOMSG
	 *     -ENOTCONN
	 *     -ENOTEMPTY
	 *     -ENOTSUPP
	 *     -ENXIO
	 *     -EOPNOTSUPP
	 *     -EPERM
	 *     -EPIPE
	 *     -EPROTO
	 *     -ESHUTDOWN
	 *     -ESRCH
	 *     -ETIMEDOUT
	 *     -EXDEV
	 *
	 */
	if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,-ENODEV)) {

		SLOT_ASSERT(0 == slot->actioned);

		if (0x1  == error_packet_exterr // No Target
		|| 0x0  == error_packet_exterr // No Error
		) {
			u8 usb_addr = slot->usb_addr;
			struct vmb_udev*udev = &vmb->udev[usb_addr];

			udev->devicefail = 1;

			port_debug(vmb,slot);

			if (0 < udev->portnumber) {
				struct vmb_port*port = &vmb->port[udev->portnumber - 1];

				SLOT_ASSERT(VPS_running == port->status || VPS_enumerating == port->status);

				port->status = VPS_error;
				port->udev = NULL;   // CLEAR USAGE COUNT ?????

			} else {
				// device probably connected through a hub	
			}
			slot->retired = 1;
			slot->running = 0;
			slot->usb_addr = 0;
			slot->jiffies = jiffies;

			free_vmb_slot(vmb,slot);
		} else {
			slot->actioned = 1;	
			//slot->retired = 1;
			//slot->usb_addr = 0;
			//slot->usb_endp = 0;
			slot->jiffies = jiffies;
	
			/*
			 * leave the slot:-
			 */
			SLOT_ASSERT(1 == slot->allocated);
			SLOT_ASSERT(1 == slot->initialized);
			SLOT_ASSERT(0 == slot->retired);
			SLOT_ASSERT(0 == slot->end_seen);
			SLOT_ASSERT(0 == slot->retry_flag);
			SLOT_ASSERT(1 == slot->running);
			SLOT_ASSERT(1 == slot->device_error);
			SLOT_ASSERT(1 == slot->pipeout);
			/*
			 * so that if the USB CORE does a retry it will get
			 *    the same slot and thus garentee that the tx fifo
			 *    is properly flushed
			 * we might have to ensure that the retry from the
			 *    USB CORE gets the "retry" bit set inside the
			 *    txfifo header --> use the "device_error" flag
			 */
		}
		return;

	} else {
		SLOT_ASSERT(slot->queue_size > 0);

		slot->jiffies = jiffies;

		goto giveback_next_urb;		
	}
}
static void fail_rejected_control_packet(struct vmb *vmb,struct vmb_slot*slot,u8 error_packet_exterr)
{
	SLOT_ASSERT(PIPE_CONTROL == slot->pipetype);

	if (0 < slot->idref_done && 0 == slot->idref_pkts) {
		slot->idref_this -= 1;
		slot->idref_done -= 1;
	}
	slot->next_urb_index -= slot->idref_done;
	slot->idref_done = 0;

	struct urb *urb = slot_to_head_urb(slot);
	SLOT_ASSERT(urb);
giveback_next_urb:;
	URB_DBG(0, "%s(failing %s -->%d:%d) urb=%p\n",__FUNCTION__
			, PIPE_INTERRUPT == slot->pipetype ? "INTERRUPT"
			: PIPE_ISOCHRONOUS == slot->pipetype ? "ISOCHRONOUS"
			: PIPE_BULK == slot->pipetype  ? "BULK"
			: "CONTROL",slot->usb_addr,slot->usb_endp,urb);

	ERR("%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) CONTROL urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
	URB_SETUP(0,urb);
	URB_DEBUG(0,urb);

	urb->actual_length = 0;

	if ((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next - slot->queue_size))) {
		slot->next_urb_index += 1;
		slot->next_urb_count = 0;
	}
	if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,-ENODEV)) {	// -ECONNRESET -EFAULT -EPIPE
	
		if (0x1  == error_packet_exterr // No Target
		|| 0x0  == error_packet_exterr // No Error
		) {
			u8 usb_addr = slot->usb_addr;
			struct vmb_udev*udev = &vmb->udev[usb_addr];

			udev->devicefail = 1;

			port_debug(vmb,slot);

			if (0 < udev->portnumber) {
				struct vmb_port*port = &vmb->port[udev->portnumber - 1];

				SLOT_ASSERT(VPS_running == port->status || VPS_enumerating == port->status);

				port->status = VPS_error;
				port->udev = NULL;   // CLEAR USAGE COUNT ?????

			} else {
				// device probably connected through a hub	
			}
		} else {
		}

		slot->retired = 1;
		slot->running = 0;
		slot->usb_addr = 0;
		SLOT_ASSERT(0 == slot->usb_endp);
		slot->jiffies = jiffies;

		free_vmb_slot(vmb,slot);

		return;
	} else {
		SLOT_ASSERT(slot->queue_size > 0);

		urb = slot_to_head_urb(slot);
		SLOT_ASSERT(urb);
		SLOT_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));
		SLOT_ASSERT(VMB_URB_STATE_CONTROL_SETUP == urb->error_count);

		slot->jiffies = jiffies;

		goto giveback_next_urb;		
	}	
}
static void rx_packet_rejected(struct vmb *vmb)	// ERROR packet
{
	ASSERT(vmb);
	SW_ASSERT_LOCKED(vmb);

	unsigned td = error_packet.utle;

	ASSERT(td < 16);
	struct vmb_slot*slot = &vmb->slot[td];
	ASSERT(td == slot->td);
	ASSERT(vmb == slot->vmb);

	if (0 == slot->allocated) {
		DBG(0,"SLOT[%X] probably dequeued whilst packet being rejected\n",slot->td);	
		return;
	} else if (0x01 == error_packet.status) { // NAK

		SLOT_ASSERT(1 == slot->initialized);
		SLOT_ASSERT(0 == slot->device_error);

		SLOT_ASSERT(slot->queue_size > 0);

		u16 idref_fail = SLOT_IDREF_MASK & error_packet.id;
		u16 packet_no = error_packet.id >> SLOT_IDREF_SHIFT;	// ZERO is the 1st in the urb
		u16 urb_count = packet_no << vmb_maxpacket_shift(slot->maxusbpktlen);
		u16 err_index = slot->idref[idref_fail].queue_this;


		SLOT_ASSERT(1 == slot->running);	// still running
		SLOT_ASSERT(0 == slot->retired);	// still active
		u16 tler = READ_TLER(vmb,td);
		TLER_ASSERT(1 == (tler & 1));	// it should still be active and flushing OUT's and SETUP's
		TLER_ASSERT((0x0000 == (0x0038 & tler)) || (0x0010 == (0x0038 & tler)));	// OUT or SETUP
	
		UTL_DEBUG(1,vmb);

		if (PIPE_INTERRUPT == slot->pipetype) {
			SLOT_ASSERT(1 == slot->pipeout);
			DBG(0, "%s(TD=%X UTLE=%04X packet.ID=%04X urb_index=%02X urb_count=%04X) INTERRUPT NAK ****ERROR****\n",__FUNCTION__,td,tler,error_packet.id,err_index,urb_count);
		} else if (PIPE_ISOCHRONOUS == slot->pipetype) {
			SLOT_ASSERT(1 == slot->pipeout);
			DBG(0, "%s(TD=%X UTLE=%04X packet.ID=%04X urb_index=%02X urb_count=%04X) ISOCHRONOUS NAK ****ERROR****\n",__FUNCTION__,td,tler,error_packet.id,err_index,urb_count);
		} else if (PIPE_BULK == slot->pipetype) {
			SLOT_ASSERT(1 == slot->pipeout);
			DBG(2, "%s(TD=%X UTLE=%04X packet.ID=%04X urb_index=%02X urb_count=%04X) BULK NAK ****ERROR****\n",__FUNCTION__,td,tler,error_packet.id,err_index,urb_count);
		} else {
			DBG(0, "%s(TD=%X UTLE=%04X packet.ID=%04X urb_index=%02X urb_count=%04X) CONTROL NAK ****ERROR****\n",__FUNCTION__,td,tler,error_packet.id,err_index,urb_count);
		}

		retry_rejected_bulk_packet(vmb,slot,error_packet.id);

		return;
	} else if (0x02 == error_packet.status) { // STALL

		UTL_ASSERT(0);

		SLOT_ASSERT(1 == slot->initialized);
		SLOT_ASSERT(0 == slot->device_error);

		SLOT_ASSERT(slot->queue_size > 0);

		u16 idref_fail = SLOT_IDREF_MASK & error_packet.id;

		struct slot_idref*idref = &slot->idref[idref_fail];
		u16 urb_index = idref->queue_this;
		struct urb*urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];

		DBG(1,"%s(STALL urb=%p packet.ID=%d) ****ERROR****\n",__FUNCTION__,urb,error_packet.id);

		return;
	} else if (0x03 == error_packet.status) { // FATAL

		SLOT_ASSERT(1 == slot->initialized);
		SLOT_ASSERT(0 == slot->device_error);

		SLOT_ASSERT(slot->queue_size > 0);

		u16 idref_fail = SLOT_IDREF_MASK & error_packet.id;
		u16 packet_no = error_packet.id >> SLOT_IDREF_SHIFT;	// ZERO is the 1st in the urb
		u16 urb_count = packet_no << vmb_maxpacket_shift(slot->maxusbpktlen);
		u16 err_index = slot->idref[idref_fail].queue_this;

		// just for debugging ....	
		{
			struct slot_idref*idref = &slot->idref[idref_fail];
			u16 urb_index = idref->queue_this;
			struct urb*urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];

			char*error = 0x5 == error_packet.exterr ? "Babble"
					: 0x4 == error_packet.exterr ? "Bad PID"
					: 0x3 == error_packet.exterr ? "Bad CRC"
					: 0x2 == error_packet.exterr ? "False EOP"
					: 0x1  == error_packet.exterr ? "No Target"
					: 0x0  == error_packet.exterr ? "No Error"
					: "unknown error";

			char*pipetype = PIPE_INTERRUPT == slot->pipetype ? "INTERRUPT"
					: PIPE_ISOCHRONOUS == slot->pipetype ? "ISOCHRONOUS"
					: PIPE_BULK == slot->pipetype  ? "BULK"
					: "CONTROL";
			char*direction = 1 == slot->pipeout ? "OUT" : "IN";

			u16 tler = READ_TLER(vmb,td);

			DBG(0, "%s(TD=%X UTLE=%04X urb=%p error_count=%d packet.ID=%04X urb_index=%02X urb_count=%04X) %s %s ****ERROR**** %d = %s\n"
				,__FUNCTION__,td,tler,urb,urb->error_count,error_packet.id,err_index,urb_count,pipetype,direction,error_packet.exterr,error);
		}	
		SLOT_ASSERT(0 == slot->retired);	// still active
		SLOT_ASSERT(1 == slot->running);	// still running
		int corrupt;
		u16 tler = READ_TLER(vmb,td);
		// there are too many holes this algorithm could fall through
		// the TLE engine might get stuck on a queued OUT in the fifo
		// if we clear the tler to soon - it might be better to delay
		// until we are sure that the fifo is flushed - that would
		// mean letting done_frame() handle the error responses
		UTL_DEBUG(1,vmb);

		if ((slot->usb_addr == ((tler>>6)&0x07)) && (slot->usb_endp == ((tler>>9)&0x0F))) {
			corrupt = 0;
		} else {
			DBG(0,"SLOT[%X] --> %d:%d tler=%04X seems corrupt\n",slot->td,slot->usb_addr,slot->usb_endp,tler);	
			corrupt = 1;
		}
		port_debug(vmb,slot);

		if (PIPE_BULK == slot->pipetype) {

			if (0 == (tler & 1)) {
				DBG(0,"SLOT[%X] BULK OUT tler=%04X seems corrupt\n",slot->td,tler);	
			}

			SLOT_ASSERT(0 == slot->actioned);
			if (0x0000 != (0x0038 & tler)) {
				corrupt = 1;
				DBG(0,"SLOT[%X] BULK OUT tler=%04X seems corrupt\n",slot->td,tler);	
			}			

			int retry = (0 < pc_retry) || (0 < corrupt);
			if (retry && (10 > slot->retries)) {
				if (corrupt) {
					u16 new_tle = 0x6 | set_slot_tle(slot,0x0,1); // OUT + FLUSH
					WRITE_TLER(vmb,td,new_tle); // OUT + FLUSH
					DBG(0,"SLOT[%X] BULK OUT resetting tler to %04X\n",slot->td,new_tle);
				}
				slot->retries += 1;	
				retry_rejected_bulk_packet(vmb,slot,error_packet.id);
			} else {
				fail_rejected_bulk_packet(vmb,slot,error_packet.exterr);
			}	
			return;
		} else {	
			if (1 == (tler & 1)) {
				DBG(0,"SLOT[%X] %s OUT tler=%04X seems corrupt\n",slot->td
						, PIPE_INTERRUPT == slot->pipetype ? "INTERRUPT"
						: PIPE_ISOCHRONOUS == slot->pipetype ? "ISOCHRONOUS"
						: PIPE_BULK == slot->pipetype  ? "BULK"
						: "CONTROL"
						,tler);	
			}

			SLOT_ASSERT(0 == slot->retired);
			SLOT_ASSERT(1 == slot->running);
			SLOT_ASSERT(1 == slot->allocated);
			SLOT_ASSERT(1 == slot->initialized);
			SLOT_ASSERT(1 == slot->actioned);
			SLOT_ASSERT(0 == slot->end_seen);
			SLOT_ASSERT(0 == slot->retry_flag);
			SLOT_ASSERT(0 == slot->device_error);
	
			struct urb *urb = slot_to_head_urb(slot);
			SLOT_ASSERT(urb);
			if (0x0010 == (0x0038 & tler)) {
				UTL_ASSERT(VMB_URB_STATE_CONTROL_SETUP == urb->error_count);	// SETUP
			} else if (0x0000 == (0x0038 & tler)) {
				UTL_ASSERT(VMB_URB_STATE_CONTROL_ACK_OUT == urb->error_count);	// ACK OUT
			} else {
				UTL_ASSERT(0);
			}
			if ((0 < pc_retry) && (10 > slot->retries)) {
				slot->retries += 1;	
				retry_rejected_control_packet(vmb,slot,error_packet.id);
			} else {
				fail_rejected_control_packet(vmb,slot,error_packet.exterr);
			}	
			return;
		}	
	} else { // SUCCESS

		UTL_ASSERT(0);

		SLOT_ASSERT(1 == slot->initialized);
		SLOT_ASSERT(0 == slot->device_error);

		SLOT_ASSERT(slot->queue_size > 0);

		u16 idref_fail = SLOT_IDREF_MASK & error_packet.id;
		struct slot_idref*idref = &slot->idref[idref_fail];
		u16 urb_index = idref->queue_this;
		struct urb*urb = slot->u_r_b[SLOT_QUEUE_MASK & urb_index];

		DBG(1,"%s(SUCCESS urb=%p packet.ID=%d) ****ERROR****\n",__FUNCTION__,urb,error_packet.id);
		return;
	}
	ASSERT(0);
}
//------------------------------------------------------------------------------
//
// IN SPD packet
//
//------------------------------------------------------------------------------
static void rx_packet_received(struct vmb *vmb,unsigned td,struct urb *urb)	// called from vmb_check_rx_fifo()
{
	ASSERT(vmb);
	SW_ASSERT_LOCKED(vmb);
	ASSERT(td < 16);
	ASSERT(td == input_packet.utle);

	struct vmb_slot *slot = &vmb->slot[td];
	ASSERT(td == slot->td);
	ASSERT(vmb == slot->vmb);

	SLOT_ASSERT(1 == slot->allocated);
	SLOT_ASSERT(1 == slot->initialized);
	SLOT_ASSERT(0 == slot->device_error);

	slot->recieved += 1;

	SLOT_ASSERT(slot->queue_size > 0);
	SLOT_ASSERT(urb == slot_to_head_urb(slot));

	u8 usb_addr = usb_pipedevice(urb->pipe);
	u8 usb_endp = usb_pipeendpoint(urb->pipe);

	URB_ASSERT(slot->pipetype == usb_pipetype(urb->pipe));
	URB_ASSERT(slot->usb_addr == usb_pipedevice(urb->pipe));
	URB_ASSERT(slot->usb_endp == usb_pipeendpoint(urb->pipe));
	URB_ASSERT(slot->pipeout == usb_pipeout(urb->pipe));	
	URB_ASSERT(slot->maxusbpktlen == usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe)));

	if (VMB_URB_STATE_CONTROL_IN == urb->error_count) {

		URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));
		SLOT_ASSERT(0 == slot->idref_done);
/*-IN-USE-*/	SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));

		URB_ASSERT((input_packet.length + urb->actual_length) <= urb->transfer_buffer_length);

		u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
		// note that transfer_buffer_length can be less than maxusbpktlen

		u16 len = input_packet.length;

		int space = urb->transfer_buffer_length - urb->actual_length;	
		u8*tb = urb->actual_length + (u8*)urb->transfer_buffer;
		u8*ip = input_packet.data;
	
		while(len-- > 0 && space-- > 0)	{
			*tb++ = *ip++ ;
			urb->actual_length += 1;
		}
		DBG(4,"%s(maxusbpktlen=%d packet.length=%d actual_length=%d)\n",__FUNCTION__,maxusbpktlen,input_packet.length,urb->actual_length);

		if (input_packet.length == maxusbpktlen) {

			slot->next_urb_count = urb->actual_length >> vmb_maxpacket_shift(slot->maxusbpktlen);

			UTL_DEBUG(4,vmb); // xfer not finished yet

		} else if (usb_addr > 0 && 0 < urb->actual_length) {

			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
			u16 tler = READ_TLER(vmb,td);

			if (tler & 0x0001) {

				TLER_ASSERT(0x0004 == (tler & 0x0007));

				// should have retired

				ASSERT(0);	// TODO - check this out

				urb->error_count = VMB_URB_STATE_CONTROL_ERR;

				SLOT_ASSERT(1 == slot->running);	// CONTROL IN finished
				clear_max_plen(vmb,td);
				WRITE_TLER(vmb,td,0x00);        // clear out the TD
				slot->running = 0;	// CONTROL IN stop

				slot->line_number = __LINE__;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
				queue_delayed_work(workqueue,&slot->status_work,2);	// TODO - handle VMB_URB_STATE_CONTROL_ERR
#else
				queue_delayed_work(workqueue,&slot->delayed_work,2);	// TODO - handle VMB_URB_STATE_CONTROL_ERR
#endif

				ERR("%s(TD=%X urb=%p TLE=%04X maxusbpktlen=%d packet.length=%d actual_length=%d) TD did NOT RETIRE\n",__FUNCTION__,td,urb,tler,maxusbpktlen,input_packet.length,urb->actual_length);
				URB_DEBUG(1,urb);
				UTL_DEBUG(1,vmb);

			} else {

				SLOT_ASSERT(1 == slot->running);	// CONTROL IN finished
				TLER_ASSERT(0 == (tler & 0x0007));

				SLOT_ASSERT(1 == slot->queue_size);
/* 08-NOV */				SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
				SLOT_ASSERT(0 == slot->next_urb_count);
	
				clear_max_plen(vmb,td);
		
				urb->error_count = VMB_URB_STATE_CONTROL_ACK_OUT;
	
				slot->ec = 0;
				//u16 new_tle = set_slot_tle(slot,0x0,1);	// OUT
				//WRITE_TLER(vmb,td,new_tle);	// OUT
				WRITE_TLER(vmb,td,0x00);        // clear out the TD
				slot->running = 0;	// CONTROL IN finished
				slot->retired = 0;
				slot->actioned = 0;	// CONTROL ACK OUT to be started
	
				HW_LOCK(vmb);
				u16 start_fc = read_fc(vmb);	// the frame currently or recently finished executing
				u16 txwp = read_txwp(vmb);
				u16 txrp = read_txrp(vmb);
				HW_UNLOCK(vmb);

				DBG(1, "%s(TD=%X urb=%p FC=%04X T=%04X H=%04X R=%04X W=%04X) idref=%d queue=%d\n",__FUNCTION__,td,urb,start_fc,vmb->frame_tail,vmb->frame_head,txrp,txwp,slot->idref_done,slot->queue_size);

				UTL_DEBUG(4,vmb);
			}
		} else {

			slot->next_urb_count = urb->actual_length;
			if (vmb->enumerating) {

				if (vmb->current_port > 0) {
	
					struct vmb_port*port = &vmb->port[vmb->current_port-1];
	
					UTL_ASSERT(VPS_enumerating == port->status);
	
				} else {
	
					UTL_ASSERT(vmb->current_udev);	// but it might not be == urb->dev
					UTL_ASSERT(VPS_monitoring == vmb->port[0].status || VPS_running == vmb->port[0].status);
					UTL_ASSERT(VPS_monitoring == vmb->port[1].status || VPS_running == vmb->port[1].status);
					UTL_ASSERT(VPS_running    == vmb->port[0].status || VPS_running == vmb->port[1].status);
				}
			} else {

				URB_ASSERT(0 == urb->actual_length);
				UTL_ASSERT(VPS_running    == vmb->port[0].status || VPS_running == vmb->port[1].status);
			}

			SLOT_ASSERT(1 == slot->running);	// CONTROL IN finished
			u16 tler = READ_TLER(vmb,td);
			TLER_ASSERT(0 == (tler & 1));	// retired on last xfer packet
			WRITE_TLER(vmb,td,0x00);	// clear out the TD
			slot->running = 0;	 // CONTROL IN being stopped
			slot->retired = 1;

			clear_max_plen(vmb,td);

			if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {

				slot->next_urb_index += 1;
				slot->next_urb_count = 0;
				slot->usb_addr = 0;
				slot->usb_endp = 0;
				slot->jiffies = jiffies;

				free_vmb_slot(vmb,slot);

				SLOT_ASSERT(0 == slot->queue_size);
			} else {
				URB_DBG(1,"%s(TD=%X CONTROL COMPLETED next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) CONTROL urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
				SLOT_ASSERT(slot->queue_size > 0);
				SLOT_ASSERT(1 == slot->retired);
				SLOT_ASSERT(1 == slot->actioned);
				SLOT_ASSERT(1 == slot->end_seen);
				SLOT_ASSERT(0 == slot->retry_flag);
				SLOT_ASSERT(0 == slot->running);
				SLOT_ASSERT(0 == slot->device_error);
				slot->jiffies = jiffies;
				SLOT_ASSERT(0 == slot->next_urb_count);	// not sure about this
				slot->next_urb_index += 1;
				slot->next_urb_count = 0;
				slot->retired = 0;
				slot->actioned = 0;
				slot->end_seen = 0;
				slot->retry_flag = 0; // should be already clear
				slot->device_error = 0;	// should be already true
	
				SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
				SLOT_DEBUG(1,slot);
				SLOT_ASSERT(0 == slot->end_seen);
			}

			UTL_DEBUG(4,vmb);
		}
	} else if (VMB_URB_STATE_CONTROL_ACK_IN == urb->error_count) {

		URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));
		SLOT_ASSERT(0 == slot->idref_done);
		SLOT_ASSERT(0 == slot->next_urb_count);
		SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (1+slot->queue_next-slot->queue_size)));
		URB_ASSERT(urb->actual_length == urb->transfer_buffer_length);	// already have sent them out
		URB_ASSERT(0 < urb->actual_length);

		UTL_ASSERT(0 == input_packet.length);

		if (0 < input_packet.length) {
			ERR("received %d bytes of data in an IN ACK\n",input_packet.length);
		} else {
		}
		DBG(4,"%s(CONTROL_OUT ACK IN packet.length=%d)\n",__FUNCTION__,input_packet.length);

		if (vmb->enumerating) {

			if (vmb->current_port > 0) {

				struct vmb_port*port = &vmb->port[vmb->current_port-1];

				UTL_ASSERT(VPS_enumerating == port->status);

			} else {

				UTL_ASSERT(vmb->current_udev);	// but it might not be == urb->dev
				UTL_ASSERT(VPS_monitoring == vmb->port[0].status || VPS_running == vmb->port[0].status);
				UTL_ASSERT(VPS_monitoring == vmb->port[1].status || VPS_running == vmb->port[1].status);
				UTL_ASSERT(VPS_running    == vmb->port[0].status || VPS_running == vmb->port[1].status);
			}
		} else {

			UTL_ASSERT(VPS_running    == vmb->port[0].status || VPS_running == vmb->port[1].status);
		}

		SLOT_ASSERT(1 == slot->running);	// CONTROL ACK IN finished
		u16 tler = READ_TLER(vmb,td);
		TLER_ASSERT(0 == (tler & 1));	// retired on last xfer packet
		WRITE_TLER(vmb,td,0x00);	// clear out the TD
		slot->running = 0;	 // CONTROL ACK IN being stopped
		slot->retired = 1;

		clear_max_plen(vmb,td);

		if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {

			slot->usb_addr = 0;
			slot->usb_endp = 0;
			slot->jiffies = jiffies;

			SLOT_ASSERT(0 == slot->queue_size);	// not whilst CONTROL

			free_vmb_slot(vmb,slot);

		} else {
			URB_DBG(1,"%s(TD=%X CONTROL COMPLETED next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) CONTROL urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
			SLOT_ASSERT(slot->queue_size > 0);
			SLOT_ASSERT(1 == slot->retired);
			SLOT_ASSERT(1 == slot->actioned);
			SLOT_ASSERT(1 == slot->end_seen);
			SLOT_ASSERT(0 == slot->retry_flag);
			SLOT_ASSERT(0 == slot->running);
			SLOT_ASSERT(0 == slot->device_error);
			slot->jiffies = jiffies;
			SLOT_ASSERT(0 == slot->next_urb_count);	// not sure about this
			slot->next_urb_count = 0;
			slot->retired = 0;
			slot->actioned = 0;
			slot->end_seen = 0;
			slot->retry_flag = 0; // should be already clear
			slot->device_error = 0;	// should be already true

			SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));
			SLOT_DEBUG(1,slot);
			SLOT_ASSERT(0 == slot->end_seen);
		}

		UTL_DEBUG(4,vmb);

	} else if (VMB_URB_STATE_INTERRUPT == urb->error_count) {

		URB_ASSERT(PIPE_INTERRUPT == usb_pipetype(urb->pipe));
		SLOT_ASSERT(1 == slot->queue_size);
		SLOT_ASSERT(1 == slot->idref_done);	// idref entry is used for timeout
		SLOT_ASSERT(1 == slot->running); // INTERRUPT IN
		SLOT_ASSERT(1 == slot->actioned);
		SLOT_ASSERT(0 == slot->retired);
		UTL_ASSERT(0 < usb_addr);

		struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & (slot->idref_this - slot->idref_done)];
		SLOT_ASSERT((SLOT_QUEUE_MASK & idref->queue_this) == (SLOT_QUEUE_MASK & slot->next_urb_index));

		slot->idref_done -= 1;	// stop the scheduler from retrying

		URB_ASSERT((input_packet.length + urb->actual_length) <= urb->transfer_buffer_length);

		u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
		URB_ASSERT((NULL == urb->transfer_buffer)||(maxusbpktlen <= urb->transfer_buffer_length));

		u16 len = input_packet.length;
	
		int space = urb->transfer_buffer_length - urb->actual_length;	
		u8*tb = urb->actual_length + (u8*)urb->transfer_buffer;
		u8*ip = input_packet.data;
	
		while(len-- > 0 && space-- > 0) {
			*tb++ = *ip++ ;
			urb->actual_length += 1;
		}

		u16 tler = READ_TLER(vmb,td);

		DBG(4,"%s(maxusbpktlen=%d packet.length=%d actual_length=%d)\n",__FUNCTION__,maxusbpktlen,input_packet.length,urb->actual_length);

		if (1 == (tler & 0x0001)) {

			SLOT_ASSERT(0);	// should never happen since we use IN ONCE
			/*
			 * we have a problem if an INTERRUPT xfer requires
			 *    more than one IN packet to complete it
			 */
			/*
			 * probably need to rely on the urb->transfer_buffer_length
			 *    and we might have to retire or change the TD to IN ONCE 
			 *
			 */
			UTL_DEBUG(4,vmb);

		} else {	// retired on last xfer packet

			URB_ASSERT(urb->transfer_buffer_length >= urb->actual_length);	// we do not handle overspill on INTERRUPT IN's

			u16 maxvmbpktlen = vmb_maxpacket_len(maxusbpktlen);
			TLER_ASSERT(maxvmbpktlen >= maxusbpktlen);

			URB_ASSERT(maxusbpktlen == urb->transfer_buffer_length);	// possibly not true - depends on device end point

			WRITE_TLER(vmb,td,0x00);	// clear out the TD
			slot->running = 0;	// INTERRUPT IN being stopped
			slot->retired = 1;
	
			clear_max_plen(vmb,td);

			if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {

				slot->next_urb_index += 1;
				slot->next_urb_count = 0;
				slot->usb_addr = 0;
				slot->usb_endp = 0;
				slot->jiffies = jiffies;

				free_vmb_slot(vmb,slot);
			} else {
				SLOT_ASSERT(0);	// this really should not happen
				ERR("%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) CONTROL urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
				SLOT_ASSERT(slot->queue_size > 0);
				slot->jiffies = jiffies;
				slot->next_urb_index += 1;
				slot->next_urb_count = 0;
			}

			UTL_DEBUG(4,vmb);
			SLOT_ASSERT(0 == slot->queue_size);	// not during INTERRUPT
		}
	} else if (VMB_URB_STATE_BULK_IN == urb->error_count) {

		URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
		SLOT_ASSERT(0 == slot->idref_done);	// not used for BULK IN's
		URB_ASSERT(usb_pipein(urb->pipe));

		URB_ASSERT(urb->actual_length < urb->transfer_buffer_length);

		UTL_ASSERT(usb_addr > 0 && usb_addr < MAX_DEV_ADDRS);

		SLOT_ASSERT(0 == slot->extra);

		u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
		u16 maxvmbpktlen = vmb_maxpacket_len(maxusbpktlen);

		TLER_ASSERT(8==maxvmbpktlen || 16==maxvmbpktlen || 32==maxvmbpktlen || 64==maxvmbpktlen);

		u16 len = input_packet.length;
	
		int space = urb->transfer_buffer_length - urb->actual_length;	
		u8*tb = urb->actual_length + (u8*)urb->transfer_buffer;
		u8*ip = input_packet.data;

store_data:

		while(len > 0 && space > 0) {
			len -= 1;
			space -= 1;
			*tb++ = *ip++ ;
			urb->actual_length += 1;
		}

		if (len > 0) {

			URB_ASSERT(urb->transfer_buffer_length == urb->actual_length);
			UTL_ASSERT(0 == space);

			DBG(1,"%s(maxusbpktlen=%d packet.length=%d actual_length=%d overspill=%d)\n",__FUNCTION__,maxusbpktlen,input_packet.length,urb->actual_length,slot->extra);

			if (input_packet.length < maxvmbpktlen) {	// SHORT PACKET
	
				u16 tler = READ_TLER(vmb,td);
	
				SLOT_ASSERT(0 == slot->running);	// BULK IN finished
				TLER_ASSERT(0 == (tler & 1));	// only retires on short packet
	
				if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {
					/*
					 * Leave this slot active to wait for next urb to mop up the data
					 *
					 */
					slot->next_urb_index += 1;
					slot->next_urb_count = 0;
					slot->retired = 1;
					slot->jiffies = jiffies;
		
					URB_DBG(1, "%s(TD=%X usb_hcd_giveback_urb=%p) BULK IN length=%d status=%d extra=%d\n",__FUNCTION__,td,urb,urb->actual_length,urb->status,slot->extra);

					SLOT_ASSERT(0 == slot->extra);	
					u16 spill = sizeof(slot->overspill);
					u8*eb = slot->extra + slot->overspill;
					while(len > 0 && spill > 0) {
	            		len -= 1;
						spill -= 1;
    	        		*eb++ = *ip++ ;
						slot->extra += 1;
					}
					UTL_ASSERT(0 == len);	// or the overspill buffer is TOO SMALL
				} else {
					DBG(1, "%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) BULK IN urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
					SLOT_ASSERT(slot->queue_size > 0);

					slot->next_urb_index += 1;
					slot->next_urb_count = 0;
					slot->retired = 1;
					slot->jiffies = jiffies;
		
					URB_DBG(1, "%s(TD=%X usb_hcd_giveback_urb=%p) BULK IN OVERSPILL length=%d status=%d extra=%d\n",__FUNCTION__,td,urb,urb->actual_length,urb->status,slot->extra);
	
					urb = slot_to_head_urb(slot);
	
					SLOT_ASSERT(urb);
					URB_ASSERT(VMB_URB_STATE_BULK_IN == urb->error_count);
					URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
					URB_ASSERT(usb_pipein(urb->pipe));
	
					slot->jiffies = jiffies;
					slot->line_number = __LINE__;
					SLOT_DEBUG(1,slot);

					URB_ASSERT(0 == urb->actual_length);
					space = urb->transfer_buffer_length;
					tb = urb->transfer_buffer;

					goto store_data;					
				}
				SLOT_ASSERT(usb_addr == slot->usb_addr);
				SLOT_ASSERT(usb_endp == slot->usb_endp);

				UTL_DEBUG(4,vmb);

			} else {
	
				TLER_ASSERT(input_packet.length == maxvmbpktlen);

				// the TD might have retired already
				// with the other packets in the fifo

				if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {
					/*
					 * leave this TD running to catch any more overspill
					 *   and to wait for next urb to mop up the data
					 *
					 */

					slot->next_urb_index += 1;
					slot->next_urb_count = 0;
					slot->retired = 0;
					slot->jiffies = jiffies;
		
					URB_DBG(1, "%s(TD=%X usb_hcd_giveback_urb=%p) BULK IN length=%d status=%d extra=%d\n",__FUNCTION__,td,urb,urb->actual_length,urb->status,slot->extra);

					SLOT_ASSERT(0 == slot->extra);	
					u16 spill = sizeof(slot->overspill);
					u8*eb = slot->extra + slot->overspill;
					while(len > 0 && spill > 0) {
	            		len -= 1;
						spill -= 1;
    	        		*eb++ = *ip++ ;
						slot->extra += 1;
					}
					UTL_ASSERT(0 == len);	// or the overspill buffer is TOO SMALL
				} else {	
					DBG(1, "%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) BULK IN urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
					SLOT_ASSERT(slot->queue_size > 0);
					slot->next_urb_index += 1;
					slot->next_urb_count = 0;
					slot->retired = 0;
					slot->jiffies = jiffies;
		
					URB_DBG(1, "%s(TD=%X usb_hcd_giveback_urb=%p) BULK IN OVERSPILL length=%d status=%d extra=%d\n",__FUNCTION__,td,urb,urb->actual_length,urb->status,slot->extra);
	
					urb = slot_to_head_urb(slot);
	
					SLOT_ASSERT(urb);
					URB_ASSERT(VMB_URB_STATE_BULK_IN == urb->error_count);
					URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
					URB_ASSERT(usb_pipein(urb->pipe));
	
					slot->jiffies = jiffies;
					slot->line_number = __LINE__;
					SLOT_DEBUG(1,slot);

					URB_ASSERT(0 == urb->actual_length);
					space = urb->transfer_buffer_length;
					tb = urb->transfer_buffer;

					goto store_data;					
				}
	
				SLOT_ASSERT(usb_addr == slot->usb_addr);
				SLOT_ASSERT(usb_endp == slot->usb_endp);

				UTL_DEBUG(4,vmb);
			}
		} else {	// no overspill

			SLOT_ASSERT(0 == len);
		
			if (input_packet.length < maxvmbpktlen) {

				SLOT_ASSERT(0 == slot->extra);

				DBG(1,"%s(maxusbpktlen=%d packet.length=%d actual_length=%d actioned=%d)\n",__FUNCTION__,maxusbpktlen,input_packet.length,urb->actual_length,slot->actioned);

				SLOT_ASSERT(1 == slot->running);	// BULK IN finished
	
				u16 tler = READ_TLER(vmb,td);

				URB_ASSERT(urb->transfer_buffer_length >= urb->actual_length);
	
				if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {
					slot->next_urb_index += 1;
					slot->next_urb_count = 0;
					slot->jiffies = jiffies;

					if (0 == (tler & 1)) {	// only retires on short packet
						slot->usb_addr = 0;
						slot->usb_endp = 0;
			
						WRITE_TLER(vmb,td,0x00);	// clear out the TD
						slot->running = 0;	// BULK IN being stopped
						slot->retired = 1;
	
						clear_max_plen(vmb,td);
			
						free_vmb_slot(vmb,slot);

						DBG(4, "%s(TD=%X) BULK IN slot is now free\n",__FUNCTION__,td);
					} else {
						SLOT_ASSERT(1 == slot->running);
						SLOT_ASSERT(0 == slot->retired);
						DBG(4, "%s(TD=%X) BULK IN slot is overspilling\n",__FUNCTION__,td);
					}
				} else {
					DBG(1, "%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) BULK IN urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
					SLOT_ASSERT(slot->queue_size > 0);
					slot->jiffies = jiffies;
					slot->next_urb_index += 1;
					slot->next_urb_count = 0;

					if (0 == (tler & 1)) {	// only retires on short packet
						WRITE_TLER(vmb,td,tler|1);	// re activate the TD
					} else {
					}

					struct urb*urb = slot_to_head_urb(slot);
					SLOT_ASSERT(urb);

					URB_ASSERT(VMB_URB_STATE_BULK_IN == urb->error_count);
					URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
					URB_ASSERT(usb_pipein(urb->pipe));

					SLOT_ASSERT(usb_addr == slot->usb_addr);
					SLOT_ASSERT(usb_endp == slot->usb_endp);

					URB_DBG(1, "%s(TD=%X) BULK IN next urb=%p\n",__FUNCTION__,td,urb);

					URB_DEBUG(4,urb);
					UTL_DEBUG(4,vmb);
					SLOT_DEBUG(1,slot);
				}

				UTL_DEBUG(4,vmb);

			} else if (urb->transfer_buffer_length == urb->actual_length) {

				DBG(4,"%s(maxusbpktlen=%d packet.length=%d actual_length=%d)\n",__FUNCTION__,maxusbpktlen,input_packet.length,urb->actual_length);

				// the TD might have retired already
				// with the other packets in the fifo

				SLOT_ASSERT(slot->queue_size > 0);
				SLOT_ASSERT(0 == slot->extra);
				SLOT_ASSERT(0 == slot->retired);

				if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {
					/*
					 * leave this TD running to catch any overspill
					 *
					 */
					slot->next_urb_index += 1;
					slot->next_urb_count = 0;
					slot->retired = 0;
					slot->jiffies = jiffies;
		
					URB_DBG(1, "%s(TD=%X usb_hcd_giveback_urb=%p) BULK IN length=%d status=%d queue_size=%d\n",__FUNCTION__,td,urb,urb->actual_length,urb->status,slot->queue_size);
				} else {
					DBG(1, "%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) BULK IN urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
					SLOT_ASSERT(slot->queue_size > 0);
					slot->next_urb_index += 1;
					slot->next_urb_count = 0;
					slot->retired = 0;
					slot->jiffies = jiffies;
		
					URB_DBG(1, "%s(TD=%X usb_hcd_giveback_urb=%p) BULK IN length=%d status=%d queue_size=%d\n",__FUNCTION__,td,urb,urb->actual_length,urb->status,slot->queue_size);
		
					struct urb*urb = slot_to_head_urb(slot);
	
					SLOT_ASSERT(urb);
					URB_ASSERT(VMB_URB_STATE_BULK_IN == urb->error_count);
					URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
					URB_ASSERT(usb_pipein(urb->pipe));
	
					slot->jiffies = jiffies;
	
					URB_ASSERT(0 == urb->actual_length);
	
					URB_DEBUG(4,urb);
					UTL_DEBUG(1,vmb);
					SLOT_DEBUG(1,slot);
				}

				SLOT_ASSERT(0 == slot->extra);
				SLOT_ASSERT(usb_addr == slot->usb_addr);
				SLOT_ASSERT(usb_endp == slot->usb_endp);

				UTL_DEBUG(4,vmb);
			} else { // xfer not finished yet
	
				slot->next_urb_count = urb->actual_length;
				DBG(4,"%s(maxusbpktlen=%d packet.length=%d actual_length=%d)\n",__FUNCTION__,maxusbpktlen,input_packet.length,urb->actual_length);

				// the TD might have retired already
				// with the other packets in the fifo
	
				// might want to run the timer here in case the USB device does not terminate with a short packet
	
				TLER_ASSERT(input_packet.length == maxvmbpktlen);
	
				UTL_DEBUG(4,vmb);
			}
		}
	} else {

		DBG(1, "%s(urb=%p error_count=%d status=%d) ****ERROR****\n",__FUNCTION__,urb,urb->error_count,urb->status);
		UTL_ASSERT(0); // find out WHY
	}
	if (0x0000 == (0x003F & slot->recieved)) {
		UTL_DEBUG(1,vmb);
	}
	return;
}
static void urb_timer_expired(struct vmb *vmb,unsigned td,struct urb *urb)
{
	ASSERT(vmb);
	SW_ASSERT_LOCKED(vmb);
	ASSERT(td < 16);

	struct vmb_slot *slot = &vmb->slot[td];
	ASSERT(td == slot->td);
	ASSERT(vmb == slot->vmb);

	SLOT_ASSERT(1 == slot->allocated);
	SLOT_ASSERT(1 == slot->initialized);
	SLOT_ASSERT(0 == slot->device_error);

	SLOT_ASSERT(urb == slot_to_head_urb(slot));

	u8 usb_addr = usb_pipedevice(urb->pipe);
	u8 usb_endp = usb_pipeendpoint(urb->pipe);

	SLOT_ASSERT(usb_addr == slot->usb_addr);
	SLOT_ASSERT(usb_endp == slot->usb_endp);

	if (VMB_URB_STATE_CONTROL_STALL == urb->error_count) {

		URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));

		URB_DBG(1, "%s(TD=%X usb_hcd_giveback_urb=%p) STALL on CONTROL length=%d buffer=%d status=%d\n",__FUNCTION__,td,urb,urb->actual_length,urb->transfer_buffer_length,urb->status);

		SLOT_ASSERT(0 == slot->running);	// CONTROL IN stalled
		u16 tler = READ_TLER(vmb,td);
		TLER_ASSERT(0 == (tler & 1));

		WRITE_TLER(vmb,td,0x00);	// clear out the TD
		slot->running = 0;	// CONTROL IN being stopped
		slot->retired = 1;

		clear_max_plen(vmb,td);

control_in_stall:		
		if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,-EPIPE)) {	// STALL
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
			slot->usb_addr = 0;
			slot->usb_endp = 0;
			slot->jiffies = jiffies;

			free_vmb_slot(vmb,slot);
		} else {
			ERR("%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) CONTROL urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
			SLOT_ASSERT(0);
			SLOT_ASSERT(slot->queue_size > 0);
			slot->jiffies = jiffies;
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;

			urb = slot_to_head_urb(slot);

			SLOT_ASSERT(urb);
			URB_ASSERT(VMB_URB_STATE_CONTROL_SETUP == urb->error_count);

			URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));

			goto control_in_stall;
		}

		UTL_DEBUG(4,vmb);

		SLOT_ASSERT(0 == slot->queue_size);	// during CONTROL
		vmb_scheduler(vmb);

	} else if (VMB_URB_STATE_CONTROL_IN_FAILED == urb->error_count) {

		URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));

		URB_DBG(1, "%s(TD=%X usb_hcd_giveback_urb=%p) FATAL on CONTROL IN length=%d buffer=%d status=%d\n",__FUNCTION__,td,urb,urb->actual_length,urb->transfer_buffer_length,urb->status);

		WRITE_TLER(vmb,td,0x00);	// clear out the TD
		slot->running = 0;	// CONTROL IN being stopped
		slot->retired = 1;

		clear_max_plen(vmb,td);

control_in_fail:		
		if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,-EPIPE)) {	// STALL
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
			slot->usb_addr = 0;
			slot->usb_endp = 0;
			slot->jiffies = jiffies;

			free_vmb_slot(vmb,slot);
		} else {
			ERR("%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) CONTROL urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
			ASSERT(0);
			SLOT_ASSERT(slot->queue_size > 0);
			slot->jiffies = jiffies;
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;

			urb = slot_to_head_urb(slot);

			SLOT_ASSERT(urb);
			URB_ASSERT(VMB_URB_STATE_CONTROL_SETUP == urb->error_count);

			URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));

			goto control_in_fail;
		}

		UTL_DEBUG(4,vmb);

		SLOT_ASSERT(0 == slot->queue_size);	// during CONTROL
		vmb_scheduler(vmb);

	} else if (VMB_URB_STATE_BULK_IN_STALL == urb->error_count) {

		URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
		URB_ASSERT(usb_pipein(urb->pipe));

		URB_DBG(1, "%s(TD=%X usb_hcd_giveback_urb=%p) STALL on BULK length=%d buffer=%d status=%d\n",__FUNCTION__,td,urb,urb->actual_length,urb->transfer_buffer_length,urb->status);

		SLOT_ASSERT(0 == slot->running);	// ?????????????
		u16 tler = READ_TLER(vmb,td);
		TLER_ASSERT(0 == (tler & 1));
		WRITE_TLER(vmb,td,0x00);	// clear out the TD
		slot->running = 0;	// stopping the BULK IN slot after timeout
		slot->retired = 1;

		clear_max_plen(vmb,td);

stall_bulk_in_urb:		
		if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,-EPIPE)) {	// STALL
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
			slot->usb_addr = 0;
			slot->usb_endp = 0;
			slot->jiffies = jiffies;

			free_vmb_slot(vmb,slot);
		} else {
			ERR("%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) BULK IN urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
			SLOT_ASSERT(0);
			SLOT_ASSERT(slot->queue_size > 0);
			slot->jiffies = jiffies;
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;

			urb = slot_to_head_urb(slot);

			SLOT_ASSERT(urb);
			URB_ASSERT(VMB_URB_STATE_BULK_IN == urb->error_count);

			URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
			URB_ASSERT(usb_pipein(urb->pipe));

			goto stall_bulk_in_urb;
		}

		UTL_DEBUG(4,vmb);

		SLOT_ASSERT(0 == slot->queue_size);	// TO DO
		vmb_scheduler(vmb);

	} else if (VMB_URB_STATE_BULK_IN_FAILED == urb->error_count) {

		URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
		URB_ASSERT(usb_pipein(urb->pipe));

		URB_DBG(0, "%s(TD=%X usb_hcd_giveback_urb=%p) FAILURE on BULK IN length=%d buffer=%d status=%d\n",__FUNCTION__,td,urb,urb->actual_length,urb->transfer_buffer_length,urb->status);

		SLOT_ASSERT(0 == slot->running);	// ?????????????
		u16 tler = READ_TLER(vmb,td);
		TLER_ASSERT(0 == (tler & 1));
		WRITE_TLER(vmb,td,0x00);	// clear out the TD
		slot->running = 0;	// stopping the BULK IN slot after timeout
		slot->retired = 1;

		clear_max_plen(vmb,td);

fail_bulk_in_urb:		
		if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,-ENOENT)) {	// TODO what error code -ENOENT -ECONNRESET -ESHUTDOWN
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
			slot->usb_addr = 0;
			slot->usb_endp = 0;
			slot->jiffies = jiffies;

			free_vmb_slot(vmb,slot);
		} else {
			SLOT_ASSERT(slot->queue_size > 0);
			slot->jiffies = jiffies;
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;

			urb = slot_to_head_urb(slot);

			SLOT_ASSERT(urb);
			URB_ASSERT(VMB_URB_STATE_BULK_IN == urb->error_count);

			URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
			URB_ASSERT(usb_pipein(urb->pipe));

			goto fail_bulk_in_urb;
		}

		UTL_DEBUG(4,vmb);

		vmb_scheduler(vmb);

	} else if (VMB_URB_STATE_INT_FAILED == urb->error_count) {

		URB_ASSERT(PIPE_INTERRUPT == usb_pipetype(urb->pipe));
		URB_ASSERT(usb_pipein(urb->pipe));

		URB_DBG(0, "%s(TD=%X usb_hcd_giveback_urb=%p) FAILURE on INTERRUPT length=%d buffer=%d status=%d\n",__FUNCTION__,td,urb,urb->actual_length,urb->transfer_buffer_length,urb->status);

		SLOT_ASSERT(0 == slot->running);	// ?????????????
		u16 tler = READ_TLER(vmb,td);
		TLER_ASSERT(0 == (tler & 1));
		WRITE_TLER(vmb,td,0x00);	// clear out the TD
		slot->running = 0;	// stopping the INTERRUPT IN slot after timeout
		slot->retired = 1;

		clear_max_plen(vmb,td);

		if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,-ENOENT)) {	// TODO what error code -ENOENT -ECONNRESET -ESHUTDOWN

			slot->next_urb_index += 1;
			slot->next_urb_count = 0;
			slot->usb_addr = 0;
			slot->usb_endp = 0;
			slot->jiffies = jiffies;

			free_vmb_slot(vmb,slot);
		} else {
			ERR("%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) INTERRUPT IN urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
			SLOT_ASSERT(0);
			SLOT_ASSERT(slot->queue_size > 0);
			slot->jiffies = jiffies;
			slot->next_urb_index += 1;
			slot->next_urb_count = 0;

		}

		UTL_DEBUG(4,vmb);

		SLOT_ASSERT(0 == slot->queue_size);	// TO DO
		vmb_scheduler(vmb);

	} else {

		UTL_ASSERT(0);		// WHY ??
	}
	return;
}
static void vmb_check_rx_fifo(struct vmb*vmb)
{
	int packets = 0;

	ASSERT(vmb);
	SW_ASSERT_LOCKED(vmb);

	check_fifo:;

	u8 rpl = read_rxrplr(vmb);
	u8 rph = read_rxrphr(vmb);
	u16 rp = rpl | (rph << 8);

	u8 wpl = read_rxwplr(vmb);
	u8 wph = read_rxwphr(vmb);
	u16 wp = wpl | (wph << 8);

	u16 length = wp - (rp & 0x3FFF);
	if (rp & 0x8000) length += 1;
	if (rp & 0x4000) length += 1;

	if (length > 0)
	{
		packets += 1;

		u8 h0 = read_fifo(vmb);
		u8 h1 = read_fifo(vmb);
	
		if (h0 & 0x80) {

			error_packet.status = (h0 >> 5) & 0x03;
			error_packet.exterr = (h0 >> 2) & 0x07;
	
			u8 f0 = read_fifo(vmb);
			u8 f1 = read_fifo(vmb);
		
			error_packet.utle = f1 >> 4;
			error_packet.id = f0 | ((f1 & 0x0F) << 8);
	
			rx_packet_rejected(vmb);

		} else {

			input_packet.data01 = (h0 >> 2) & 0x01;
			u16 len = input_packet.length = h1 | ((0x03 & h0) << 8);
	
			u8*b = input_packet.data;
			while(len-- > 0) {
				u8 d = read_fifo(vmb);
				*b++ = d;
			}
		
			u8 crc0 = read_fifo(vmb);
			u8 crc1 = read_fifo(vmb);
			input_packet.crc = crc0 + (crc1 << 8);
			u8 f0 = read_fifo(vmb);
			u8 f1 = read_fifo(vmb);
		
			input_packet.utle = f1 >> 4;
			input_packet.frame = f0 | ((f1 & 0x0f) << 8);
	
			struct vmb_slot*slot = &vmb->slot[input_packet.utle];
			ASSERT(input_packet.utle == slot->td);
			ASSERT(vmb == slot->vmb);

			SLOT_ASSERT(1 == slot->initialized);	// find out why - should not happen
			SLOT_ASSERT(1 == slot->allocated);	// find out why - should not happen
			SLOT_ASSERT(0 == slot->device_error);	// find out why - should not happe

			struct urb *urb = slot_to_head_urb(slot);
	
			if (urb) {
	
				PKT_DEBUG(3,urb);

				rx_packet_received(vmb,input_packet.utle,urb);
	
			} else if (slot->usb_addr > 0 && slot->usb_endp > 0) {
	
				rx_packet_contined(vmb,input_packet.utle);
	
			} else {

				DBG(1,"SLOT %X REMOVED %d:%02d extra=%d jiffies=%08lX\n",input_packet.utle,slot->usb_addr,slot->usb_endp,slot->extra,slot->jiffies);
	
				UTL_ASSERT(0);	// find out why - should not happen
			}
		}
		goto check_fifo;
	
	} else {

		DBG(3, "%s() %d packets completed\n",__FUNCTION__,packets);

		return;
	}
}
/*
 * All the functions following are called from the single workqueue
 * which is shared between all the attached VMB's
 * (this code must be re-entrant)
 *
 * There is only a single kernel thread handling all the inserted VMB's
 *
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_rxthresh_event(void*data)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = data;
#else		
static void vmb_rxthresh_event(struct work_struct*work)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = rxthresh_to_vmb(work);
#endif

	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	UTL_DEBUG(4,vmb);

	u32 count = vmb->rxthresh.kicked - vmb->rxthresh.worked++;	// could get out of step

	u8 rpl = read_rxrplr(vmb);
	u8 rph = read_rxrphr(vmb);
	u16 rp = rpl | (rph << 8);

	u8 wpl = read_rxwplr(vmb);
	u8 wph = read_rxwphr(vmb);
	u16 wp = wpl | (wph << 8);

	u16 length = wp - (rp & 0x3FFF);
	if (rp & 0x8000) length += 1;
	if (rp & 0x4000) length += 1;

	DBG(4, "%s(RX THRESH) kicked=%X worked=%X count=%d rxR = %04X rxW = %04X length=%d\n",__FUNCTION__,vmb->rxthresh.kicked,vmb->rxthresh.worked,count,rp,wp,length);

	vmb_check_rx_fifo(vmb);

	HW_LOCK(vmb);

	vmb->sr_mask &= ~ 0x08;	// RX THRESH
	write_sr(vmb,vmb->sr_mask);
	
	HW_UNLOCK(vmb);

	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);

	vmb_giveback_urbs(vmb);
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_rxto_event(void*data)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = data;	
#else		
static void vmb_rxto_event(struct work_struct*work)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = rxto_to_vmb(work);
#endif

	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	DBG(1, "%s(starting)\n",__FUNCTION__);

	VMB_DEBUG(1,vmb);

	u32 count = vmb->rxto.kicked - vmb->rxto.worked;

	vmb->rxto.worked += 1;

	UTL_ASSERT((int)count > 0);

	DBG(1, "%s(completed)\n",__FUNCTION__);

	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);

	vmb_giveback_urbs(vmb);
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_txthresh_event(void*data)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = data;	
#else		
static void vmb_txthresh_event(struct work_struct*work)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = txthresh_to_vmb(work);
#endif

	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	vmb_check_rx_fifo(vmb);	// must make sure write errors are handled first

	PTR_DEBUG(0,vmb);

	VMB_DEBUG(0,vmb);

	u32 count = vmb->txthresh.kicked - vmb->txthresh.worked;

	int slots = MAX_VMB_SLOTS;
	int n = 0;

	while(slots-- > 0)
	{
		u8 td = n++;
		struct vmb_slot *slot = &vmb->slot[td];
		ASSERT(td == slot->td);
		ASSERT(vmb == slot->vmb);

		if        (vmb->unplugged) {	
			continue;
		} else if (0 == slot->initialized) {
			continue;
		} else if (0 == slot->running) {
			continue;
		} else if (1 == slot->end_seen) {
			continue;
		} else {
			SLOT_ASSERT(slot->allocated);	
			u16 tler = READ_TLER(vmb,td);
	
			if        (0xFFFF == tler) {
				DBG(0, "%s(TD=%X tler reads as 0xFFFF)\n",__FUNCTION__,td);
			} else if (0x0001 & tler) {
				// still active
			} else if (0x0006 & tler){
				UTL_ASSERT(0);	// status failed NAK or STALL or FATAL
			} else if (0x0010 == (0x0038 & tler)){
				slot->end_seen = 1;	// SETUP complete
				DBG(1, "%s(TD=%X SETUP complete)\n",__FUNCTION__,td);
			} else if (0x0000 == (0x0038 & tler)){
				slot->end_seen = 1;	// OUT   complete
				DBG(1, "%s(TD=%X OUT complete)\n",__FUNCTION__,td);
			} else {
			}
		}
	}

	vmb->txthresh.worked += 1;

	UTL_ASSERT((int)count > 0);

	DBG(4, "%s(count=%d kicked=%d worked=%d) completed\n",__FUNCTION__,count,vmb->txthresh.kicked,vmb->txthresh.worked);

	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);

	KICK_SCHEDULER(vmb);

	vmb_giveback_urbs(vmb);
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_ovr_event(void*data)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = data;	
#else		
static void vmb_ovr_event(struct work_struct*work)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = ovr_to_vmb(work);
#endif

	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	DBG(1, "%s(starting)\n",__FUNCTION__);

	VMB_DEBUG(1,vmb);

	u32 count = vmb->ovr.kicked - vmb->ovr.worked;

	vmb->ovr.worked += 1;

	UTL_ASSERT((int)count > 0);

	DBG(1, "%s(completed)\n",__FUNCTION__);

	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);

	vmb_giveback_urbs(vmb);
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_rxerr_event(void*data)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = data;	
#else		
static void vmb_rxerr_event(struct work_struct*work)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = rxerr_to_vmb(work);
#endif

	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	UTL_DEBUG(1,vmb);

	u32 count = vmb->rxerr.kicked - vmb->rxerr.worked;

	DBG(1, "%s(RXERR) kicked=%X worked=%X count=%d\n",__FUNCTION__,vmb->rxerr.kicked,vmb->rxerr.worked,count);

	vmb->rxerr.worked += 1;

	UTL_ASSERT((int)count > 0);

	int S = MAX_VMB_SLOTS;
	int n = 0;
			
 	while(S-- > 0)
	{
		unsigned td = n++;
		struct vmb_slot *slot = &vmb->slot[td];
		ASSERT(td == slot->td);
		ASSERT(vmb == slot->vmb);

		if (0 == slot->initialized) {
			continue; // nothing doing here
		} else if (slot->pipeout) {

			SLOT_ASSERT(slot->allocated);	
			struct urb *urb = slot_to_head_urb(slot);
	
			if (urb) {

				if (VMB_URB_STATE_CONTROL_IN == urb->error_count) {

					u16 tler = READ_TLER(vmb,td);
		
					if        (0x0000 == (0x0001 & tler)) {
						// not active
					} else if (0x0000 == (0x0038 & tler)) {
						// OUT
					} else if (0x0008 == (0x0038 & tler)) {
						// IN
					} else if (0x0010 == (0x0038 & tler)) {
						// SETUP
					} else if (0x0018 == (0x0038 & tler)) {
						// IN ONCE
					} else if (0x0020 == (0x0038 & tler)) {
						// OUT ISO
					} else if (0x0028 == (0x0038 & tler)) {
						// IN ISO
					} else if (0x0038 == (0x0038 & tler)) {
						// Not USED
					} else if (0x0002 == (0x0006 & tler)) {
						UTL_ASSERT(0); // IN SPD - status failed NAK
					} else if (0x0004 == (0x0006 & tler)) {
						td_stalled(vmb,td,urb); // IN SPD - status failed STALL
					} else if (0x0006 == (0x0006 & tler)) {
						UTL_ASSERT(0);	// IN SPD - status failed FATAL
					} else {
						// IN SPD - status is OK
					}
				} else {
					// no STALL's or NAK's on OUT transfers
				}
			} else {
			}

		} else {
			SLOT_ASSERT(slot->allocated);	
			SLOT_ASSERT(0 == slot->pipeout);

			if (0 == slot->running) {	// an rxerr but (probably) not in the slot
	
				continue; // nothing doing here
	
			} else if (0 == slot->actioned) {	// ??????????????????
				SLOT_ASSERT(0); // TO DO - check WHY ??????????????????	
				continue; // nothing doing here
	
			} else {
	
				struct urb *urb = slot_to_head_urb(slot);
		
				if (urb) {
					u16 tler = READ_TLER(vmb,td);
		
					if        (0x0000 == (0x0001 & tler)) {
						// not active
					} else if (0x0000 == (0x0038 & tler)) {
						// OUT
					} else if (0x0008 == (0x0038 & tler)) {
						// IN
					} else if (0x0010 == (0x0038 & tler)) {
						// SETUP
					} else if (0x0018 == (0x0038 & tler)) {
						// IN ONCE
					} else if (0x0020 == (0x0038 & tler)) {
						// OUT ISO
					} else if (0x0028 == (0x0038 & tler)) {
						// IN ISO
					} else if (0x0038 == (0x0038 & tler)) {
						// Not USED
					} else if (0x0002 == (0x0006 & tler)) {
						URB_ASSERT(VMB_URB_STATE_INTERRUPT == urb->error_count);	// IN SPD - status failed NAK
					} else if (0x0004 == (0x0006 & tler)) {
						td_stalled(vmb,td,urb); // IN SPD - status failed STALL
					} else if (0x0006 == (0x0006 & tler)) {
						td_failure(vmb,td,urb); // IN SPD - status failed FATAL
					} else {
						// IN SPD - status is OK
					}
				} else {
				}
			}
		}
	}

	DBG(5, "%s() completed\n",__FUNCTION__);

	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);

	vmb_giveback_urbs(vmb);
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_eot_event(void*data)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = data;	
#else		
static void vmb_eot_event(struct work_struct*work)	// workqueue API triggered from vmb_hcd_irq()
{
	struct vmb *vmb = eot_to_vmb(work);
#endif

	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	u32 count = vmb->eot.kicked - vmb->eot.worked;

	DBG(4, "%s(EOT) kicked=%X worked=%X count=%d\n",__FUNCTION__,vmb->eot.kicked,vmb->eot.worked,count);

	UTL_DEBUG(4,vmb);

	vmb->eot.worked += 1;

	UTL_ASSERT((int)count > 0);

	int S = MAX_VMB_SLOTS;
	int n = 0;
			
 	while(S-- > 0)
	{
		unsigned td = n++;

		struct vmb_slot *slot = &vmb->slot[td];
		ASSERT(td == slot->td);
		ASSERT(vmb == slot->vmb);

		if        (vmb->unplugged) {	
		} else if (0 == slot->initialized) {
			continue; // nothing doing here
		} else if (1 == slot->end_seen) {
			continue;
		} else if (0 == slot->running) {
			continue;
		} else {
			SLOT_ASSERT(slot->allocated);	
			u16 tler = READ_TLER(vmb,td);

			if        (0xFFFF == tler) {
				DBG(0, "%s(TD=%X tler reads as 0xFFFF)\n",__FUNCTION__,td);
			} else if (0x0001 & tler) {
				// still active
			} else if (0x0006 & tler){
				UTL_DEBUG(4,vmb);	// status failed NAK or STALL or FATAL
			} else if (0x0010 == (0x0038 & tler)){
				slot->end_seen = 1;	// SETUP complete
				DBG(1, "%s(TD=%X SETUP complete) queue_work-->vmb.scheduler\n",__FUNCTION__,td);
			} else if (0x0000 == (0x0038 & tler)){
				slot->end_seen = 1;	// OUT   complete
				DBG(1, "%s(TD=%X OUT complete) queue_work-->vmb.scheduler\n",__FUNCTION__,td);
			} else {
			}
		}
	}

	URB_DBG(1, "%s(count=%d kicked=%d worked=%d) completed\n",__FUNCTION__,count,vmb->eot.kicked,vmb->eot.worked);

	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);

	KICK_SCHEDULER(vmb);

	vmb_giveback_urbs(vmb);
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_single_port(void*data)	// workqueue API triggered from port_reset() and vmb_hub_control()
{
	struct vmb_port*vmb_port = data;
#else
static void vmb_single_port(struct work_struct*work)	// workqueue API triggered from port_reset() and vmb_hub_control()
{
	struct vmb_port *vmb_port = container_of((void *) work, struct vmb_port, delayed_work.work);
#endif

	assert_valid_vmb(vmb_port->vmb,__FUNCTION__);
	ASSERT(! in_atomic());

	struct vmb*vmb = port_to_vmb(vmb_port);

	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	DBG(4, "%s(starting) PORT %d\n",__FUNCTION__,vmb_port->number);

	VMB_DEBUG(1,vmb);

	if (vmb_port->reset && time_after_eq(jiffies,vmb_port->finish))
	{
		DBG(4, "%s(TIMEOUT port %d) vmb = %p\n",__FUNCTION__,vmb_port->number,vmb);

		port_enable(vmb,vmb_port->number,1);

		if ((1 == vmb->enumerating) && (vmb_port->number == vmb->current_port) && (VPS_resetting == vmb_port->status))
		{
			vmb_port->status = VPS_enabled;

			vmb_port->Status |= (1 << USB_PORT_FEAT_ENABLE);
			vmb_port->Status &= ~(1 << USB_PORT_FEAT_RESET);
		}
		else
		{
			DBG(4, "%s(port %d) vmb enumerating=%d current_port=%d vmb_port->status=%d\n",__FUNCTION__,vmb_port->number,vmb->enumerating,vmb->current_port,vmb_port->status);
		}
		VMB_DEBUG(1,vmb);
	} else {
		UTL_ASSERT(0);
	}

	DBG(4, "%s() completed\n",__FUNCTION__);

	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);

	vmb_giveback_urbs(vmb);
}
//----------------------------------------------------------------------------------------
//
// workqueue API callback function, triggered by:
//
//           queue_delayed_work(workqueue,&slot->delayed_work,jiffies);
//
// from:  slot_enqueue_new()   VMB_URB_STATE_QUEUED_FOR_ACTION
//        td_stalled()         VMB_URB_STATE_CONTROL_STALL
//        td_stalled()         VMB_URB_STATE_BULK_IN_STALL
//        rx_packet_received() 6
//        rx_packet_received() 1
//
//----------------------------------------------------------------------------------------
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_slot_event(void*data)	// workqueue API
{
	struct vmb_slot*slot = data;
#else
static void vmb_slot_event(struct work_struct*work_struct)	// workqueue API
{
	struct vmb_slot *slot = container_of((void *) work_struct, struct vmb_slot, delayed_work.work);
#endif

	ASSERT(! in_atomic());

	struct vmb *vmb = vmb_slot_to_vmb(slot);

	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	int line_number = slot->line_number;

	slot->line_number = 0;

	if (1 == slot->device_error) {
		ERR("%s(TD=%X called from line_number=%d)\n",__FUNCTION__,slot->td,line_number);
	} else if (1 == slot->initialized) {
		SLOT_ASSERT(slot->allocated);
	
		VMB_DEBUG(5,vmb);
	
		SLOT_ASSERT(SLOT_QUEUE_MASK & (slot->next_urb_index ^ slot->queue_next));	// slot->next_urb_index <> slot->queue_next
		SLOT_ASSERT(0 == slot->next_urb_count);
	
		unsigned td = slot->td;
	
		struct urb *urb = slot_to_head_urb(slot);
	
		if (NULL == urb) {
	
			DBG(1, "%s(TD=%X NO urb)\n",__FUNCTION__,td);
	
			UTL_ASSERT(0);	// just ignore it
	
		} else {
	
			DBG(4, "%s(TD=%X urb=%p error_count=%d)\n",__FUNCTION__,td,urb,urb->error_count);
		
			UTL_ASSERT(vmb == urb_to_vmb(urb));
		
			if (VMB_URB_STATE_CONTROL_STALL == urb->error_count) {
		
				urb_timer_expired(vmb,td,urb);
		
			} else if (VMB_URB_STATE_BULK_IN_STALL == urb->error_count) {
		
				urb_timer_expired(vmb,td,urb);
	
			} else if (VMB_URB_STATE_BULK_IN_FAILED == urb->error_count) {
		
				urb_timer_expired(vmb,td,urb);
	
			} else if (VMB_URB_STATE_INT_FAILED == urb->error_count) {
		
				urb_timer_expired(vmb,td,urb);
	
			} else if (VMB_URB_STATE_CONTROL_IN_FAILED == urb->error_count) {
		
				urb_timer_expired(vmb,td,urb);
	
			} else {
	
				ASSERT(0);	// WHY ???
			}

			DBG(5, "%s(TD=%X urb=%p) completed\n",__FUNCTION__,slot->td, urb);
			URB_DEBUG(5,urb);	
			UTL_DEBUG(5,vmb);
		}
	} else if (1 == slot->allocated) {

		u16 queue_head = slot->queue_next - slot->queue_size;
		struct urb*urb = slot->u_r_b[SLOT_QUEUE_MASK & queue_head];
		ERR("%s(TD=%X queue_size=%d head urb=%p called from line_number=%d)\n",__FUNCTION__,slot->td,slot->queue_size,urb,line_number);
	} else {

		SLOT_ASSERT(0 == slot->queue_size);
		ERR("%s(TD=%X called from line_number=%d)\n",__FUNCTION__,slot->td,line_number);
	}
	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);

	vmb_giveback_urbs(vmb);
}
static int slot_enqueue_new(struct vmb*vmb,struct vmb_slot*slot,struct urb*urb)
{
	ASSERT(0 < in_vmb_scheduler);

	ASSERT(vmb);
	ASSERT(slot);
	ASSERT(urb);

	SW_ASSERT_LOCKED(vmb);

	SLOT_ASSERT(slot->allocated);
	SLOT_ASSERT(0 == slot->initialized);
	URB_ASSERT(0 == urb->error_count);
	ASSERT(vmb == slot->vmb);

	URB_QUEUED(vmb,urb,slot);
	urb->hcpriv = vmb;

	if (PIPE_INTERRUPT == usb_pipetype(urb->pipe)) {

		URB_ASSERT(usb_pipein(urb->pipe));
		URB_ASSERT(0 < urb->interval && 500 > urb->interval); // less than 500ms

		URB_ASSERT(slot->usb_addr > 0);
		URB_ASSERT(slot->usb_endp > 0);

		struct vmb_udev*udev = &vmb->udev[slot->usb_addr];

		URB_ASSERT(udev->usb_device == urb->dev);

		urb->actual_length = 0;

		set_max_plen(vmb,slot->td,urb);

		urb->error_count = VMB_URB_STATE_INT_WAIT;
		slot->actioned = 1;

		struct slot_idref*idref = &slot->idref[SLOT_IDREF_MASK & slot->idref_this++];
		slot->idref_done += 1;
		slot->idref_pkts = 0;

		HW_LOCK(vmb);
		u16 start_fc = read_fc(vmb);    // the frame currently or recently finished executing
		HW_UNLOCK(vmb);

		u16 urb_index = slot->queue_next;
		idref->queue_this = urb_index;
		idref->frame_last = UTL_FRAME_MASK & (start_fc + 100+urb->interval);
		SLOT_ASSERT(NULL == slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next]);
		slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next++] = urb;
		slot->queue_size += 1;

		SLOT_ASSERT(0 == slot->running);
		SLOT_ASSERT(0 == slot->retired);
		u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
		URB_DBG(1,"INTERRUPT %d:%02d urb len=%d maxusbpktlen=%d interval=%d start_fc=%04X frame_last=%04X\n",slot->usb_addr,slot->usb_endp,urb->transfer_buffer_length,maxusbpktlen,urb->interval,start_fc,idref->frame_last);

		UTL_DEBUG(1,vmb);

	} else if (PIPE_ISOCHRONOUS == usb_pipetype(urb->pipe)) {

		URB_ASSERT(slot->usb_addr > 0);
		URB_ASSERT(slot->usb_endp > 0);

		int delay = 5;	// just for the hell of it
		urb->error_count = VMB_URB_STATE_QUEUED_FOR_ACTION;
		slot->jiffies = jiffies + delay;

		SLOT_ASSERT(NULL == slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next]);
		slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next++] = urb;
		slot->queue_size += 1;

		slot->line_number = __LINE__;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
		queue_delayed_work(workqueue,&slot->status_work,delay);
#else
		queue_delayed_work(workqueue,&slot->delayed_work,delay);
#endif

	} else if (PIPE_BULK == usb_pipetype(urb->pipe)) {

		URB_ASSERT(slot->usb_addr > 0);
		URB_ASSERT(slot->usb_endp > 0);

		struct vmb_udev*udev = &vmb->udev[slot->usb_addr];

		URB_ASSERT(udev->usb_device == urb->dev);

		URB_ASSERT(0 == urb->actual_length);
		SLOT_ASSERT(0 == slot->extra);

		if (usb_pipein(urb->pipe)) {

			u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));

			DBG(4,"BULK IN %d:%02d urb->transfer_buffer_length=%d maxusbpktlen=%d\n",slot->usb_addr,slot->usb_endp,urb->transfer_buffer_length,maxusbpktlen);

			urb->error_count = VMB_URB_STATE_BULK_IN;
			urb->actual_length = 0;

			set_max_plen(vmb,slot->td,urb);
		} else {

			urb->error_count = VMB_URB_STATE_BULK_OUT;
		}

		u16 urb_index = slot->queue_next;
		slot->toggle[SLOT_QUEUE_MASK & urb_index] = usb_gettoggle(urb->dev, slot->usb_endp,1);	// only needed to outward bound packets
		SLOT_ASSERT(NULL == slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next]);
		slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next++] = urb;
		slot->queue_size += 1;

		UTL_DEBUG(4,vmb);

		KICK_SCHEDULER(vmb);

	} else {	// PIPE_CONTROL == usb_pipetype(urb->pipe)

		URB_ASSERT(PIPE_CONTROL == usb_pipetype(urb->pipe));
		SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));

		if (slot->usb_addr) {
		} else {

			UTL_ASSERT(vmb->enumerating);

			if (vmb->current_port > 0) {

				struct vmb_port*port = &vmb->port[vmb->current_port-1];

				UTL_ASSERT(VPS_enumerating == port->status);

			} else {

				URB_ASSERT(vmb->current_udev == urb->dev);
				UTL_ASSERT(VPS_monitoring == vmb->port[0].status || VPS_running == vmb->port[0].status);
				UTL_ASSERT(VPS_monitoring == vmb->port[1].status || VPS_running == vmb->port[1].status);
				UTL_ASSERT(VPS_running    == vmb->port[0].status || VPS_running == vmb->port[1].status);
			}
		}

		set_max_plen(vmb,slot->td,urb);
		urb->error_count = VMB_URB_STATE_CONTROL_SETUP;
//......................................................................

		SLOT_ASSERT(NULL == slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next]);
		slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next++] = urb;
		slot->queue_size += 1;

		SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next-slot->queue_size)));

		URB_SETUP(0,urb);

		KICK_SCHEDULER(vmb);
	}

	slot->initialized = 1;

	return 0;
}
static int slot_enqueue_add(struct vmb*vmb,struct vmb_slot*slot,struct urb*urb)
{
	ASSERT(0 < in_vmb_scheduler);

	ASSERT(vmb);
	ASSERT(slot);
	ASSERT(urb);

	SW_ASSERT_LOCKED(vmb);

	SLOT_ASSERT(slot->allocated);

	URB_ASSERT(slot->usb_addr == usb_pipedevice(urb->pipe));
	URB_ASSERT(slot->usb_endp == usb_pipeendpoint(urb->pipe));
	URB_ASSERT(slot->pipetype == usb_pipetype(urb->pipe));
	URB_ASSERT(slot->pipeout  == usb_pipeout(urb->pipe));
	URB_ASSERT(slot->maxusbpktlen == usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe)));

	if (slot->queue_size == SLOT_QUEUE_SIZE) {

		DBG(1, "%s() URB QUEUE FILLED\n",__FUNCTION__);
		UTL_ASSERT(0);

		return -ENOMEM;

	} else if (0 == slot->initialized) {

		SLOT_ASSERT(slot->queue_size > 0);
		SLOT_ASSERT(0 == slot->next_urb_count);

		URB_QUEUED(vmb,urb,slot);
		urb->hcpriv = vmb;

		if (PIPE_BULK == usb_pipetype(urb->pipe)) {

			URB_ASSERT(slot->usb_addr > 0);
			URB_ASSERT(slot->usb_endp > 0);
	
			struct vmb_udev*udev = &vmb->udev[slot->usb_addr];
	
			URB_ASSERT(udev->usb_device == urb->dev);
	
			if (usb_pipein(urb->pipe)) {
	
				u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
	
				DBG(4,"BULK IN %d:%02d urb %p transfer_buffer_length=%d maxusbpktlen=%d\n",slot->usb_addr,slot->usb_endp,urb,urb->transfer_buffer_length,maxusbpktlen);
	
				URB_ASSERT(0 == urb->actual_length);
				SLOT_ASSERT(0 == slot->extra);
	
				urb->error_count = VMB_URB_STATE_BULK_IN;
				urb->actual_length = 0;
	
				SLOT_ASSERT(check_max_plen(vmb,slot->td,urb));
		
			} else {
	
				SLOT_ASSERT(0 == slot->extra);
				URB_ASSERT(0 == urb->actual_length);
	
				urb->error_count = VMB_URB_STATE_BULK_OUT;
	
				UTL_DEBUG(4,vmb);
			}
		} else if (PIPE_ISOCHRONOUS == usb_pipetype(urb->pipe)) {
			ASSERT(0);	// TODO
			return -EINVAL;
		} else if (PIPE_CONTROL == usb_pipetype(urb->pipe)) {

			URB_ASSERT(slot->usb_addr > 0);	// it cannot be enumerating
	
			struct vmb_udev*udev = &vmb->udev[slot->usb_addr];
	
			URB_ASSERT(udev->usb_device == urb->dev);
	
			urb->error_count = VMB_URB_STATE_CONTROL_SETUP;

			if (usb_pipein(urb->pipe)) {
	
				u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
	
				URB_ASSERT(0 == urb->actual_length);
				SLOT_ASSERT(0 == slot->extra);
	
				SLOT_ASSERT(check_max_plen(vmb,slot->td,urb));
		
				URB_DBG(1,"CONTROL IN %d:%02d urb %p transfer_buffer_length=%d maxusbpktlen=%d\n",slot->usb_addr,slot->usb_endp,urb,urb->transfer_buffer_length,maxusbpktlen);
	
				urb->actual_length = 0;
	
			} else {
	
				u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
	
				SLOT_ASSERT(0 == slot->extra);
				URB_ASSERT(0 == urb->actual_length);
	
				URB_DBG(1,"CONTROL OUT %d:%02d urb %p transfer_buffer_length=%d maxusbpktlen=%d\n",slot->usb_addr,slot->usb_endp,urb,urb->transfer_buffer_length,maxusbpktlen);
	
			}
			URB_SETUP(0,urb);
			UTL_DEBUG(4,vmb);
		} else {
			UTL_ASSERT(0);	// expect only PIPE_BULK || PIPE_ISOCHRONOUS || PIPE_CONTROL
			return -EINVAL;
		}
		slot->queue_size += 1;
		SLOT_ASSERT(slot->queue_size > (SLOT_QUEUE_MASK & (slot->queue_next - slot->next_urb_index)));
		SLOT_ASSERT(NULL == slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next]);
		slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next++] = urb;
		KICK_SCHEDULER(vmb);	// this can now be done synchronously --- TODO
		return 0;

	} else if (slot->queue_size > 0) {

		// slot->recieved might still be zero if the urb enqueue's happen very quickly

		URB_QUEUED(vmb,urb,slot);
		urb->hcpriv = vmb;

		if (PIPE_BULK == usb_pipetype(urb->pipe)) {

			URB_ASSERT(slot->usb_addr > 0);
			URB_ASSERT(slot->usb_endp > 0);
	
			struct vmb_udev*udev = &vmb->udev[slot->usb_addr];
	
			URB_ASSERT(udev->usb_device == urb->dev);
	
			if (usb_pipein(urb->pipe)) {
	
				u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
	
				DBG(4,"BULK IN %d:%02d urb %p transfer_buffer_length=%d maxusbpktlen=%d\n",slot->usb_addr,slot->usb_endp,urb,urb->transfer_buffer_length,maxusbpktlen);
	
				URB_ASSERT(0 == urb->actual_length);
				SLOT_ASSERT(0 == slot->extra);
	
				urb->error_count = VMB_URB_STATE_BULK_IN;
				urb->actual_length = 0;
	
				SLOT_ASSERT(check_max_plen(vmb,slot->td,urb));
		
			} else {
	
				SLOT_ASSERT(0 == slot->extra);
				URB_ASSERT(0 == urb->actual_length);
	
				urb->error_count = VMB_URB_STATE_BULK_OUT;
	
				UTL_DEBUG(4,vmb);
			}
		} else if (PIPE_ISOCHRONOUS == usb_pipetype(urb->pipe)) {
			ASSERT(0);	// TODO
			return -EINVAL;
		} else if (PIPE_CONTROL == usb_pipetype(urb->pipe)) {

			URB_ASSERT(slot->usb_addr > 0);	// it cannot be enumerating
	
			struct vmb_udev*udev = &vmb->udev[slot->usb_addr];
	
			URB_ASSERT(udev->usb_device == urb->dev);
	
			urb->error_count = VMB_URB_STATE_CONTROL_SETUP;

			if (usb_pipein(urb->pipe)) {
	
				u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
	
				URB_ASSERT(0 == urb->actual_length);
				SLOT_ASSERT(0 == slot->extra);
	
				SLOT_ASSERT(check_max_plen(vmb,slot->td,urb));
		
				urb->actual_length = 0;
	
				URB_DBG(0,"CONTROL IN %d:%02d urb %p transfer_buffer_length=%d maxusbpktlen=%d\n",slot->usb_addr,slot->usb_endp,urb,urb->transfer_buffer_length,maxusbpktlen);
	
			} else {
	
				u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));
	
				SLOT_ASSERT(0 == slot->extra);
				URB_ASSERT(0 == urb->actual_length);
	
				URB_DBG(1,"CONTROL OUT %d:%02d urb %p transfer_buffer_length=%d maxusbpktlen=%d\n",slot->usb_addr,slot->usb_endp,urb,urb->transfer_buffer_length,maxusbpktlen);
	
			}
			UTL_DEBUG(4,vmb);
		} else {
			UTL_ASSERT(0);	// expect only PIPE_BULK || PIPE_ISOCHRONOUS || PIPE_CONTROL
			return -EINVAL;
		}
		slot->queue_size += 1;
		SLOT_ASSERT(slot->queue_size > (SLOT_QUEUE_MASK & (slot->queue_next - slot->next_urb_index)));
		SLOT_ASSERT(NULL == slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next]);
		slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next++] = urb;
		KICK_SCHEDULER(vmb);	// this can now be done synchronously --- TODO
		return 0;

	} else {	// 0 == slot->queue_size

		SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
		SLOT_ASSERT(0 == slot->next_urb_count);

		urb->hcpriv = vmb;

		URB_QUEUED(vmb,urb,slot);

		slot->jiffies = jiffies;

		if (PIPE_CONTROL == usb_pipetype(urb->pipe)) {
			UTL_ASSERT(0);	// TO DO
			return -EINVAL;
		} else if (PIPE_INTERRUPT == usb_pipetype(urb->pipe)) {
			UTL_ASSERT(0);	// TO DO
			return -EINVAL;
		} else if (PIPE_ISOCHRONOUS == usb_pipetype(urb->pipe)) {
			UTL_ASSERT(0);	// TO DO
			return -EINVAL;
		} else if (usb_pipeout(urb->pipe)) { // BULK OUT
			// slot->recieved might be zero if the error occured at start of xfer	
			URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
			SLOT_ASSERT(0 == slot->idref_done);
			SLOT_ASSERT(0 == slot->extra);
			SLOT_ASSERT(1 == slot->device_error);
			URB_ASSERT(0 == urb->actual_length);	// not sure ??

			SLOT_ASSERT(1 == slot->actioned);
			SLOT_ASSERT(0 == slot->retry_flag);

			SLOT_ASSERT(0 == slot->next_urb_count);
			SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));

			slot->actioned = 0;
			if (1 == slot->retired) {
				u16 tler = READ_TLER(vmb,slot->td);
				ASSERT(0 == (0x0001 & tler));
				WRITE_TLER(vmb,slot->td,0x00);        // clear out the TD

				SLOT_ASSERT(0 == slot->running);
				SLOT_ASSERT(1 == slot->end_seen);
				slot->retired = 0;	
				slot->end_seen = 0;	
				slot->device_error = 0;
			} else {
				SLOT_ASSERT(1 == slot->running);
				SLOT_ASSERT(0 == slot->end_seen);
				SLOT_ASSERT(0 == slot->actioned);
				slot->retry_flag = 1;
				slot->device_error = 0;
			}
			urb->error_count = VMB_URB_STATE_BULK_OUT;

			UTL_DEBUG(4,vmb);

			SLOT_ASSERT(NULL == slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next]);
			slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next++] = urb;
			slot->queue_size += 1;

			slot->line_number = __LINE__;

			KICK_SCHEDULER(vmb);	// this can now be done synchronously --- TODO
			return 0;
		} else {	// BULK IN
			URB_ASSERT(PIPE_BULK == usb_pipetype(urb->pipe));
			SLOT_ASSERT(0 == slot->idref_done);	// not used for BULK IN
			SLOT_ASSERT(slot->recieved > 0);
			urb->error_count = VMB_URB_STATE_BULK_OVERSPILL;
			SLOT_ASSERT(NULL == slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next]);
			slot->u_r_b[SLOT_QUEUE_MASK & slot->queue_next++] = urb;
			slot->queue_size += 1;
	
			slot->line_number = __LINE__;

			urb_overspill(vmb,slot,urb);

			return 0;
		}
	}
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_en_queue_event(void*data)	// workqueue API triggered from vmb_urb_enqueue()
{
	struct vmb *vmb = data;
#else
static void vmb_en_queue_event(struct work_struct*work)	// workqueue API triggered from vmb_urb_enqueue()
{
	struct vmb *vmb = urb_en_queue_to_vmb(work);
#endif

	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	DBG(1, "WORK:%s(en_queue_size=%d)\n",__FUNCTION__,vmb->en_queue_size);
	
	UTL_DEBUG(1,vmb);

	while(vmb->en_queue_size > 0) {

		int status = 0;

		HW_LOCK(vmb);
		struct urb*urb = vmb->en_queue[URB_EN_QUEUE_MASK & (vmb->en_queue_next - vmb->en_queue_size--)];
		HW_UNLOCK(vmb);

		ASSERT(urb);
		URB_ASSERT(0 == urb->error_count);

		u8 usb_addr = usb_pipedevice(urb->pipe);
		u8 usb_endp = usb_pipeendpoint(urb->pipe);

		if (usb_pipeout(urb->pipe) && (PIPE_BULK == usb_pipetype(urb->pipe))) {
			char msg[2+4*16];
			int n = (sizeof(msg)/sizeof(char) - 2)/4;
			int len = min(urb->transfer_buffer_length,n);
			char*m=msg;
			int B = len;
			int A = len;
			u8*b = urb->transfer_buffer;
			u8*a = urb->transfer_buffer;

			while(B-- > 0) {
				m += sprintf(m," %02X",*b++);
			}	
			m += sprintf(m," ");
			while(A-- > 0) {
				char c = *a++;	
				m += sprintf(m,"%c",isprint(c)?c:'?');
			}	
			*m = 0;

			//INFO("%d bytes BULK OUT =%s\n",urb->transfer_buffer_length,msg);
		}
		if (0 == usb_addr) { // enumerating

			struct vmb_slot *slot = find_empty_vmb_slot(vmb,urb);

			if (slot)
			{
				if (0 == vmb->enumerating) {
		
					ASSERT(0 == vmb->current_port);
		
					vmb->enumerating = 1;
		
					ASSERT(NULL == vmb->current_udev);
		
					u8 u = MAX_DEV_ADDRS;
		
					while(u-- > 0)
					{
						if (urb->dev == vmb->udev[u].usb_device) {
		
							u8 portnumber = vmb->udev[u].portnumber;
							if (portnumber > 0) {	
								DBG(0, "%s(re-ENUMERATION starting for device %d on port %d)\n",__FUNCTION__,u,portnumber);
							} else {
								DBG(0, "%s(re-ENUMERATION starting for device %d)\n",__FUNCTION__,u);
							}
		
							usb_put_dev(vmb->udev[u].usb_device);
							vmb->udev[u].usb_device = NULL;
							vmb->udev[u].portnumber = 0;
							vmb->udev[u].devicefail = 0;
						}
					}
		
					vmb->current_udev = urb->dev;
					usb_get_dev(vmb->current_udev);	// increment reference count
		
				} else if (0 < vmb->current_port) {
		
					ASSERT(NULL == vmb->current_udev);
		
					struct vmb_port*port = &vmb->port[vmb->current_port - 1];
		
					ASSERT(VPS_enumerating == port->status);
		
					if (port->udev) {
						URB_ASSERT(port->udev == urb->dev);
					} else {
						port->udev = urb->dev;
						usb_get_dev(port->udev);	// increment reference count
					}
		
				} else {
		
					URB_ASSERT(vmb->current_udev == urb->dev);
				}
	
				status = slot_enqueue_new(vmb,slot,urb);
			}
			else
			{
				ASSERT(0 == vmb->slot_count);
				ERR("%s() RAN OUT OF SLOTS\n",__FUNCTION__);
				UTL_ASSERT(0);

				status = -ENOMEM;
			}
	
		} else if (usb_addr < MAX_DEV_ADDRS) { // enumeration is over
	
			struct vmb_udev*udev = &vmb->udev[usb_addr];
			u8 pipeout = usb_pipeout(urb->pipe);

			if (udev->devicefail) {
				ERR("%s(DEVICE NOT RESPONDING) **ERROR**\n",__FUNCTION__);
				status = -ENODEV;	
			} else if (NULL == udev->usb_device) { // first access to device after enumeration
				UTL_ASSERT(NULL == find_matching_vmb_slot(vmb,usb_addr,usb_endp,pipeout));
	
				struct vmb_slot *slot = find_empty_vmb_slot(vmb,urb);
			
				if (slot)
				{
					ASSERT(1 == vmb->enumerating);
		
					if (0 < vmb->current_port) {
		
						ASSERT(NULL == vmb->current_udev);
		
						struct vmb_port*port = &vmb->port[vmb->current_port - 1];
		
						ASSERT(VPS_enumerating == port->status);
		
						DBG(0, "%s(ENUMERATION finished for device %d on port %d)\n",__FUNCTION__,usb_addr,port->number);
		
						if (port->udev) {
							URB_ASSERT(port->udev == urb->dev);
						} else {
							port->udev = urb->dev;
							usb_get_dev(port->udev);	// increment reference count
						}
		
						vmb->udev[usb_addr].portnumber = vmb->current_port; // directly connected
						vmb->udev[usb_addr].devicefail = 0;
						vmb->udev[usb_addr].usb_device = urb->dev;
						usb_get_dev(vmb->udev[usb_addr].usb_device);
		
						port->status = VPS_running;
						vmb->current_port = 0;
		
					} else {
		
						DBG(0, "%s(ENUMERATION finished for device %d)\n",__FUNCTION__,usb_addr);
		
						ASSERT(VPS_monitoring == vmb->port[0].status || VPS_running == vmb->port[0].status);
						ASSERT(VPS_monitoring == vmb->port[1].status || VPS_running == vmb->port[1].status);
						ASSERT(VPS_running    == vmb->port[0].status || VPS_running == vmb->port[1].status);
		
						URB_ASSERT(vmb->current_udev == urb->dev);
		
						usb_put_dev(vmb->current_udev);
						vmb->current_udev = NULL;
		
						vmb->udev[usb_addr].portnumber = 0; // connected via a hub
						vmb->udev[usb_addr].devicefail = 0;
						vmb->udev[usb_addr].usb_device = urb->dev;
						usb_get_dev(vmb->udev[usb_addr].usb_device);
					}
		
					vmb->enumerating = 0;
	
					status = slot_enqueue_new(vmb,slot,urb);
				}
				else
				{
					ASSERT(0 == vmb->slot_count);
					ERR("%s() RAN OUT OF SLOTS\n",__FUNCTION__);
					UTL_ASSERT(0);

					status = -ENOMEM;
				}
			} else {  // existing device - have accessed the device previously
	
				URB_ASSERT(udev->usb_device == urb->dev);
	
				if (0 < udev->portnumber) {
	
					struct vmb_port*port = &vmb->port[udev->portnumber - 1];
	
					ASSERT(VPS_running == port->status);
					ASSERT(port->udev == udev->usb_device);
				}
	
				struct vmb_slot *slot = find_matching_vmb_slot(vmb,usb_addr,usb_endp,pipeout);
	
				if (slot) {
	
					status = slot_enqueue_add(vmb,slot,urb);
	
				} else if ( (slot = find_empty_vmb_slot(vmb,urb)) ) {
	
					status = slot_enqueue_new(vmb,slot,urb);
	
				} else {
	
					ASSERT(0 == vmb->slot_count);
					ERR("%s() RAN OUT OF SLOTS\n",__FUNCTION__);
					UTL_ASSERT(0);

					status = -ENOMEM;
				}
			}
	
		} else { // the VMB only has 3 bits for the device address
	
			ERR("%s() MAXIMUM DEVICES REACHED\n",__FUNCTION__);

			status = -EPIPE;
		}
	
		if (status) {
			ASSERT(0 > status);	// should be a negative error code	
			urb->actual_length = 0;
			urb->error_count = VMB_URB_STATE_RETURNED_TO_CALLER;
			urb->status = status;
			urb->hcpriv = NULL;
			URB_DBG(0, "%s(urb_hcd_giveback_urb=%p status=%d)\n",__FUNCTION__,urb,status);
			vmb_urb_rejected(vmb,urb);

		} else {
		}

		// TODO - need to add a long stop timer callback
		//        to cover the hole in which a urb has been
		//        added to the queue after the previous has
		//        been dealt with but before this work queue
		//        function has exited and the queueing of a
		//        new event failed.
	}

	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);

	vmb_giveback_urbs(vmb);
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_de_queue_event(void*data)	// workqueue API triggered from vmb_urb_dequeue()
{
	struct vmb *vmb = data;
#else	
static void vmb_de_queue_event(struct work_struct*work)	// workqueue API triggered from vmb_urb_dequeue()
{
	struct vmb *vmb = urb_de_queue_to_vmb(work);
#endif

	/*
	 * how do we stop an OUT ????
	 *
	 * - I think we have to wait until the txfifo flushes the relevant urb's
	 *
	 * - TODO ---> need to have a "dequeuing" state on urb's in the txfifo
	 */

	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	URB_DBG(1, "%s(de_queue_size=%d)\n",__FUNCTION__,vmb->de_queue_size);
	
	UTL_DEBUG(1,vmb);

	while(vmb->de_queue_size > 0) {

		HW_LOCK(vmb);
		struct urb*urb = vmb->de_queue[URB_DE_QUEUE_MASK & (vmb->de_queue_next - vmb->de_queue_size--)];
		HW_UNLOCK(vmb);

		ASSERT(urb);

		struct vmb_slot *slot = urb_to_slot(urb);
	
		if (slot) {

			SLOT_DEBUG(1,slot);	
			SLOT_ASSERT(0 < slot->queue_size);	
			SLOT_ASSERT(slot->allocated);	
			u8 td = slot->td;

			URB_DBG(0, "%s(de_queue_size=%d TD=%X urb=%p running=%d actioned=%d)\n",__FUNCTION__,vmb->de_queue_size,td,urb,slot->running,slot->actioned);
	
			UTL_ASSERT(vmb == urb_to_vmb(urb));
			DBG(1,"%s:%d(DEQUEUE urb=%p TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X)\n"
				,__FUNCTION__,__LINE__,urb,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size);
			URB_DEBUG(1,urb);

			if (slot_to_head_urb(slot) == urb) {
	
				WRITE_TLER(vmb,td,0x00);        // clear out the TD
				//u16 tler = READ_TLER(vmb,td);
				//WRITE_TLER(vmb,td,tler & 0xFFFE); // clear out the TD
				slot->running = 0;	// being dequeued
				slot->retired = 1;
		
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
				cancel_delayed_work(&slot->status_work);
	#else
				cancel_delayed_work(&slot->delayed_work);
	#endif
	
				urb->actual_length = 0;
	
				if ((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & (slot->queue_next - slot->queue_size))) {
					slot->next_urb_index += 1;
					slot->next_urb_count = 0;
				}
	
				if (VMB_URB_COMPLETED(vmb,slot,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0)) {
					SLOT_ASSERT(0 == slot->next_urb_count);
					URB_DBG(4,"%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) DEQUEUE urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
	
					UTL_DEBUG(5,vmb);
					slot->idref_done = 0;
					slot->actioned = 1;
					slot->usb_addr = 0;
					slot->usb_endp = 0;
					slot->extra = 0;
					slot->jiffies = jiffies;
	
					free_vmb_slot(vmb,slot);
		
					clear_max_plen(vmb,td);
				} else {
					URB_DBG(4,"%s(TD=%X next_urb_count=%04X next_urb_index=%04X queue_next=%04X queue_size=%04X) DEQUEUE urb=%p\n",__FUNCTION__ ,slot->td,slot->next_urb_count,slot->next_urb_index,slot->queue_next,slot->queue_size,urb);
					SLOT_ASSERT(slot->queue_size > 0);
					slot->jiffies = jiffies;
				}
		
				VMB_DEBUG(1,vmb);
				UTL_DEBUG(1,vmb);

			} else {	// we need to remove a urb that is not at the head of the list
				u16 queue_item = slot->queue_next-- - --slot->queue_size;
				u8 queue_size = slot->queue_size;
				URB_ASSERT(queue_size > 0);

				struct urb **urbp = NULL;
				while(queue_size-- > 0) {
					urbp = &slot->u_r_b[SLOT_QUEUE_MASK & queue_item];
					if (urb == slot->u_r_b[SLOT_QUEUE_MASK & queue_item++]) {
						while(queue_size-- > 0) {
							*urbp = slot->u_r_b[SLOT_QUEUE_MASK & queue_item];
							urbp = &slot->u_r_b[SLOT_QUEUE_MASK & queue_item++];
						}
						break;
					} else {
						continue;
					}
				}
				*urbp = NULL;

				urb->actual_length = 0;

				URB_DBG(0,"%s(%s:%d is stacking urb=%p from TD=%X onto giveback queue) queue size = %d\n",__FUNCTION__,__FUNCTION__,__LINE__,urb,slot->td,vmb->giveback_size);

				vmb_urb_giveback(__FUNCTION__,__LINE__,vmb,urb,VMB_URB_STATE_RETURNED_TO_CALLER,0);

				VMB_DEBUG(1,vmb);
				UTL_DEBUG(1,vmb);
			}
		} else {

			/*
			 * we have almost certainly just suffered a race condition
			 * between usb_hcd_giveback_urb(urb) and vmb_urb_dequeue(urb)
			 */
			URB_DBG(0, "%s(urb=%p) failed to find slot\n",__FUNCTION__,urb);
		}
	}

	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);

	vmb_giveback_urbs(vmb);
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void vmb_diagnostics_event(void*data)	// workqueue API triggered from initialise_vmb(+5000) and here
{
	struct vmb *vmb = data;	
#else
static void vmb_diagnostics_event(struct work_struct*work_struct)	// workqueue API triggered from initialise_vmb(+5000) and here
{
	struct vmb *vmb = container_of((void *) work_struct, struct vmb, diagnostics.work);
#endif

	assert_valid_vmb(vmb,__FUNCTION__);
	ASSERT(! in_atomic());

	ASSERT(0==in_vmb_scheduler++);
	SW_LOCK(vmb);

	//UTL_DEBUG(0,vmb);
	QUEUE_DEBUG(0,vmb);

	queue_delayed_work(workqueue,&vmb->diagnostics,msecs_to_jiffies(5000));

	KICK_SCHEDULER(vmb);

	SW_UNLOCK(vmb);
	ASSERT(--in_vmb_scheduler==0);
}
